/*
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: auth.c,v 1.7 1995/02/23 07:44:34 janssen Exp $ */

#include <iluntrnl.h>
#include "authschm.h"

/*L1, L2, Main unconstrained*/

/***********************************************************************/
/***********************************************************************/
/***  The Passport struct					     ***/
/***********************************************************************/
/***********************************************************************/

struct _ilu_Passport_s {

  ilu_cardinal nCreds;
  ilu_Credential *Creds;
};

/***********************************************************************/
/***********************************************************************/
/***  Functions on Passports and Credentials			     ***/
/***********************************************************************/
/***********************************************************************/

ilu_Principal
  ilu_GetPrincipal (RETAIN(ilu_Passport) passport,
		    ilu_AuthenticationScheme auth_scheme,
		    ILU_ERRS((NoCredentialForAuthScheme, BadCredential)) *err)
/* returns the user description of "p", using credentials for the
   authentication scheme specified by "auth_scheme" */
{
  ilu_Principal p;
  int i;
  ilu_Error err2 = NIL;

  p.ap_auth_scheme = NIL;
  p.ap_user_id = NIL;
  for (i = 0;  i < passport->nCreds;  i++)
    if (passport->Creds[i]->cr_auth_scheme == auth_scheme)
      {
	err2 = authscheme_get_user_id(auth_scheme, passport->Creds[i], &p.ap_user_id);
	ILU_ERR_SWITCH(err2) {
	  ILU_SUCCESS_CASE
	    p.ap_auth_scheme = auth_scheme;
	  ILU_ERR_CASE(BadCredential, e) {
	    *err = err2;
	  };
	} ILU_ERR_ENDSWITCH;
      }
  if (p.ap_auth_scheme == NIL)
    {
      ILU_BIND_ERR(NoCredentialForAuthScheme, err, e) {
	e->as = auth_scheme;
	e->p = passport;
      } ILU_END_BIND_ERR;
    }
  return (p);
}

ilu_Principal ilu_PrincipalOfCredential (RETAIN(ilu_Credential) cred, ILU_ERRS((BadCredentials)) *err)
/* return principal for the given credential */
{
  ilu_Principal p;

  p.ap_auth_scheme = cred->cr_auth_scheme;
  *err = authscheme_get_user_id(cred->cr_auth_scheme, cred, &p.ap_user_id);
  return (p);  
}

ilu_Passport ilu_CreatePassport ()
/* Create and return new Passport */
{
  ilu_Passport p = (ilu_Passport) ilu_must_malloc(sizeof(struct _ilu_Passport_s));
  p->nCreds = 0;
  p->Creds = NIL;
  return (p);
}

void ilu_DestroyPassport (ilu_Passport p)
{
  if (p->Creds != NIL)
    {
      while (p->nCreds-- > 0)
	{
	  if (p->Creds[p->nCreds] != NIL)
	    ilu_free(p->Creds[p->nCreds]);
	}
      ilu_free(p->Creds);
    }
  ilu_free(p);
}

ilu_Passport ilu_AddCredential (ilu_Passport p, PASS(ilu_Credential) cred)
/* adds the opaque credential in "cred" to "p". */
{
  if (p->Creds == NIL)
    p->Creds = (ilu_Credential *) ilu_must_malloc(sizeof(ilu_Credential *) * (p->nCreds + 1));
  else
    p->Creds = (ilu_Credential *) ilu_realloc(p->Creds, sizeof(ilu_Credential *) * (p->nCreds + 1));
  p->Creds[p->nCreds] = cred;
  p->nCreds++;
  return (p);
}

ilu_Credential ilu_FindCredential (ilu_Passport p, ilu_AuthenticationScheme auth_scheme)
/* returns NIL if "p" contains no credential of the right type */
{
  int i;

  for (i = 0;  i < p->nCreds;  i++)
    if (p->Creds[i]->cr_auth_scheme == auth_scheme)
      return (p->Creds[i]);
  return (NIL);
}

/***********************************************************************/
/***********************************************************************/
/***  Registration and lookup of authentication schemes		     ***/
/***********************************************************************/
/***********************************************************************/

static HashTable /* auth_scheme_name -> auth_scheme */ AuthenticationSchemes = NIL;

ILU_ERRS((AuthenticationSchemeAlreadyRegistered)) ilu_RegisterAuthenticationScheme(PASS(ilu_AuthenticationScheme) as)
{
  ilu_AuthenticationScheme a;
  ilu_Error err;

  if (AuthenticationSchemes == NIL)
    {
      _ilu_Assert((AuthenticationSchemes = _ilu_hash_MakeNewTable(3, NULLFN, NULLFN)) != NIL,
		  "ilu_RegisterAuthenticationScheme MakeNewTable AuthenticationSchemes");
    }
  if ((a = _ilu_hash_FindInTable(AuthenticationSchemes, authscheme_name(as))) != NIL)
    {
      ILU_BIND_ERR(AuthenticationSchemeAlreadyRegistered, &err, e) {
	e->as = a;
      } ILU_END_BIND_ERR;
      return (err);
    }
  else
    {
      _ilu_Assert(_ilu_hash_AddToTable (AuthenticationSchemes, authscheme_name(as), as),
		  "ilu_RegisterAuthenticationScheme AddToTable AuthenticationSchemes");
      return (NULL);
    }
}

#ifdef USE_SUNRPC_UNIX_AUTHENTICATION
extern struct _ilu_AuthenticationScheme_s _ilu_AuthenticationScheme_SunRPCUnix;
#endif /* def USE_SUNRPC_UNIX_AUTHENTICATION */

OPTIONAL(ilu_AuthenticationScheme) ilu_FindAuthenticationSchemeByName (RETAIN(char *) name)
{

#ifdef USE_SUNRPC_UNIX_AUTHENTICATION
  {
    static initialized = ilu_FALSE;
    if (! initialized)
      {
	ilu_RegisterAuthenticationScheme (&_ilu_AuthenticationScheme_SunRPCUnix);
	initialized = ilu_TRUE;
      }
  }
#endif /* def USE_SUNRPC_UNIX_AUTHENTICATION */

  if (AuthenticationSchemes == NIL)
    return (NIL);
  return (_ilu_hash_FindInTable(AuthenticationSchemes, name));
}

/***********************************************************************/
/***********************************************************************/
/***  Errors							     ***/
/***********************************************************************/
/***********************************************************************/

ILU_DEF_ERR(BadCredential, "Credential struct is malformed or corrupted") {
  /* nothing to do */
}

ILU_QUADEF(BadCredential);

ILU_DEF_ERR(NoCredentialForAuthScheme, "Passport contains no credential for specified authentication scheme")
{
  /* nothing to do */
}

ILU_QUADEF(NoCredentialForAuthScheme);

ILU_DEF_ERR(AuthenticationSchemeAlreadyRegistered, "The specified authentication scheme has already been registered")
{
  /* nothing to do */
}

ILU_QUADEF(AuthenticationSchemeAlreadyRegistered);

ILU_DEF_ERR(BadUserID, "The user id is a malformed user id for the given authentication scheme")
{
  /* nothing to free */
}

ILU_QUADEF(BadUserID);

ILU_DEF_ERR(CantCreateCredential, "The user id cannot be proven")
{
  /* nothing to free */
}

ILU_QUADEF(CantCreateCredential);

