/*
Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: courier.c,v 1.68 1994/12/06 23:41:24 spreitze Exp $ */
/* Last edited by Mike Spreitzer December 6, 1994 3:41 pm PST */

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluntrnl.h"
#include "packet.h"
#include "courier.h"

#include "call.h"
#include "protocol.h"
#include "connect.h"
#include "transprt.h"
#include "port.h"
#include "object.h"
#include "type.h"
#include "method.h"
#include "mooring.h"

/*************************************************************/
/*		Implement XDR PACKET			     */
/*************************************************************/

/*L1, Main unconstrained*/
/*L2 >= call's conn's iomu*/

#define ODD(x)		(((x)&0x1)!=0)
#define EVEN(x)		(((x)&0x1)==0)  
#define PADDED_SIZE(x)	((((unsigned) (x))+1) & (~0x1))
  
static unsigned long endianNessTest = 0x01020304;
#ifdef BIG_ENDIAN
#undef BIG_ENDIAN
#endif
#define BIG_ENDIAN	((((unsigned char *)(&endianNessTest))[0])==0x01)

static ilu_boolean _cdr_get_u_short (PACKET p, ilu_shortcardinal *l)
{
  if (packet_allocation(p) - packet_size(p) < 2)
    return(ilu_FALSE);
  *l = (p->next[0] << 8) + p->next[1];
  p->next += 2;
  return (ilu_TRUE);
}

static ilu_boolean _cdr_get_u_long (PACKET p, ilu_cardinal *l)
{
  return (_cdr_get_u_short(p, &((ilu_shortcardinal *) l)[BIG_ENDIAN ? 0 : 1]) &&
	  _cdr_get_u_short(p, &((ilu_shortcardinal *) l)[BIG_ENDIAN ? 1 : 0]));
}

static ilu_boolean _cdr_get_long (PACKET p, ilu_integer *l)
{
  return (_cdr_get_u_long (p, (ilu_cardinal *) l));
}

static ilu_boolean _cdr_get_short (PACKET p, short *l)
{
  return (_cdr_get_u_short(p, (ilu_shortcardinal *) l));
}

static ilu_boolean _cdr_get_byte (PACKET p, char *l)
{
  _ilu_Assert (0, "Attempt to input 1 byte octet in Courier!");
  return (ilu_FALSE);
}

static ilu_boolean _cdr_get_u_byte (PACKET p, ilu_byte *l)
{
  _ilu_Assert (0, "Attempt to input 1 byte octet in Courier!");
  return (ilu_FALSE);
}

static ilu_boolean _cdr_get_float(PACKET p, float *l)
{
  return (_cdr_get_u_long (p, (ilu_cardinal *) l));
}

static ilu_boolean _cdr_get_double (PACKET p, double *l)
{
  return (_cdr_get_u_long (p, ((ilu_cardinal *) l) + (BIG_ENDIAN ? 0 : 1))
	  && _cdr_get_u_long(p, ((ilu_cardinal *) l) + (BIG_ENDIAN ? 1 : 0)));
}

static ilu_cardinal CourierMaxStringSize = 100000;

static ilu_boolean _cdr_get_bytes(PACKET p, ilu_byte **b, ilu_cardinal *l, ilu_cardinal limit)
{
  ilu_shortcardinal l3;
  ilu_cardinal l2;
  register ilu_byte *p1, *p2;
  register ilu_cardinal c;

  if (limit != 0 && limit <= 0xFFFF)
    {	/* short (native Courier) string */
      if ((!_cdr_get_u_short (p, &l3))
	  || packet_allocation(p) - packet_size(p) < PADDED_SIZE(l3))
	return (ilu_FALSE);
      l2 = l3;
    }
  else
    {
      if ((!_cdr_get_u_long (p, &l2))
	  || packet_allocation(p) - packet_size(p) < PADDED_SIZE(l2))
	return (ilu_FALSE);
    }

  *l = 0;
  if (l2 > CourierMaxStringSize OR (limit > 0 AND l2 > limit))
    {
      fprintf (stderr,
	       "%s %lu, which exceeds CourierMaxStringSize value of %lu or call limit of %lu.\n",
	       "Attempt to read string of length", l2,
	       CourierMaxStringSize, limit);
      return (ilu_FALSE);
    }

  {
    register ilu_integer i;

    if (*b == NULL)
      *b = (ilu_byte *) ilu_malloc(l2 + 1);
    for (p1 = *b, p2 = p->next, c = l2;  c > 0;  c--)
      *p1++ = *p2++;
    for (i = PADDED_SIZE(l2) - l2;  i > 0;  i--)
      p2++;	/* eat padding characters */
    *p1 = '\0';
    p->next = p2;
    *l = l2;
    return (ilu_TRUE);
  }
}

static ilu_boolean _cdr_get_opaque (PACKET p, ilu_byte **b, ilu_cardinal len)
{
  register ilu_byte *p1, *p2;
  register ilu_cardinal c;
  register ilu_integer i;

  if (packet_allocation(p) - packet_size(p) < PADDED_SIZE(len))
    return (ilu_FALSE);

  if (*b == NULL)
    *b = (ilu_byte *) ilu_malloc(len);
  for (p1 = *b, p2 = p->next, c = len;  c > 0;  c--)
    *p1++ = *p2++;
  for (i = PADDED_SIZE(len) - len;  i > 0;  i--)
    p2++;	/* eat padding characters */
  p->next = p2;
    return (ilu_TRUE);
}

static void _cdr_destroy (PACKET p)
{
  if (p->buf != NULL && p->count > 0)
    ilu_free(p->buf);
  ilu_free(p);
}

static PACKET _cdr_CreatePacket (ilu_bytes buf, ilu_cardinal size)
{
  PACKET p = (PACKET) ilu_malloc (sizeof(struct ilu_packet_s));
  if (buf != NULL)
    p->buf = buf;
  else if (size > 0)
    p->buf = (ilu_byte *) ilu_malloc(size);
  else
    p->buf = NULL;
  p->count = size;
  p->next = p->buf;

  p->put_long = NULLFN;
  p->put_byte = NULLFN;
  p->put_short = NULLFN;
  p->put_u_long = NULLFN;
  p->put_u_byte = NULLFN;
  p->put_u_short = NULLFN;
  p->put_float = NULLFN;
  p->put_double = NULLFN;
  p->put_bytes = NULLFN;
  p->put_opaque = NULLFN;

  p->get_long = _cdr_get_long;
  p->get_byte = _cdr_get_byte;
  p->get_short = _cdr_get_short;
  p->get_u_long = _cdr_get_u_long;
  p->get_u_byte = _cdr_get_u_byte;
  p->get_u_short = _cdr_get_u_short;
  p->get_float = _cdr_get_float;
  p->get_double = _cdr_get_double;
  p->get_bytes = _cdr_get_bytes;
  p->get_opaque = _cdr_get_opaque;

  p->destroy = _cdr_destroy;

  return (p);
}

/**********************************************************************
  End of packet implementation
***********************************************************************/

/*L1, L2, Main unconstrained*/

static ilu_refany _courier_CreateDataBlock (void)
{
  struct Courier *new = (struct Courier *) ilu_malloc(sizeof(struct Courier));

  new->cr_in = NULL;
  return ((ilu_refany) new);
}

static void _courier_FreeDataBlock (struct Courier *d)
{
  if (d->cr_in != NULL)
    {
      packet_destroy (d->cr_in);
      d->cr_in = NULL;
    }
  ilu_free(d);
}

static void formRegistryName (ilu_string buf)
{
  ilu_string registryhome = (ilu_string ) getenv("ILUREGISTRY");

#ifdef MACOS
  if (registryhome != NULL)
    sprintf (buf, "%s:%s", registryhome, PROGRAM_NUMBER_REGISTRY_NAME);
  else
    sprintf (buf, "%s:%s", REGISTRY_LAST_RESORT,
	     PROGRAM_NUMBER_REGISTRY_NAME);
#else
  if (registryhome != NULL)
    sprintf (buf, "%s/%s", registryhome, PROGRAM_NUMBER_REGISTRY_NAME);
  else
    {
      sprintf (buf, "%s/%s", REGISTRY_LAST_RESORT,
	       PROGRAM_NUMBER_REGISTRY_NAME);
/*      fprintf (stderr, "ILU:  the environment variable ILUREGISTRY is not set, so we're using \"%s\" as the name of the Courier program number registry.\n", buf); */
    }
#endif
}

/*L1 >= {prmu}*/

static HashTable ProgramNumberHashTable = NULL;
/* program number -> courierinfo* */

static HashTable RegistryHashTable = NULL;
/* type unique_id -> courierinfo* */

static void _courier_EnsureRegistries(void)
{
  courierinfo *s = NULL;

  if (RegistryHashTable == NULL)
    {
      char buf[1000], line[1000], name[1000], id[1000];
      FILE *f;
      long lineno;

      RegistryHashTable = _ilu_hash_MakeNewTable (137, NULLFN, NULLFN);
      ProgramNumberHashTable = _ilu_hash_MakeNewTable (137,
		_ilu_hash_HashPointer, _ilu_hash_PointerCompare);

      formRegistryName (buf);
      if ((f = fopen(buf, "r")) == NULL)
	{
	  fprintf (stderr, "%s program number registry file %s.\n",
		   "ILU:  _courier_EnsureRegistries:  Can't access", buf);
	  return;
	}
      lineno = 0;
      while (fgets(line, sizeof(line), f) != NULL)
	{
	  lineno++;
     	  s = (courierinfo *) ilu_malloc(sizeof(courierinfo));
	  s->cri_class = NULL;
	  if (strncmp(line, ".LastAssignedProgramNumber ", 27) == 0)
	    ; /* skip this record */
	  else {
#ifdef MACOS
            /* AC: all the following code 'fixed' to do a hand-parse due
	       to broken sscanf() */
                /* NB: code now does a return if any line from registry
			is malformed */
		/* NB: We also don't use isspace() to advance over
		 * whitespace, because isspace() is also broken on the
		 * mac...*/
              if ((tmp=strchr(line,':'))==NULL) {
                  fprintf (stderr, "%s %s\"%s\":  %s",
			   "_courier_EnsureRegistries:  Invalid record",
			   "(1) in program number registry file",
			   buf, line);
                  return;
                }
              strncpy(name,line,tmp-line);
              name[tmp-line]='\0';
              tmp++;
              start=tmp;
              ch=*tmp;
              while ((ch!=' ') && (ch!='\0')) {
                  tmp++;
                  ch=*tmp;
                }
              if (ch=='\0') {
                  fprintf (stderr, "%s %s \"%s\":  %s",
			   "_courier_EnsureRegistries:  Invalid record",
			   "(2) in program number registry file",
			   buf, line);
                  return;
                }
              strncpy(id,start,tmp-start);
              id[tmp-start]='\0';
              tmp++;
              /* advance past white-space since strtoul is too stupid
		 to do it for us... */
              while (*tmp==' ')
                        tmp++;
              s->cri_pnumber=_ilu_strtoul(tmp,&tmp,16);
              if (tmp==NULL) {
                  fprintf (stderr, "%s %s \"%s\":  %s",
			   "_courier_EnsureRegistries:  Invalid record",
			   "(3) in program number registry file",
			   buf, line);
                  return;
                }
              while (*tmp==' ')
                  tmp++;
              s->cri_version=_ilu_strtoul(tmp,NULL,10);
#else /* MACOS */
	      if (4 != sscanf (line, "%1000[^:]:%1000s %lx %lu", name, id,
			       &s->cri_pnumber, &s->cri_version))
                  fprintf (stderr, "%s line %lu %s \"%s\":  %s",
			   "_courier_EnsureRegistries:  Invalid record at",
			   lineno, "in program number registry file",
			   buf, line);
	      else
#endif	/* MACOS */
	       {
	         if (_ilu_hash_FindInTable (RegistryHashTable, id)
	   	  OR _ilu_hash_FindInTable (ProgramNumberHashTable,
					    (void *) s->cri_pnumber))
	   	{
		  fprintf (stderr,
			   "%s %s %lu %s \"%s\":  <%s> (ignoring)\n",
			   "_courier_EnsureRegistries:  Duplicate type id",
			   "or program number at line", lineno,
			   "in program number registry", buf, line);
		  continue;
		}
	      s->cri_type_id = _ilu_Strdup(id);
	      _ilu_Assert(_ilu_hash_AddToTable(RegistryHashTable,
					       s->cri_type_id, s),
			  "Courier AddToTable RegistryHashTable");
	      _ilu_Assert(_ilu_hash_AddToTable(ProgramNumberHashTable,
				      (void *) s->cri_pnumber, s),
		     "Courier AddToTable ProgramNumberHashTable");
	     }
          }
	}
      fclose (f);
    }
}

static courierinfo *_courier_CourierInformationForClass (ilu_Class class)
{
  courierinfo *s = NULL;

  _courier_EnsureRegistries();

  s = (courierinfo *) _ilu_hash_FindInTable (RegistryHashTable,
					    class->cl_unique_id);
  if (s != NULL)
    s->cri_class = class;

  DEBUG(COURIER_DEBUG,
	(stderr, "%s \"%s:%s\", pnumber is 0x%lx, version is %ld.\n",
	 "_courier_CourierInformationForClass:  Class", class->cl_name,
	 class->cl_unique_id, s == NULL ? 0 : s->cri_pnumber,
	 s == NULL ? 0 : s->cri_version));

  return (s);
}

static courierinfo *_courier_ClassFromProgramNumber (ilu_cardinal pnumber,
						   ilu_cardinal version)
{
  courierinfo *s = NULL;

  _courier_EnsureRegistries();

  s = (courierinfo *) _ilu_hash_FindInTable (ProgramNumberHashTable,
					    (void *) pnumber);

  DEBUG(COURIER_DEBUG,
	(stderr,
	 "%s 0x%lx, version is %lu(%lu), class is \"%s\" (\"%s:%s\").\n",
	 "_courier_ClassFromProgramNumber:  pnumber is", pnumber, version,
	 (s == NULL) ? 0 : s->cri_version,
	 (s == NULL) ? "unknown" : s->cri_type_id,
	 (s == NULL || s->cri_class == NULL) ? "" : s->cri_class->cl_unique_id,
	 (s == NULL || s->cri_class == NULL) ? "" : s->cri_class->cl_name));

  return (s);
}

/*======================================================================*/
/*======================== Basic I/O code ==============================*/
/*======================================================================*/

#define INPUT_ERROR		1
#define OUTPUT_ERROR		2

/*L1, L2, Main unconstrained (this is only for calling from debugger)*/
static ilu_cardinal ilu_courier_SetMaxStringSize (ilu_cardinal size)
{
  ilu_cardinal old_size = CourierMaxStringSize;
  if (size > 0)
    CourierMaxStringSize = size;
  return (old_size);  
}

/*L2 >= {call's connection's iomu}*/
/*L1, Main unconstrained*/

/* ==================== short cardinal ==================== */

static ilu_boolean _courier_OutputShortCardinal (ilu_Call call, ilu_shortcardinal i)
{
  register unsigned char * buf;
  ilu_Error err = ILU_NO_ERR;

  buf = transport_get_output_buffer(connection_transport(call_connection(call)), 2, &err);
  if (err != NULL)
    {
      fprintf (stderr, "(ILU)_courier_OutputShortCardinal:  Error \"%s: %s\" signalled on attempt to fetch 2 bytes of buffer.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
      ilu_FreeError(err);
    }
  else
    {
      *buf++ = (i >> 8) & 0xFF;
      *buf++ = i & 0xFF;
      return (ilu_TRUE);
    }
  return (err == NULL);
}

static ilu_boolean _courier_InputShortCardinal (ilu_Call call, ilu_shortcardinal *i)
{
  return (packet_get_u_short (COURIER(call_connection(call))->cr_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfShortCardinal (ilu_Call call, ilu_shortcardinal i)
{
  return (2);
}

/* ==================== cardinal ==================== */

static ilu_boolean _courier_OutputCardinal (ilu_Call call, ilu_cardinal l)
{
  register unsigned char * buf;
  ilu_Error err = ILU_NO_ERR;

  buf = transport_get_output_buffer(connection_transport(call_connection(call)), 4, &err);
  if (err != NULL)
    {
      fprintf (stderr, "(ILU)_courier_OutputCardinal:  Error \"%s: %s\" signalled on attempt to fetch 4 bytes of buffer.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
      ilu_FreeError(err);
    }
  else
    {
      *buf++ = (l >> 24) & 0xFF;
      *buf++ = (l >> 16) & 0xFF;
      *buf++ = (l >> 8) & 0xFF;
      *buf++ = l & 0xFF;
      return (ilu_TRUE);
    }
  return (err == NULL);
}

static ilu_boolean _courier_InputCardinal (ilu_Call call, ilu_cardinal *i)
{
  return (packet_get_u_long (COURIER(call_connection(call))->cr_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfCardinal (ilu_Call call, ilu_cardinal i)
{
  return (4);
}

/* ==================== integer ==================== */

static ilu_boolean _courier_OutputInteger (ilu_Call call, ilu_integer i)
{
  return _courier_OutputCardinal(call, (ilu_cardinal) i);
}

static ilu_boolean _courier_InputInteger (ilu_Call call, ilu_integer *i)
{
  return (packet_get_long (COURIER(call_connection(call))->cr_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfInteger (ilu_Call call, ilu_integer i)
{
  return (4);
}

/* ==================== short integer ==================== */

static ilu_boolean _courier_OutputShortInteger (ilu_Call call, ilu_shortinteger i)
{
  return (_courier_OutputCardinal(call, (ilu_cardinal) i));
}

static ilu_boolean _courier_InputShortInteger (ilu_Call call, ilu_shortinteger *i)
{
  return (packet_get_short (COURIER(call_connection(call))->cr_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfShortInteger (ilu_Call call, ilu_shortinteger i)
{
  return (2);
}

/* ==================== long integer ==================== */

static ilu_boolean _courier_OutputLongInteger (ilu_Call call, ilu_longinteger i)
{
  return (_courier_OutputCardinal(call, (ilu_cardinal) i.high) &&
	  _courier_OutputCardinal(call, (ilu_cardinal) i.low));
}

static ilu_boolean _courier_InputLongInteger (ilu_Call call, ilu_longinteger *i)
{
  ilu_boolean stat = ilu_TRUE;
  PACKET pkt = COURIER(call_connection(call))->cr_in;
  stat = (packet_get_long (pkt, &i->high));
  return (stat && (packet_get_u_long (pkt, &i->low)));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfLongInteger (ilu_Call call, ilu_longinteger i)
{
  return (8);
}

/* ==================== long cardinal ==================== */

static ilu_boolean _courier_OutputLongCardinal (ilu_Call call, ilu_longcardinal i)
{
  return (_courier_OutputCardinal(call, (ilu_cardinal) i.high) &&
	  _courier_OutputCardinal(call, (ilu_cardinal) i.low));
}

static ilu_boolean _courier_InputLongCardinal (ilu_Call call, ilu_longcardinal *i)
{
  ilu_boolean stat;
  PACKET pkt = COURIER(call_connection(call))->cr_in;

  stat = (packet_get_u_long (pkt, &i->high));
  return (stat && (packet_get_u_long (pkt, &i->low)));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfLongCardinal (ilu_Call call, ilu_longcardinal i)
{
  return (8);
}

/* ==================== enumeration ==================== */

static ilu_boolean _courier_OutputEnumeration (ilu_Call call, ilu_shortcardinal i)
{
  return (_courier_OutputShortCardinal(call, i));
}

static ilu_boolean _courier_InputEnumeration (ilu_Call call, ilu_shortcardinal *i)
{
  return(packet_get_u_short (COURIER(call_connection(call))->cr_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfEnumeration (ilu_Call call, ilu_shortcardinal i)
{
  return (2);
}

/* ==================== real ==================== */

static ilu_boolean _courier_OutputReal (ilu_Call call, double d)
{
  double l2 = d;

  return (_courier_OutputCardinal(call, ((ilu_cardinal *) &l2)[BIG_ENDIAN ? 0 : 1]) &&
	  _courier_OutputCardinal(call, ((ilu_cardinal *) &l2)[BIG_ENDIAN ? 1 : 0]));
}

static ilu_boolean _courier_InputReal (ilu_Call call, double *d)
{
  return (packet_get_double (COURIER(call_connection(call))->cr_in, d));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfReal (ilu_Call call, double d)
{
  return (8);
}

/* ==================== long real ==================== */

static ilu_boolean _courier_OutputLongReal (ilu_Call call, ilu_longreal d)
{
  register unsigned char * buf;
  ilu_Error err = NULL;

  buf = transport_get_output_buffer(connection_transport(call_connection(call)), 16, &err);
  if (err != NULL)
    {
      fprintf (stderr, "(ILU)_courier_OutputLongReal:  Error \"%s: %s\" signalled on attempt to fetch 16 bytes of buffer.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
      ilu_FreeError(err);
      return (ilu_FALSE);
    }
  else
    {
      register int index;
      for (index = 0;  index < 16;  index++)
	*buf++ = d.b[index];
      return (ilu_TRUE);
    }
}

static ilu_boolean _courier_InputLongReal (ilu_Call call, ilu_longreal *d)
{
  unsigned char *ref = d->b;
  return(packet_get_opaque (COURIER(call_connection(call))->cr_in, &ref, 16));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfLongReal (ilu_Call call, ilu_longreal d)
{
  return (16);
}

/* ==================== short real ==================== */

static ilu_boolean _courier_OutputShortReal (ilu_Call call, float f)
{
  float f2;

  f2 = f;
  return (_courier_OutputCardinal(call, *((ilu_cardinal *) &f2)));
}

static ilu_boolean _courier_InputShortReal (ilu_Call call, float *f)
{
  return (packet_get_float (COURIER(call_connection(call))->cr_in, f));
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfShortReal (ilu_Call call, float d)
{
  return (4);
}

/* ==================== bytes ==================== */

static ilu_boolean _courier_OutputBytes (ilu_Call call, ilu_bytes s, ilu_cardinal len, ilu_cardinal limit)
{
  ilu_Error err = ILU_NO_ERR;
  ilu_cardinal size;

  if (limit > 0)
    size = MIN(len, limit);
  else
    size = len;

  if (limit > 0 && limit <= 0xFFFF)
    {
      if (NOT _courier_OutputShortCardinal (call, size))
	return (ilu_FALSE);
    }
  else
    {
      if (NOT _courier_OutputCardinal (call, size))
	return (ilu_FALSE);
    }
  err = transport_write_bytes(call_transport(call), s, PADDED_SIZE(size));
  if (err != NULL)
    {
      fprintf (stderr, "(ILU)_courier_OutputBytes:  Error \"%s: %s\" on output of string %p, len %lu, limit %lu.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err), s, len, limit);
      ilu_FreeError(err);
    }
  return (err == NULL);
}

static ilu_boolean _courier_InputBytes (ilu_Call call, ilu_bytes *s, ilu_cardinal *len, ilu_cardinal limit)
{
  *s = NULL;
  return (packet_get_bytes (COURIER(call_connection(call))->cr_in, s, len, limit));
}

  /*ARGSUSED*/
static ilu_cardinal _courier_SizeOfBytes (ilu_Call call, ilu_bytes i, ilu_cardinal l, ilu_cardinal limit)
{
  if (((limit > 0) && (l > limit)) || l > CourierMaxStringSize)
    {
      fprintf (stderr, "_courier_SizeOfBytes:  Attempt to pass bytestring of length %lu, which exceeds either CourierMaxStringSize of %lu, or limit on this sequence type of %lu.\n",
	       l, CourierMaxStringSize, limit);
      return (0);
    }
  return (((limit != 0 && limit <= 0xFFFF) ? 2 : 4) + PADDED_SIZE(l));
}

/* ==================== string ==================== */

static ilu_boolean _courier_OutputString (ilu_Call call, ilu_string s, ilu_cardinal len, ilu_cardinal limit)
{
  return (_courier_OutputBytes(call, (ilu_bytes) s, len, limit));
}

static ilu_boolean _courier_InputString (ilu_Call call, ilu_string *s, ilu_cardinal *len, ilu_cardinal limit)
{
  *s = NULL;
  if (!packet_get_bytes (COURIER(call_connection(call))->cr_in,
	(ilu_byte **) s, len, limit))
    return (ilu_FALSE);
  return (ilu_TRUE);
}

  /*ARGSUSED*/
static ilu_cardinal _courier_SizeOfString (ilu_Call call, ilu_string i, ilu_cardinal l, ilu_cardinal limit)
{
  if (l > CourierMaxStringSize)
    {
      fprintf (stderr, "_courier_SizeOfString:  Attempt to pass string of length %lu, which exceeds either CourierMaxStringSize value of %lu, or limit on this sequence type of %lu.\n",
	       l, CourierMaxStringSize, limit);
      return (0);
    }

  return (4 + PADDED_SIZE(l));
}

/* ==================== byte ==================== */

static ilu_boolean _courier_OutputByte (ilu_Call call, ilu_byte b)
{
  return (_courier_OutputShortCardinal(call, (ilu_shortcardinal) b));
}

static ilu_boolean _courier_InputByte (ilu_Call call, ilu_byte *b)
{
  ilu_shortcardinal l;
  ilu_boolean stat;

  stat = packet_get_u_short (COURIER(call_connection(call))->cr_in, &l);
  if (stat)
    *b = l & 0xFF;
  return (stat);
}

/*ARGSUSED*/
static ilu_cardinal _courier_SizeOfByte (ilu_Call call, ilu_byte i)
{
  return (2);
}

/* ==================== opaque ==================== */

static ilu_boolean _courier_OutputOpaque (ilu_Call call, ilu_bytes o, ilu_cardinal len)
{
  ilu_Error err;

  err = transport_write_bytes(call_transport(call), o, PADDED_SIZE(len));
  if (err != NULL)
    {
      fprintf (stderr, "(ILU)_courier_OutputOpaque:  Error \"%s: %s\" on output of string %p, len %lu.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err), o, len);
      ilu_FreeError(err);
    }
  return (err == NULL);
}

static ilu_boolean _courier_InputOpaque (ilu_Call call, ilu_bytes *o, ilu_cardinal len)
{
  return(packet_get_opaque (COURIER(call_connection(call))->cr_in, o, len));
}

static ilu_cardinal _courier_SizeOfOpaque (ilu_Call call, ilu_bytes o, ilu_cardinal len)
{
  return (PADDED_SIZE(len));
}

/* ==================== sequence ==================== */

static ilu_boolean _courier_OutputSequence (ilu_Call c, ilu_cardinal sequenceLength, ilu_cardinal limit)
{
  if (limit != 0 && limit <= 0xFFFF)
    return (_courier_OutputShortCardinal (c, (ilu_shortcardinal) sequenceLength));
  else
    return (_courier_OutputCardinal (c, sequenceLength));
}

static ilu_boolean _courier_OutputSequenceMark (ilu_Call c, ilu_cardinal extent)
{
  return (ilu_TRUE);
}

static ilu_boolean _courier_InputSequenceMark (ilu_Call c, ilu_cardinal extent)
{
  return (ilu_TRUE);
}

static ilu_boolean _courier_InputSequence (ilu_Call c, ilu_cardinal *sequenceLength, ilu_cardinal limit)
{
  ilu_boolean stat;
  ilu_shortcardinal len;

  if (limit != 0 && limit <= 0xFFFF)
    {
      stat = _courier_InputShortCardinal (c, &len);
      if (stat)
	*sequenceLength = len;
    }
  else
    stat = _courier_InputCardinal (c, sequenceLength);
  return(stat);
}

static ilu_boolean _courier_EndSequence (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _courier_SizeOfSequence (ilu_Call c, ilu_cardinal length, ilu_cardinal limit)
{
  if (limit != 0 && limit <= 0xFFFF)
    return (_courier_SizeOfShortCardinal(c, (ilu_shortcardinal) length));
  else
    return (_courier_SizeOfCardinal(c, length));
}

/* ==================== union ==================== */

static ilu_boolean _courier_OutputUnion (ilu_Call c, ilu_shortcardinal typeIndex)
{
  return (_courier_OutputShortCardinal (c, typeIndex));
}

static ilu_boolean _courier_InputUnion (ilu_Call c, ilu_shortcardinal *typeIndex)
{
  return(_courier_InputShortCardinal (c, typeIndex));
}

static ilu_boolean _courier_EndUnion (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _courier_SizeOfUnion (ilu_Call c, ilu_shortcardinal typeIndex)
{
  return (_courier_SizeOfShortCardinal(c, typeIndex));
}

/* ==================== array ==================== */

static ilu_boolean _courier_OutputArray (ilu_Call c, ilu_cardinal len)
{
  return(ilu_TRUE);
}

static ilu_boolean _courier_InputArray (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _courier_EndArray (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _courier_SizeOfArray (ilu_Call c, ilu_cardinal len)
{
  return(0);
}

/* ==================== record ==================== */

static ilu_boolean _courier_OutputRecord (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _courier_InputRecord (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _courier_EndRecord (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _courier_SizeOfRecord (ilu_Call c)
{
  return(0);
}

/*======================================================================
**======================================================================
**====================  Non-I/O code ===================================
**======================================================================
**====================================================================*/

/*L2, Main unconstrained*/
/*L1_sup < prmu*/
static ilu_string _courier_FormProtocolHandle (ilu_Object obj)
{
  char buf[100];
  courierinfo *s = NULL;
  ilu_Class class = object_class(obj);

  _ilu_AcquireMutex(ilu_prmu);
  s = _courier_CourierInformationForClass (class);
  if (s == NULL)
    {
      _ilu_ReleaseMutex(ilu_prmu);
      fprintf (stderr, "%s program#/version for class %s.\n",
	       "_courier_FormProtocolHandle:  Can't figure", class->cl_name);
      return (NULL);
    }
  else
    {
      sprintf (buf, "courier_0x%lx_%ld", s->cri_pnumber, s->cri_version);
      _ilu_ReleaseMutex(ilu_prmu);
      return (_ilu_Strdup(buf));
    }
}
     
/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static ilu_bytes _courier_ReadPacket (ilu_Transport bs,
			      ilu_cardinal *packetLength,
			      ilu_PacketType *type, ilu_cardinal *sn)
{
  ilu_bytes packet;
  ilu_cardinal serialNumber, packetType, size;

  if (!transport_read_message(bs, &packet, &size)) {
    DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	  (stderr, "_courier_ReadPacket:  error on transport_read_message\n"));
    return (NULL);
  };

  packetType = (packet[0] << 8) | packet[1];
  serialNumber = (packet[2] << 8) | packet[3];

  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_ReadPacket:  read packet of length %lu, SN %lu, type %lu (%s).\n",
	 size, serialNumber, packetType,
	 ((packetType == 0) ? "request"
	  : ((packetType == 2) ? "reply"
	     : ((packetType == 3 || packetType == 1) ? "exception"
		: "unknown type")))));

  *packetLength = size;
  *type = (packetType == 0) ? ilu_PacketType_Request : ilu_PacketType_Reply;
  *sn = serialNumber;
  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_ReadPacket:  returning packet SN %lu\n", serialNumber));
  return (packet);  
}

static ilu_boolean _courier_InterpretRequest (ilu_Call call, ilu_bytes packet,
				      ilu_cardinal len)
{
  struct Courier *c = COURIER(call_connection(call));
  ilu_cardinal programNumber;
  ilu_shortcardinal programVersion, methodID, junk;
  ilu_Class class;
  courierinfo *s;

  c->cr_in = _cdr_CreatePacket (packet, len);

  _courier_InputShortCardinal (call, &junk);	/* packet type */
  _courier_InputShortCardinal (call, &junk);	/* serial no. of call */
  _courier_InputCardinal (call, &programNumber);
  _courier_InputShortCardinal (call, &programVersion);
  _courier_InputShortCardinal (call, &methodID);
  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_InterpretRequest:  (call SN %lu) class ID = 0x%lx (version %u), method ID = %u\n",
	 call->ca_SN, programNumber, programVersion, methodID));

  _ilu_AcquireMutex(ilu_prmu);
  s = _courier_ClassFromProgramNumber(programNumber, programVersion);
  class = (s == NULL) ? NULL
	: (s->cri_class == NULL) ? ilu_FindClassFromID(s->cri_type_id)
	: s->cri_class;
  call_intro_type(call) = class;
  _ilu_ReleaseMutex(ilu_prmu);
  if (NULL == s OR NULL == class)
    {
      DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	    (stderr, "_courier_InterpretRequest:  (call %lu) Can't find ilu_Class with pn 0x%lx, version %u\n",
	     call->ca_SN, programNumber, programVersion));
      return (FALSE);
    }

  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_InterpretRequest:  (call SN %lu) intro_type is %s\n",
	 call->ca_SN, class->cl_name));

  call->ca_method = ilu_FindMethodByID (class, methodID);
  
  if (call->ca_method == NULL)
    {
      DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	    (stderr, "_courier_InterpretRequest:  (call %lu) Can't find method on class 0x%lx with methodID = %u.\n",
	     call->ca_SN, programNumber, methodID));
      return (FALSE);
    }
  else
    {
      DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	    (stderr, "_courier_InterpretRequest:  record for method %d is %p (%s).\n",
	     methodID, call->ca_method, call->ca_method->me_name));
    }
  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_InterpretRequest:  returning TRUE\n"));
  return (TRUE);
}

/*L1, Main unconstrained*/
/*L2 >= {conn's iomu}*/

static ilu_ProtocolException _courier_InterpretReply (ilu_Call call, ilu_bytes packet, ilu_cardinal len, ilu_cardinal *estatus)
{
  struct Courier *c = COURIER(call->ca_connection);
  ilu_shortcardinal replyStatus, junk;
  static ilu_ProtocolException replyStatusExceptions[] = {
    ilu_ProtocolException_NoSuchClassAtServer,
    ilu_ProtocolException_ClassVersionMismatch,
    ilu_ProtocolException_NoSuchMethodOnClass,
    ilu_ProtocolException_GarbageArguments
    };

  c->cr_in = _cdr_CreatePacket (packet, len);

  _courier_InputShortCardinal (call, &replyStatus);
  _courier_InputShortCardinal (call, &junk);		/* actually, serial number */

  DEBUG((INCOMING_DEBUG | COURIER_DEBUG),
	(stderr, "_courier_InterpretReply:  SN %lu, %lu bytes, replyStatus is %u\n",
	 call->ca_SN, len, replyStatus));

  if (replyStatus == 2)		/* MSG_ACCEPTED */
    {
      *estatus = 0;
      return (ilu_ProtocolException_Success);
    }
  else if (replyStatus == 3)
    {
      ilu_shortcardinal e;

      _courier_InputShortCardinal (call, &e);
      *estatus = e;
      return (ilu_ProtocolException_Success);
    }
  else if (replyStatus == 1)
    {
      _courier_InputShortCardinal (call, &replyStatus);
      switch (replyStatus)
	{
	case 1:
	  _courier_InputShortCardinal (call, &junk);	/* lowest version number */
	  _courier_InputShortCardinal (call, &junk);	/* highest version number */

	case 0:
	case 2:
	case 3:
	  return (replyStatusExceptions[replyStatus]);

	default:
	  return (ilu_ProtocolException_Unknown);
	}
    }
  return (ilu_ProtocolException_Unknown);
}

/*L1_sup < prmu*/

static ilu_boolean GetCourierProgramNumberAndVersion (ilu_Class pclass, ilu_cardinal *pnumber, ilu_cardinal *version)
{
  courierinfo *s = NULL;

  _ilu_AcquireMutex(ilu_prmu);
  if ((s = _courier_CourierInformationForClass (pclass)) != NULL)	/* probably consult external DB */
    {
      *pnumber = s->cri_pnumber;
      *version = s->cri_version;
    }
  _ilu_ReleaseMutex(ilu_prmu);
  return (s != NULL);
}

static ilu_boolean _courier_BeginRequest (ilu_Call call, ilu_Class pclass, ilu_Method method, ilu_cardinal argSize)
{
  ilu_cardinal packetSize;
  ilu_cardinal pnumber, version;

  call->ca_method = method;

  DEBUG(CALL_DEBUG, (stderr, "_courier_BeginRequest:  call %p (sn %lu), argSize %lu, class %s (%s), method %s (%ld)\n",
		     call, call_serial_number(call), argSize,
		     class_name(pclass), class_unique_id(pclass), method_name(method), method_id(method)));

  if (!GetCourierProgramNumberAndVersion (pclass, &pnumber, &version))
    {
      fprintf (stderr, "_courier_BeginRequest:  Can't determine program-number/version of class \"%s\" on call %p.\n", pclass->cl_name, call);
      return (FALSE);
    }

  DEBUG(CALL_DEBUG, (stderr, "_courier_BeginRequest:  pnum/vers 0x%lx/%ld\n", pnumber, version));

  packetSize = argSize
    + 12;	/* for fields of header */

  _courier_OutputShortCardinal (call, 0);	/* CallMessageBody */
  _courier_OutputShortCardinal (call, call->ca_SN);	/* transaction ID */
  _courier_OutputCardinal (call, pnumber);
  _courier_OutputShortCardinal (call, version);
  _courier_OutputShortCardinal (call, method->me_id);
  
  DEBUG(CALL_DEBUG, (stderr, "_courier_BeginRequest:  request %lu begun (size %lu).\n",
		       call->ca_SN, packetSize));
  return (TRUE);
}

/*L1, Main unconstrained*/
/*L2 >= {conn's iomu}*/

static void _courier_DropInput (ilu_Call call)
{
  ilu_Connection  conn = call_connection(call);
  struct Courier *s = COURIER(conn);
  if (connection_closed(conn)) {
    _ilu_Assert(s == NULL,
		"protocol_data not NULL for closed connection");
    return;
  }
  if (s->cr_in != NULL)
    packet_destroy(s->cr_in);
  s->cr_in = NULL;
}

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static void _courier_FlushBuffer (ilu_Call call)
{
  transport_flush_output(call_transport(call));
}

static void _courier_DropRequest (ilu_Call call)
{
  transport_flush_output(connection_transport(call_connection(call)));
}

static          ilu_boolean
_courier_SendPacket(ilu_Call call, ilu_boolean isReply)
{
  ilu_Error       err;

  ilu_Transport   bs = call_transport(call);
  err = transport_flush_output(bs);
  if (err != NULL) {
    fprintf(stderr,
	    "(ILU)_courier_SendPacket:  Error \"%s: %s\" in write of packet.\n",
	    ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
    ILU_ERR_SWITCH(err) {
      ILU_ERR_CASE(WriteError, ev) {
	fprintf(stderr,
		"(ILU)_courier_SendPacket:  write error on packet, errno %d, \"%s\".\n",
		ev->ep_errno, ev->errstr);
	ilu_free(ev->buffer);	/* pretty useless by now */
      }
      ILU_ERR_ELSE
	fprintf(stderr,
		"(ILU)_courier_SendPacket:  Error \"%s: %s\" in write of packet.\n",
		ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
    } ILU_ERR_ENDSWITCH;
    ilu_FreeError(err);
    return (ilu_FALSE);
  }
  if (bs->tr_class->tc_timesout) {
    if (isReply) {
      err = _ilu_CacheCall(call, bs->tr_packet, bs->tr_size);
      ILU_ERR_SWITCH(err) {
	ILU_SUCCESS_CASE;
	ILU_ERR_CASE(TInfoTooLong, x)
	  fprintf(stderr, "(ILU)_courier_SendPacket:"
		  "  cant cache reply because transport info"
		  " (%s) too long", x->tinfo);
      } ILU_ERR_ENDSWITCH;
      ilu_FreeError(err);
    } else
      ilu_free(bs->tr_packet);
    bs->tr_packet = NULL;
  }
  return (ilu_TRUE);
}

static ilu_boolean _courier_FinishRequest (ilu_Call call)
{
  return (_courier_SendPacket(call, ilu_FALSE));
}

static ilu_boolean _courier_FinishReply (ilu_Call call)
{
  return (_courier_SendPacket (call, ilu_TRUE));
}

static ilu_boolean _courier_FinishException (ilu_Call call)
{
  return (_courier_SendPacket(call, ilu_TRUE));
}

/*L1, Main unconstrained*/
/*L2 >= {call's conn's iomu}*/

/*L1, Main unconstrained*/
/*L2 >= {call's conn's iomu}*/

static ilu_boolean _courier_BeginReply (ilu_Call call, ilu_boolean exceptions, ilu_cardinal argSize)
{
  ilu_cardinal packetSize;

  DEBUG(PACKET_DEBUG, (stderr, "_courier_BeginReply:  SN %lu, argSize %lu, exceptions %s, fd %d.\n",
		       call->ca_SN, argSize, exceptions ? "TRUE" : "FALSE",
		       transport_file_descriptor(connection_transport(call_connection(call)))));

  packetSize = argSize
    + 4;			/* for the basic header fields */

  _courier_OutputShortCardinal (call, 2); /* message type == REPLY */
  _courier_OutputShortCardinal (call, call->ca_SN);

  DEBUG(PACKET_DEBUG, (stderr, "_courier_BeginReply:  started reply %lu (size %lu).\n", call->ca_SN, packetSize));
  return (TRUE);
}

static ilu_boolean _courier_BeginException (ilu_Call call, ilu_cardinal evalue, ilu_cardinal argSize)
                   
                         	/* the exception value */
                          
{
  ilu_cardinal packetSize;

  if (evalue == 0)		/* send protocol error */
    packetSize = 6;		/* no values to send */
  else
    packetSize = argSize + 6;	/* for the fields of the header */

  _courier_OutputShortCardinal (call, (evalue == 0) ? 1 : 3);		/* message type == REJECT or ABORT */
  _courier_OutputShortCardinal (call, call->ca_SN);
  _courier_OutputShortCardinal (call, (evalue == 0) ? argSize : evalue);	/* exception value */

  DEBUG(PACKET_DEBUG, (stderr, "_courier_BeginException:  exception started to peer %d:  SN %lu, size %lu, evalue %lu.\n",
		       transport_file_descriptor(connection_transport(call_connection(call))),
		       call_serial_number(call), packetSize, (evalue == 0) ? argSize : evalue));
  return (TRUE);
}

/*L2, Main unconstrained*/
/*L1 >= {prmu}*/

static ilu_Protocol _courier_NewCourier (void)
{
  ilu_Protocol new = (ilu_Protocol)
		     ilu_malloc(sizeof(struct _ilu_Protocol_s));

  new->pr_concurrent_requests = ilu_FALSE;

  new->pr_begin_request = _courier_BeginRequest;
  new->pr_finish_request = _courier_FinishRequest;
  new->pr_resend_request = _courier_FinishRequest;
  new->pr_drop_request = _courier_DropRequest;
  new->pr_begin_reply = _courier_BeginReply;
  new->pr_finish_reply = _courier_FinishReply;
  new->pr_begin_exception = _courier_BeginException;
  new->pr_finish_exception = _courier_FinishException;

  new->pr_output_integer = _courier_OutputInteger;
  new->pr_input_integer = _courier_InputInteger;
  new->pr_size_of_integer = _courier_SizeOfInteger;

  new->pr_output_shortinteger = _courier_OutputShortInteger;
  new->pr_input_shortinteger = _courier_InputShortInteger;
  new->pr_size_of_shortinteger = _courier_SizeOfShortInteger;

  new->pr_output_longinteger = _courier_OutputLongInteger;
  new->pr_input_longinteger = _courier_InputLongInteger;
  new->pr_size_of_longinteger = _courier_SizeOfLongInteger;

  new->pr_output_cardinal = _courier_OutputCardinal;
  new->pr_input_cardinal = _courier_InputCardinal;
  new->pr_size_of_cardinal = _courier_SizeOfCardinal;

  new->pr_output_shortcardinal = _courier_OutputShortCardinal;
  new->pr_input_shortcardinal = _courier_InputShortCardinal;
  new->pr_size_of_shortcardinal = _courier_SizeOfShortCardinal;

  new->pr_output_longcardinal = _courier_OutputLongCardinal;
  new->pr_input_longcardinal = _courier_InputLongCardinal;
  new->pr_size_of_longcardinal = _courier_SizeOfLongCardinal;

  new->pr_output_real = _courier_OutputReal;
  new->pr_input_real = _courier_InputReal;
  new->pr_size_of_real = _courier_SizeOfReal;

  new->pr_output_shortreal = _courier_OutputShortReal;
  new->pr_input_shortreal = _courier_InputShortReal;
  new->pr_size_of_shortreal = _courier_SizeOfShortReal;

  new->pr_output_longreal = _courier_OutputLongReal;
  new->pr_input_longreal = _courier_InputLongReal;
  new->pr_size_of_longreal = _courier_SizeOfLongReal;

  new->pr_output_optional =
	(ilu_boolean (*)(ilu_Call,ilu_boolean)) _courier_OutputCardinal;
  new->pr_input_optional =
	(ilu_boolean (*)(ilu_Call,ilu_boolean *)) _courier_InputCardinal;
  new->pr_size_of_optional =
	(ilu_cardinal (*)(ilu_Call,ilu_boolean)) _courier_SizeOfCardinal;

  new->pr_output_enum_code = _courier_OutputEnumeration;
  new->pr_input_enum_code = _courier_InputEnumeration;
  new->pr_size_of_enum_code = _courier_SizeOfEnumeration;

  new->pr_output_byte = _courier_OutputByte;
  new->pr_input_byte = _courier_InputByte;
  new->pr_size_of_byte = _courier_SizeOfByte;

  new->pr_output_character = _courier_OutputShortCardinal;
  new->pr_input_character = _courier_InputShortCardinal;
  new->pr_size_of_character = _courier_SizeOfShortCardinal;

  new->pr_output_boolean =
	(ilu_boolean (*)(ilu_Call,ilu_boolean)) _courier_OutputCardinal;
  new->pr_input_boolean =
	(ilu_boolean (*)(ilu_Call,ilu_boolean *)) _courier_InputCardinal;
  new->pr_size_of_boolean =
	(ilu_cardinal (*)(ilu_Call,ilu_boolean)) _courier_SizeOfCardinal;

  new->pr_output_shortchar = _courier_OutputByte;
  new->pr_input_shortchar = _courier_InputByte;
  new->pr_size_of_shortchar = _courier_SizeOfByte;

  new->pr_output_string = _courier_OutputString;
  new->pr_input_string = _courier_InputString;
  new->pr_size_of_string = _courier_SizeOfString;

  new->pr_output_bytes = _courier_OutputBytes;
  new->pr_input_bytes = _courier_InputBytes;
  new->pr_size_of_bytes = _courier_SizeOfBytes;

  new->pr_output_opaque = _courier_OutputOpaque;
  new->pr_input_opaque = _courier_InputOpaque;
  new->pr_size_of_opaque = _courier_SizeOfOpaque;

  new->pr_output_object_id = _ilu_OutputObjectID;
  new->pr_input_object_id = _ilu_InputObjectID;
  new->pr_size_of_object_id = _ilu_SizeOfObjectID;

  new->pr_output_stringvec =
	(ilu_boolean (*)(ilu_Call,ilu_string,ilu_cardinal))
	_courier_OutputOpaque;
  new->pr_input_stringvec =
	(ilu_boolean (*)(ilu_Call,ilu_string *,ilu_cardinal))
	_courier_InputOpaque;
  new->pr_size_of_stringvec =
	(ilu_cardinal (*)(ilu_Call,ilu_string,ilu_cardinal))
	_courier_SizeOfOpaque;

  new->pr_output_pipe = NULLFN;
  new->pr_input_pipe = NULLFN;
  new->pr_size_of_pipe = NULLFN;

  new->pr_output_sequence = _courier_OutputSequence;
  new->pr_output_sequence_mark = _courier_OutputSequenceMark;
  new->pr_input_sequence = _courier_InputSequence;
  new->pr_input_sequence_mark = _courier_InputSequenceMark;
  new->pr_end_sequence = _courier_EndSequence;
  new->pr_size_of_sequence = _courier_SizeOfSequence;

  new->pr_output_record = _courier_OutputRecord;
  new->pr_input_record = _courier_InputRecord;
  new->pr_end_record = _courier_EndRecord;
  new->pr_size_of_record = _courier_SizeOfRecord;

  new->pr_output_array = _courier_OutputArray;
  new->pr_input_array = _courier_InputArray;
  new->pr_end_array = _courier_EndArray;
  new->pr_size_of_array = _courier_SizeOfArray;

  new->pr_output_union = _courier_OutputUnion;
  new->pr_input_union = _courier_InputUnion;
  new->pr_end_union = _courier_EndUnion;
  new->pr_size_of_union = _courier_SizeOfUnion;

  new->pr_form_handle = _courier_FormProtocolHandle;

  new->pr_read_packet = _courier_ReadPacket;
  new->pr_interpret_pipe_request = NULLFN;
  new->pr_interpret_request = _courier_InterpretRequest;
  new->pr_request_read = _courier_DropInput;
  new->pr_interpret_reply = _courier_InterpretReply;
  new->pr_finish_call = _courier_DropInput;

  new->pr_create_data_block = _courier_CreateDataBlock;
  new->pr_free_data_block = (void (*)(void *)) _courier_FreeDataBlock;
  new->pr_flush = _courier_FlushBuffer;

  return (new);
}

/*L1_sup < prmu*/

ilu_Protocol _ilu_courier_Protocol(void)
{
  /*L1 >= {prmu}*/
  static ilu_Protocol StandardCourier = NULL;
  _ilu_AcquireMutex(ilu_prmu);
  if (StandardCourier == NULL)
    StandardCourier = _courier_NewCourier();
  _ilu_ReleaseMutex(ilu_prmu);
  return (StandardCourier);
}
