/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: error.c,v 1.16 1994/12/07 03:21:25 janssen Exp $ */
/* $Log: error.c,v $
 * Revision 1.16  1994/12/07  03:21:25  janssen
 * Remove "unused statement" complained about by gcc.
 *
 * Revision 1.15  1994/12/06  07:34:17  janssen
 * Added BadSBH and BadURL.
 *
 * Revision 1.14  1994/12/05  21:43:13  spreitze
 * Replaced "NULL" with "NIL";
 * "Last tweaked by" with "Last edited by".
 *
 * Revision 1.13  1994/11/17  08:29:08  janssen
 * Updated for new OS partitioning
 *
 * Revision 1.12  1994/11/04  05:16:56  spreitze
 * Eliminated the one method of the GC callback class.
 * Other minor improvements, mainly in error reporting.
 *
 * Revision 1.11  1994/10/13  20:01:36  spreitze
 * "free" -> "ilu_free";
 * ilu_ParseOID==0 doesn't imply non-null parts;
 * moved co_protocol_data into the variable part of a connection,
 * to be accessed only under the connection's iomu ---
 * so we can NULL it out when we free the protocol data block.
 *
 * Revision 1.10  1994/10/06  14:37:16  spreitze
 * Added #include <string.h> for memset.
 *
 * Revision 1.9  1994/09/20  02:13:57  janssen
 * Added new errors for use by marshalling code.
 *
 * Revision 1.8  1994/09/05  21:19:19  spreitze
 * Changed malloc calls to ilu_malloc calls.
 * Liberalized SetLockTech.
 * Generalize AddRegisterersToDefault
 * Renamed RegisterLockTech to SetLockTech.
 *
 * Revision 1.7  1994/08/01  17:18:03  spreitze
 * Added some errors internal to the kernel.
 * Small changes and additions to error system.
 *
 * Revision 1.6  1994/07/25  15:38:09  spreitze
 * Error system revision; header file rationalization.
 *
 * Revision 1.5  1994/07/06  07:10:14  janssen
 * Changed to allow dynamic registration of protocols and transports.
 * Removed ilu_TransportType and ilu_ProtocolType.
 *
 * Revision 1.4  1994/06/23  06:59:28  janssen
 * Output buffers now responsibility of transport, rather than protocol.
 * Modifications to TransportClass to allow this.
 * Note that tr_flush_buffer now serves as end-of-message marker.
 * */

/* Standard error machinery, and defs of standard errors */

#define _POSIX_SOURCE

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluerror.h"
#include "iluntrnl.h"

void ilu_FreeError(ilu_Error e)
{
  ilu_ErrorType   t;
  if (e == NIL)
    return;
  t = e->type;
  (*t->freeproc) (e);
  ilu_free(e);
  return;
}

void * _ilu_ErrCons (ilu_ErrorType et, unsigned int size)
{
  ilu_Error f;
  int newsize = sizeof(struct _ilu_ErrorBase_s) + size;

  f = (ilu_Error) ilu_must_malloc(newsize);
  memset ((char *) f, 0, newsize);
  f->type = et;
  f->data = ((char *) f) + sizeof(struct _ilu_ErrorBase_s);
  return f;
}

ILU_DEF_ERR(KernelBroken, "Internal bug in kernel") {
  /* nothing to free */
}

ILU_QUADEF(KernelBroken);

ILU_DEF_ERR(BadDataStructure, "Pointer to invalid data structure encountered") {
  /* assume that all strings are constants */
}

ILU_QUADEF(BadDataStructure);

ILU_DEF_ERR(MaxCountExceeded, "An attempt was made to fill an array or structure beyond its fixed limit") {
  /* assume that all strings are constants */
}

ILU_QUADEF(MaxCountExceeded);

ILU_DEF_ERR(ProtocolAlreadyRegistered, "Specified protocol was already in registry of protocols") {
  /* assume that all strings are constants */
}

ILU_QUADEF(ProtocolAlreadyRegistered);

ILU_DEF_ERR(TransportAlreadyRegistered, "Specified transport was already in registry of transports") {
  /* assume that all strings are constants */
}

ILU_QUADEF(TransportAlreadyRegistered);

ILU_DEF_ERR(WrongLocks, "Locking precondition not met") {
  /* Nothing to free */
}

ILU_QUADEF(WrongLocks);

ILU_DEF_ERR(BadContactInfo, "Bogus contact info") {
  ilu_free(e->x);
}

ILU_QUADEF(BadContactInfo);

ILU_DEF_ERR(BadTransportInfo, "Bogus transport info") {
  ilu_free(e->x);
}

ILU_QUADEF(BadTransportInfo);

ILU_DEF_ERR(BadProtocolInfo, "Bogus protocol info") {
  ilu_free(e->x);
}

ILU_QUADEF(BadProtocolInfo);

ILU_DEF_ERR(MalformedKernelObject, "Kernel ilu_Object that has bad fields") {
  ilu_free(e->reason);
}

ILU_QUADEF(MalformedKernelObject);

ILU_DEF_ERR(NoConnection, "Attempt to marshall or unmarshall to/from a non-existent connection") {
}

ILU_QUADEF(NoConnection);

ILU_DEF_ERR(SizeOfFailure, "Couldn't determine size of something") {
  if (e->sub_error != NIL)
    ilu_FreeError (e->sub_error);
}

ILU_QUADEF(SizeOfFailure);

ILU_DEF_ERR(InputFailure, "Couldn't input something") {
  if (e->sub_error != NIL)
    ilu_FreeError (e->sub_error);
}

ILU_QUADEF(InputFailure);

ILU_DEF_ERR(OutputFailure, "Couldn't output something") {
  if (e->sub_error != NIL)
    ilu_FreeError (e->sub_error);
}

ILU_QUADEF(OutputFailure);

ILU_DEF_ERR(WrongServer, "Received discriminant for wrong server") {
  ilu_free(e->oid);
}

ILU_QUADEF(WrongServer);

ILU_DEF_ERR(ClosedServer, "Received discriminant for closed server") {
  ilu_free(e->oid);
}

ILU_QUADEF(ClosedServer);

ILU_DEF_ERR(ObjectNotFoundInServer, "Object can't be located on specified server") {
  ilu_free(e->oid);
}

ILU_QUADEF(ObjectNotFoundInServer);

ILU_DEF_ERR(BadParseOfOID, "OID cannot be parsed into instance handle and server ID") {
  ilu_free(e->oid);
}

ILU_QUADEF(BadParseOfOID);

ILU_DEF_ERR(NoObjectForSBH, "Can't determine object") {
  ilu_free(e->sbh);
}

ILU_QUADEF(NoObjectForSBH);

ILU_DEF_ERR(BadArguments, "Invalid arguments for function") {
  ilu_free(e->reason);
}

ILU_QUADEF(BadArguments);

ILU_DEF_ERR(GcRegFailed, "[Un]RegisterGCInterest failed") {
}

ILU_QUADEF(GcRegFailed);

ILU_DEF_ERR(BadSBH, "Malformed ILU string binding handle") {
  ilu_free(e->theSBH);
}

ILU_QUADEF(BadSBH);

ILU_DEF_ERR(BadURL, "Malformed ILU universal resource locator") {
  ilu_free(e->theURL);
}

ILU_QUADEF(BadURL);

