/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: iluerror.h,v 1.11 1995/02/10 23:07:02 spreitze Exp $ */
/* Last edited by Mike Spreitzer February 10, 1995 10:01 am PST */

#ifndef _ILUERROR_H_
#define _ILUERROR_H_

/***********************************************************************
Error signalling in the ILU kernel.

Each routine that can raise errors does so by returning an ilu_Error,
either as the result or by storing through an ilu_Error* parameter
(normally the last parameter).  A successful invocation will return
a null pointer value (an ilu_Error is a pointer).
An exceptional invocation involves creating an ilu_Error
and returning it.

The macro
	ILU_NO_ERR
expands to a null pointer expression of type ilu_Error; the prudent
programmer will make a practice of writing
	*errptr = ILU_NO_ERR;
so that when he accidentally types
	errptr = ILU_NO_ERR;
instead the compiler will issue a warning.

Each error is an instance of some `type' of errors.
An error carries parameters, which are declared with the error's type
and called the "members" of that error (or error type).  Each type of
error is `declared' once, in a header file that is #included into all
the modules that raise or handle errors of that type; if there's
only one such module, the error type can be declared in that module.
An error type is declared using the following syntax:

	ILU_DECL_ERR(<id>) {
		<memberlist>
	}
	ILU_END_DECL_ERR;

Different error types must use different <id>s; there are no other
restrictions on the <id>s.  The <memberlist> is as in a structure
definition.  Due to C brain-damage, the <memberlist> cannot be empty
(but its one member can be otherwise ignored).  The invocation
	
	ILU_DECL_PARMLESS_ERR(<id>);

declares an error type <id> that has no significant members.


A procedure that can raise errors should declare in its signature the set of possible error types it can raise.  This can be done by using a macro invocation like
	ILU_ERRS((<id1>, <id2>, ... <idN>))
which is defined to expand to
	ilu_Error
The following are example usages:
	
	extern char* substr(char* base, int start, int len,
			    ILU_ERRS((alloc, bad_index)) *err);
	
	extern ILU_ERRS((bad_index)) string_set(char* s,
						int i, char c);

The form
	ILU_ERRMEMP_T(<id>)
expands to an identifier typedef'd to "pointer to a struct containing
the members of error type <id>".  This is needed only for the second
form of error decoding below.

Each error type is `defined' exactly once, using the following syntax:

	ILU_DEF_ERR(<id>, "<descriptive string>") {
		<statements>
	}

Here <id> identifies the error type.  The
<descriptive string> is an arbitrary C string (no embedded \000's)
that describes the error, and is used only for purposes of debugging
the error system itself.  The <statements> form the body of the
procedure used to free errors of this type; they free the members of
the error, but not the error structure itself.  These
statements assume the existence of a variable named `e' containing
a pointer to a struct containing the error's members.

Raising an error requires creating an appropriate error instance and
storing it somewhere (ie, either in a local variable that will be
returned or through the ilu_Error* parameter).  This is done like so:

	ILU_BIND_ERR(<id>, <ep>, <mp>)
		<statement>
	ILU_END_BIND_ERR;

Here <id> identifies the error type, <ep> is an expression that
evaluates to a pointer to where the ilu_Error is to be stored, and
<mp> is an identifier that will be bound in the scope of the
<statement> to a pointer to a struct containing the error's members.
<ep> may be evaluated more than once in the expansion of this
macrology.  It is a checked runtime error for <ep> to evaluate to
a null pointer.  *<ep> is set to point to freshly allocated data
structure, whose members are not guaranteed to be initialized in
any particular way.  The <statement> then sets the structure's
members to appropriate values.

A common special case of the above is when the <statement> just
assigns to members of the error.  This special case is handled more
compactly by a family of macros, illustrated by the following member:

	ILU_ERR_CONS2(<id>, <ep>, <mn1>, <mv1>, <mn2>, <mv2>, <fv>)

This expands to an expression that assigns to *<ep> a fully-created
error instance, and then evaluates <fv>.  The type and value of the
expression are those of <fv>.  The side-effects of the expression are
equivalent to those of:

	ILU_BIND_ERR(<id>, <ep>, ev) {
		ev -> <mn1> = <mv1>;
		ev -> <mn2> = <mv2>;
		}
	ILU_END_BIND_ERR;
	<fv>;

except that no identifier "ev" is used.  Note that the member value
expressions are evaluated once, in the order given, and that <fv> is
evaluated at the end.  This family has macros for errors (error types)
with 0, 1, 2, and 3 members.  Here's an example usage:

	char* substr(char* s, int start, int len, ilu_Error *e)
	{int slen;
	slen = strlen(s);
	if (start<0 || start>slen)
		return ILU_ERR_CONS1(BadIndex, e, i, start, NIL);
	...}

(this example requires that NIL expands to ((void*)0), rather than 0:
"(expr, ..., expr, 0)" (which form the expansion must necessarily
take) is not required to coerce to a char*).

Finally, the erring procedure returns.

A returned error is decoded with the following syntax:

	ILU_ERR_SWITCH(<expr>) {
	  ILU_SUCCESS_CASE
	  	<statement-S>
	  ILU_ERR_CASE(<tid-1>, <vid-1>)
		<statement-1>
	  ILU_ERR_CASE(<tid-2>, <vid-2>)
		<statement-2>
	  ...
	  ILU_ERR_ELSE
		<statement-E>
	}
	ILU_ERR_ENDSWITCH;

where: <expr> evaluates to the ilu_Error that was returned;
<statement-S> is to be executed in the success case (<expr>==NIL);
for each i>0, <statement-i> is to be executed in the extent of <vid-i>,
which will be a variable initialized to a pointer to a struct containing
the error members, iff the error is of the type identified by <tid-i>;
and <statement-E> is to be executed if none of the earlier cases applies.
The "ILU_ERR_ELSE <statement-E>" part can be omitted when the earlier
cases are expected to handle all posibilities.  It is a checked
runtime error for none of the actual cases to apply.  Multiple
types of errors can be handled by one arm using abbreviations like
	ILU_ERR_CASE2(<tid1>, <tid2>)
		<statement>
which handles errors of types <tid1> and <tid2> by the given <statement>;
this abbreviation is available for handling 2, 3, and 4 error types.

There is another convenience invocation
	ILU_MUST_BE_SUCCESS(<expr>);
that is equivalent to
	ILU_ERR_SWITCH(<expr>) { ILU_SUCCESS_CASE ; } ILU_ERR_ENDSWITCH;
This is useful in some places at the boundary of the kernel.

Alternatively, the ilu_Error can be decoded by using
automatically-generated procedures that effectively implement the arms
of such type-switch statements.  The declaration of error type <id>
expands to, among other things, a declaration like

  extern ilu_ERRMEMP_T(<id>) ilu_ErrQua_<id>(ilu_Error e);

This procedure takes a non-null ilu_Error (passing NIL is a checked
runtime error) and returns a pointer to its members iff the error is
of the identified type, or NIL if the error is not of the identified
type.  An advantage of this decoding style is that the error handling
code has no macro invocations; the error type declaration and
definition suffice.

Temporarily, until the error system is pervasive, there is a need to
convert into some other error handling regime.  This typically
involves printing a message to stderr.  For this interim purpose only,
the the name and description of an ilu_Error value may be accessed
with the macros declared by the following pseudocode:

	const char *ILU_ERR_NAME(ilu_Error err);
	const char *ILU_ERR_DESCRIPTION(ilu_Error err);

After an error is handled, it must be freed by a call on the following routine:

	void ilu_FreeError (ilu_Error e);

which does not itself raise any errors.  ilu_FreeError may be passed
NIL, in which case it will do nothing.

C code that uses this error system in the way prescribed above ---
using only the defined macro syntax, not any knowledge about how the
macros expand --- effectively achieves type safety through runtime
typing of errors.  An error value can only be created by use of the
ILU_BIND_ERR macro, which always tags the error value consistently
with the structure allocated.  Members of error structures are only
accessible within ILU_ERR_SWITCH constructs, which invoke
error-type-specific code only on errors tagged with that error type,
or through the use of ilu_ErrQua_<id> procedures.

For non-C code (ie, non-C language-specific runtimes), achieving type
safety is a little harder.  Of course, we can't even consider it
unless the language's type system can describe the C data types used
in the errors; let's consider what to do when it can.

This error system affords two levels of safety.  The first is provided
by the ilu_ErrQua_<id> procedures.  These procedure encapsulate the
run-time type checking.  If the non-C code accesses error parameters
using only these procedures, it will be type-safe --- provided the
ilu_ErrQua_<id> procedures are declared correctly in the non-C
language.

The ilu_ErrQua_<id> procedures need to be defined as well as declared.
To minimize kernel size in some configurations (eg, C LS runtime
only), we put those definitions in a separate file.  That is, we might
expect that purely C programs will not use the ilu_ErrQua_<id> style of
decoding, and so need not include the code for the ilu_ErrQua_<id>
procedures.  For each error type <id>, the macro

	ILU_QUADEF(<id>);

is invoked in that separate file, and expands into the definition
of ilu_ErrQua_<id>.

The second, higher, level of type-safety is built on the first, by
automatically transliterating the C header (".h") files into the non-C
language of interest.  This guarantees that the ilu_ErrQua_<id>
procedures are declared correctly.  Sadly, I don't know of any program
that does such automatic transliteration to the non-C languages of
interest to us.

Following is the stuff you shouldn't depend on.
*/

typedef struct _ilu_ErrorBase_s *ilu_Error;

typedef struct {
  char *name;
  char *description;
  void (*freeproc) (ilu_Error);
} ilu_ErrorTypeStruct, *ilu_ErrorType;

struct _ilu_ErrorBase_s {
  ilu_ErrorType type;
  void *data;
};

#define ILU_ERRNIL ((void*)0)

#define ILU_NO_ERR ((ilu_Error) ILU_ERRNIL)

#define ILU_ERRMEMP_T(x)	ilu_EV2_##x

#ifdef ILU_SOLARIS2_SUNPRO_CPLUSPLUS_CPP_CONCAT_BUG
#define ILU_DECL_ERR(x)		extern ilu_ErrorTypeStruct ilu_ETS_##x ; \
typedef struct ilu_EVS_##x##_s ilu_EVS_##x , *ilu_EV_##x ; \
typedef struct ilu_EVS2_##x##_s ilu_EVS2_##x, *ilu_EV2_##x; \
extern ILU_ERRMEMP_T(x) ilu_ErrQua_##x (ilu_Error e); \
struct ilu_EVS_##x##_s { ilu_ErrorType type;   ilu_EV2_##x (data); }; \
struct ilu_EVS2_##x##_s
/* We write "(data)", instead of "data", in the penultimate line to
   work around a bug in SunPro C++ for Solaris2.3. */
#else
#define ILU_DECL_ERR(x)		extern ilu_ErrorTypeStruct ilu_ETS_##x ; \
typedef struct ilu_EVS_##x##_s ilu_EVS_##x , *ilu_EV_##x ; \
typedef struct ilu_EVS2_##x##_s ilu_EVS2_##x, *ilu_EV2_##x; \
extern ILU_ERRMEMP_T(x) ilu_ErrQua_##x (ilu_Error e); \
struct ilu_EVS_##x##_s { ilu_ErrorType type;   ilu_EV2_##x data; }; \
struct ilu_EVS2_##x##_s
#endif

#define ILU_END_DECL_ERR

#define ILU_DECL_PARMLESS_ERR(id) ILU_DECL_ERR(id) {int ignoreme;} ILU_END_DECL_ERR

#define ILU_ERRS(x) ilu_Error

#ifdef ILU_SOLARIS2_SUNPRO_CPLUSPLUS_CPP_CONCAT_BUG
#define ILU_DEF_ERR(id,xdescrip)    static void ilu_EF_##id (ilu_Error); \
ilu_ErrorTypeStruct ilu_ETS_##id = { # id , xdescrip, ilu_EF_##id }; \
static void ilu_EF2_##id (ilu_EV2_##id); \
static void ilu_EF_##id (ilu_Error ee) \
{ ilu_EF2_##id (((ilu_EV_##id) ee)->data); } \
static void ilu_EF2_##id (ilu_EV2_##id (e))
#else
#define ILU_DEF_ERR(id,xdescrip)    static void ilu_EF_##id (ilu_Error); \
ilu_ErrorTypeStruct ilu_ETS_##id = { # id , xdescrip, ilu_EF_##id }; \
static void ilu_EF2_##id (ilu_EV2_##id); \
static void ilu_EF_##id (ilu_Error ee) \
{ ilu_EF2_##id (((ilu_EV_##id) ee)->data); } \
static void ilu_EF2_##id (ilu_EV2_##id e)
#endif

#define _ILU_ERR_CONS(x)		((ilu_EV_##x ) _ilu_ErrCons(&ilu_ETS_##x , sizeof(struct ilu_EVS2_##x##_s)))

extern void * _ilu_ErrCons (ilu_ErrorType et, unsigned int size);

#define ILU_QUA_ERR(x)		((ilu_Error) (x))

#ifdef ILU_SOLARIS2_SUNPRO_CPLUSPLUS_CPP_CONCAT_BUG
#define ILU_BIND_ERR(id,ep,mp)   do {ilu_EV2_##id (mp); if (ep == ILU_ERRNIL) _ilu_Assert(0, #ep " is null"); else { *ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)); mp = ((ilu_EV_##id) *ep)->data; } if (ep != ILU_ERRNIL)
#else
#define ILU_BIND_ERR(id,ep,mp)   do {ilu_EV2_##id mp; if (ep == ILU_ERRNIL) _ilu_Assert(0, #ep " is null"); else { *ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)); mp = ((ilu_EV_##id) *ep)->data; } if (ep != ILU_ERRNIL)
#endif

#define ILU_END_BIND_ERR } while (0)

/* The following assume "_ilu_Assert(b,s)" is an expression. */

#define ILU_ERR_CONS0(id, ep, fv) ( \
	_ilu_Assert(ep != ILU_ERRNIL, #ep " is null"), \
	*ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)), \
	fv )

#define ILU_ERR_CONS1(id, ep, mn1, mv1, fv) ( \
	_ilu_Assert(ep != ILU_ERRNIL, #ep " is null"), \
	*ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)), \
	((ilu_EV_##id) *ep)->data->mn1 = mv1, \
	fv )

#define ILU_ERR_CONS2(id, ep, mn1, mv1, mn2, mv2, fv) ( \
	_ilu_Assert(ep != ILU_ERRNIL, #ep " is null"), \
	*ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)), \
	((ilu_EV_##id) *ep)->data->mn1 = mv1, \
	((ilu_EV_##id) *ep)->data->mn2 = mv2, \
	fv )

#define ILU_ERR_CONS3(id, ep, mn1, mv1, mn2, mv2, mn3, mv3, fv) ( \
	_ilu_Assert(ep != ILU_ERRNIL, #ep " is null"), \
	*ep = ILU_QUA_ERR(_ILU_ERR_CONS(id)), \
	((ilu_EV_##id) *ep)->data->mn1 = mv1, \
	((ilu_EV_##id) *ep)->data->mn2 = mv2, \
	((ilu_EV_##id) *ep)->data->mn3 = mv3, \
	fv )

#define ILU_ERR_SWITCH(x)	\
do { ilu_Error __ilu_CurrentError = (ilu_Error) (x); \
     if (0)

#ifdef ILU_SOLARIS2_SUNPRO_CPLUSPLUS_CPP_CONCAT_BUG
#define ILU_ERR_CASE(tid,vid)		} else if (__ilu_CurrentError != ILU_ERRNIL && __ilu_CurrentError->type == &ilu_ETS_##tid ) { \
ilu_EV2_##tid (vid) ; \
vid = ((ilu_EV_##tid ) __ilu_CurrentError)->data;
#else
#define ILU_ERR_CASE(tid,vid)		} else if (__ilu_CurrentError != ILU_ERRNIL && __ilu_CurrentError->type == &ilu_ETS_##tid ) { \
ilu_EV2_##tid vid ; \
vid = ((ilu_EV_##tid ) __ilu_CurrentError)->data;
#endif

#define ILU_SUCCESS_CASE	} else if (__ilu_CurrentError == ILU_ERRNIL) {

#define ILU_ERR_ELSE		} else if ( 1 ) {

#define ILU_ERR_ENDSWITCH	else {char buf[1000]; sprintf(buf, "unhandled error %s at line %d in file %s", ((__ilu_CurrentError!=ILU_ERRNIL)?__ilu_CurrentError->type->name:"SUCCESS"), __LINE__, __FILE__); _ilu_Assert(0, buf);} } while (0)

#define ILU_ERR_CASE2(t1, t2)	} else if (__ilu_CurrentError != ILU_ERRNIL && (__ilu_CurrentError->type == &ilu_ETS_##t1 || __ilu_CurrentError->type == &ilu_ETS_##t2) ) {

#define ILU_ERR_CASE3(t1, t2, t3)	} else if (__ilu_CurrentError != ILU_ERRNIL && (__ilu_CurrentError->type == &ilu_ETS_##t1 || __ilu_CurrentError->type == &ilu_ETS_##t2 || __ilu_CurrentError->type == &ilu_ETS_##t3) ) {

#define ILU_ERR_CASE4(t1, t2, t3, t4)	} else if (__ilu_CurrentError != ILU_ERRNIL && (__ilu_CurrentError->type == &ilu_ETS_##t1 || __ilu_CurrentError->type == &ilu_ETS_##t2 || __ilu_CurrentError->type == &ilu_ETS_##t3 || __ilu_CurrentError->type == &ilu_ETS_##t4) ) {

#define ILU_MUST_BE_SUCCESS(x) ILU_ERR_SWITCH(x) { ILU_SUCCESS_CASE ; } ILU_ERR_ENDSWITCH;

#define ILU_ERR_NAME(err)	 ((err)->type->name)
#define ILU_ERR_DESCRIPTION(err) ((err)->type->description)

#define ILU_QUADEF(x)		ILU_ERRMEMP_T(x) ilu_ErrQua_##x (ilu_Error e) \
{ \
  if (e->type == &ilu_ETS_##x ) \
	return ((ilu_EV_##x ) e)->data; \
  else return (ILU_ERRNIL); \
} \
ILU_ERRMEMP_T(x) ilu_ErrQua_##x (ilu_Error e)

extern void ilu_FreeError (ilu_Error);	



#endif /* ndef _ILUERROR_H_ */
