/*
Copyright (c) 1991-1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: iluerrs.h,v 1.12 1995/02/23 02:43:06 janssen Exp $ */
/* Last tweaked by Mike Spreitzer November 3, 1994 9:07 am PST */

#ifndef _ILUERRS_H_
#define _ILUERRS_H_

/* Some standard errors for use throughout the kernel */

/*
 * Used internally in the kernel to report that some invariant
 * doesn't hold.  That is, a bug has been detected in the kernel.
 * Used only for invariants that don't require cooperation from the
 * app or LS runtime.  When raised, the called procedure has done
 * an unspecified part of its job.  BadDataStructure is a special
 * case, and is used instead where it applies.
 */
ILU_DECL_ERR(KernelBroken)
{
  char           *what;		/* constant string that alludes to
				 * the invariant that's broken */
} ILU_END_DECL_ERR;

/*
 * Used to report violation of some invariant that requires
 * cooperation of the app or LS runtime.  Called proc has done an
 * unspecified part of its job.  While we should never return such
 * a vague error to apps, it might be OK to use it in the kernel
 * and let the LS runtime figure out a more specific thing to
 * report.  Then again, it might be better to just use a more
 * specific error in the kernel in the first place.
 */
ILU_DECL_ERR(RuntimeBroken)
{
  char           *what;		/* constant string that alludes to
				 * the invariant that's broken */
} ILU_END_DECL_ERR;

/*
 * Used internally in the kernel to report that something in the
 * kernel is "broken" in some structural way (e.g., a pointer is
 * NULL that shouldn't be).  Called proc has done an unspecified
 * part of its job. This is a special case of KernelBroken.
 */
ILU_DECL_ERR(BadDataStructure)
{
  void           *data_str_ptr;
  char           *supposed_type;/* pointer to constant string,
				 * never freed */
  char           *why_bad;	/* pointer to constant string,
				 * never freed */
} ILU_END_DECL_ERR;

#define BIND_BadDataStructure(e,p,st,wb) \
ILU_BIND_ERR(BadDataStructure, e, ev) \
      { \
	  ev->data_str_ptr = p; \
	  ev->supposed_type = st; \
	  ev->why_bad = wb; \
      } \
ILU_END_BIND_ERR;

/* Raised when there is some reason to believe that a kernel
   object is broken.  The "reason" is some explanation as to
   why that belief is held. */

ILU_DECL_ERR(MalformedKernelObject) {
  ilu_Object obj;
  ilu_string reason;
} ILU_END_DECL_ERR;

/* Raised when a fixed-length array in the impl overflows. */

ILU_DECL_ERR(MaxCountExceeded) {
  int max_count;
}
ILU_END_DECL_ERR;

/* Signalled by ilu_RegisterProtocol */

ILU_DECL_ERR(ProtocolAlreadyRegistered) {
  char * name;
  ilu_Protocol (*old_protocol) (void);
  ilu_Protocol (*new_protocol) (void);
}
ILU_END_DECL_ERR;

/* Signalled by ilu_RegisterTransport */

ILU_DECL_ERR(TransportAlreadyRegistered) {
  char *name;
  ilu_TransportClass (*old_transport) (void);
  ilu_TransportClass (*new_transport) (void);
}
ILU_END_DECL_ERR;

/*
 * Raised when the caller violates a locking precondition of a
 * procedure.  /which/ identifies a mutex relevent to the violated
 * constraint.  Signalled by lots of stuff.  When raised, nothing
 * else has been done, unless otherwise noted at the procedure in
 * question.
 */

typedef enum {
  ilu_lock_smu, ilu_lock_otmu, ilu_lock_cmu, ilu_lock_prmu,
  ilu_lock_trmu, ilu_lock_gcmu, ilu_lock_timu, ilu_lock_server,
  ilu_lock_io, ilu_lock_call
}               ilu_lock;

ILU_DECL_ERR(WrongLocks)
{
  ilu_lock        which;
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadContactInfo)
{
  ilu_string      x;		/* owned by caller of erring proc */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadTransportInfo)
{
  ilu_string      x;		/* owned by error */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadProtocolInfo)
{
  ilu_string      x;		/* owned by error */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadArguments)
{
  ilu_string      function_name;/* global -- not to be freed */
  ilu_string      reason;	/* explanation of what was wrong;
				 * owned by err */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(GcRegFailed)
{
  ilu_string      why;		/* global -- not to be freed */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadSBH)
{
  ilu_string theSBH;		/* copy of offending sbh */
} ILU_END_DECL_ERR;

ILU_DECL_ERR(BadURL)
{
  ilu_string theURL;		/* copy of offending URL */
} ILU_END_DECL_ERR;

#endif /* ndef _ILUERRS_H_ */

