/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: locks.c,v 1.25 1995/02/10 17:40:48 spreitze Exp $ */
/* Last tweaked by Mike Spreitzer February 9, 1995 2:34 pm PST */

#define _POSIX_SOURCE

#include "iluntrnl.h"

/*L1 not meaningful; this module implements L1 locks;
  L2, Main unconstrained*/


typedef struct {
  ilu_string d1, d2;
  int held;
} DefaultMutex;

static ilu_private Default_CreateMutex(ilu_string d1, ilu_string d2)
{ DefaultMutex *dm = (DefaultMutex*) ilu_malloc(sizeof(DefaultMutex));
  dm->d1 = _ilu_Strdup(d1);
  dm->d2 = _ilu_Strdup(d2);
  dm->held = 0;
  return ((ilu_private) dm);
}

static void Default_AcquireMutex(ilu_private m)
{ DefaultMutex *dm = (DefaultMutex*) m;
  ASSERT(!dm->held, buf,
	 (buf, "Trying to acquire held mutex %s%s", dm->d1, dm->d2));
  dm->held = 1;
}

static void Default_HoldMutex(ilu_private m)
{ DefaultMutex *dm = (DefaultMutex*) m;
  ASSERT(dm->held, buf,
	 (buf, "Not holding mutex %s%s", dm->d1, dm->d2));
}

static void Default_ReleaseMutex(ilu_private m)
{ DefaultMutex *dm = (DefaultMutex*) m;
  ASSERT(dm->held, buf,
	 (buf, "Trying to release non-held mutex %s%s", dm->d1, dm->d2));
  dm->held = 0;
}

static void Default_DestroyMutex(ilu_private m)
{ return; }

static ilu_LockTech Default_LockTech[1] = {
	{Default_CreateMutex, Default_AcquireMutex,
	 Default_HoldMutex, Default_ReleaseMutex, Default_DestroyMutex,
	 NULLFN, NULLFN, NULLFN, NULLFN} };

static ilu_LockTech *theLockTech = Default_LockTech;
static int ltPhase = 0;
static int nIn = 0;
static int stdDumped = 0;

static DefaultMutex def_smu = {"global ", "smu", 0};
static DefaultMutex def_otmu = {"global ", "otmu", 0};
static DefaultMutex def_cmu = {"global ", "cmu", 0};
static DefaultMutex def_prmu = {"global ", "prmu", 0};
static DefaultMutex def_trmu = {"global ", "trmu", 0};
static DefaultMutex def_gcmu = {"global ", "gcmu", 0};
static DefaultMutex def_daimu = {"global ", "daimu", 0};

ilu_Mutex ilu_smu = &def_smu;
ilu_Mutex ilu_otmu = &def_otmu;
ilu_Mutex ilu_cmu = &def_cmu;
ilu_Mutex ilu_prmu = &def_prmu;
ilu_Mutex ilu_trmu = &def_trmu;
ilu_Mutex ilu_gcmu = &def_gcmu;
ilu_Mutex ilu_daimu = &def_daimu;

ilu_Mutex ilu_CreateMutex(ilu_string d1, ilu_string d2)
{ return _ilu_CreateMutex(d1, d2); }

ilu_Mutex _ilu_CreateMutex(ilu_string d1, ilu_string d2)
{
  ilu_private m;
  _ilu_AutoSetDebugLevel();
  ltPhase = 1;
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_CreateMutex:  %s, %s\n", d1, d2));
  m = (theLockTech->lt_mcreate)(d1, d2);
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_CreateMutex:  => %p\n", m));
  return m;
}

void ilu_AcquireMutex(ilu_Mutex m)
{ _ilu_AcquireMutex(m); }

void _ilu_AcquireMutex(ilu_Mutex m)
{
  _ilu_AutoSetDebugLevel();
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_AcquireMutex:  %p\n", m));
  nIn++;
  if (!stdDumped) {
    DEBUG(LOCK_DEBUG, (stderr, "ilu_smu = %p\n", ilu_smu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_otmu = %p\n", ilu_otmu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_cmu = %p\n", ilu_cmu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_prmu = %p\n", ilu_prmu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_trmu = %p\n", ilu_trmu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_gcmu = %p\n", ilu_gcmu));
    DEBUG(LOCK_DEBUG, (stderr, "ilu_daimu = %p\n", ilu_daimu));
    stdDumped = 1;
  }
  (theLockTech->lt_acquire) (m);
  return;
}

void ilu_HoldMutex(ilu_Mutex m)
{ _ilu_HoldMutex(m); }

void _ilu_HoldMutex(ilu_Mutex m)
{
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_HoldMutex:  %p\n", m));
  (theLockTech->lt_hold)(m);
  return;
}

void ilu_ReleaseMutex(ilu_Mutex m)
{ _ilu_ReleaseMutex(m); }

void _ilu_ReleaseMutex(ilu_Mutex m)
{
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_ReleaseMutex:  %p\n", m));
  (theLockTech->lt_release)(m);
  nIn--;
  return;
}

void ilu_DestroyMutex(ilu_Mutex m)
{ _ilu_DestroyMutex(m); }

void _ilu_DestroyMutex(ilu_Mutex m)
{
  ltPhase = 1;
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_DestroyMutex:  %p\n", m));
  (theLockTech->lt_mdestroy)(m);
  return;
}

ilu_boolean ilu_CanCondition(void)
{ return _ilu_CanCondition(); }

ilu_boolean _ilu_CanCondition(void)
{
  if (theLockTech->lt_ccreate != NULLFN)
       return TRUE;
  else return FALSE;
}

ILU_DEF_ERR(CantCondition, "Condition variables not available to kernel")
{}

ilu_Condition 
ilu_CreateCondition(ilu_string d1, ilu_string d2,
		     ILU_ERRS((CantCondition)) * err)
{ return _ilu_CreateCondition(d1, d2, err); }

ilu_Condition 
_ilu_CreateCondition(ilu_string d1, ilu_string d2,
		     ILU_ERRS((CantCondition)) * err)
{
  ilu_Condition     c;

  ltPhase = 1;
  if (theLockTech->lt_ccreate == NULLFN)
    return ILU_ERR_CONS0(CantCondition, err, NIL);
  else *err = ILU_NO_ERR;
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_CreateCondition:  %s, %s\n", d1, d2));
  c = (*theLockTech->lt_ccreate) (d1, d2);
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_CreateCondition:  => %p\n", c));
  return (c);
}

ILU_ERRS((CantCondition)) ilu_NotifyCondition(ilu_Condition c)
{ return _ilu_NotifyCondition(c); }

ILU_ERRS((CantCondition)) _ilu_NotifyCondition(ilu_Condition c)
{
  ilu_Error       err;
  ltPhase = 1;
  if (theLockTech->lt_notify == NULLFN)
    return ILU_ERR_CONS0(CantCondition, &err, err);
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_NotifyCondition:  %p\n", c));
  (*theLockTech->lt_notify) (c);
  return NIL;
}

ILU_ERRS((CantCondition)) ilu_DestroyCondition(ilu_Condition c)
{ return _ilu_DestroyCondition(c); }

ILU_ERRS((CantCondition)) _ilu_DestroyCondition(ilu_Condition c)
{
  ilu_Error       err;
  ltPhase = 1;
  if (theLockTech->lt_cdestroy == NULLFN)
    return ILU_ERR_CONS0(CantCondition, &err, err);
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_DestroyCondition:  %p\n", c));
  (*theLockTech->lt_cdestroy)(c);
  return NIL;
}

ILU_ERRS((CantCondition)) ilu_WaitCondition(ilu_Condition c, ilu_Mutex m)
{ return _ilu_WaitCondition(c, m); }

ILU_ERRS((CantCondition)) _ilu_WaitCondition(ilu_Condition c, ilu_Mutex m)
{
  ilu_Error       err;
  ltPhase = 1;
  if (theLockTech->lt_wait == NULLFN)
    return ILU_ERR_CONS0(CantCondition, &err, err);
  DEBUG(LOCK_DEBUG, (stderr, "_ilu_WaitCondition:  c = %p, m = %p\n",
		     c, m));
  (*theLockTech->lt_wait)(c, m);
  return NIL;
}

void ilu_SetLockTech(ilu_LockTech *lt)
{
  _ilu_AutoSetDebugLevel();
  if (lt != NIL)
    theLockTech = lt;
  DEBUG(LOCK_DEBUG,
	(stderr,
	"ilu_RegisterLockTech (%p), ltPhase == %d, nIn == %d\n",
	 lt, ltPhase, nIn));
  _ilu_Assert(ltPhase == 0 && nIn == 0, "RegisterLockTech");
  ilu_smu = _ilu_CreateMutex("global ", "smu");
  ilu_otmu = _ilu_CreateMutex("global ", "otmu");
  ilu_cmu = _ilu_CreateMutex("global ", "cmu");
  ilu_prmu = _ilu_CreateMutex("global ", "prmu");
  ilu_trmu = _ilu_CreateMutex("global ", "trmu");
  ilu_gcmu = _ilu_CreateMutex("global ", "gcmu");
  ilu_daimu = _ilu_CreateMutex("global ", "daimu");
  return;
}

