/*
Copyright (c) 1991-1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: object.c,v 1.78 1995/02/23 23:01:08 spreitze Exp $ */
/* Last edited by Mike Spreitzer February 23, 1995 2:34 pm PST */

#define _POSIX_SOURCE

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluntrnl.h"

#include "object.h"
#include "server.h"
#include "type.h"
#include "protocol.h"
#include "port.h"
#include "vector.h"

#include "version.h"	/* defines ILU_VERSION */

/*L1, L2, Main unconstrained*/

static const ilu_string _ilu_id = "$" "Id: ILU version " ILU_VERSION_STRING ".  Copyright 1990-1994 Xerox Corporation.  All Rights Reserved. $" ;
static const ilu_string _ilu_ilu_version = ILU_VERSION_STRING ;
static const ilu_cardinal _ilu_ilu_major_version = ILU_MAJOR_VERSION ;
static const ilu_cardinal _ilu_ilu_minor_version = ILU_MINOR_VERSION ;

/*L1 >= {cmu}*/
static ilu_ObjectNoter theNoter = NULLFN;
/* The proc to call when the "very interested" predicate changes. */

/*L1_sup < cmu*/
void ilu_SetNoter(ilu_ObjectNoter n)
{
  _ilu_Assert(n!=NULLFN, "SetNoter: given NIL noter");
  _ilu_Assert(theNoter==NULLFN, "SetNoter: already set");
  _ilu_AcquireMutex(ilu_cmu);
  theNoter = n;
  _ilu_ReleaseMutex(ilu_cmu);
  return;
}

/*L1, L2, Main unconstrained*/

static ilu_Server DefaultServer = NIL;

ilu_string ilu_GetILUVersion (void)
{
  return (_ilu_ilu_version);
}

ilu_cardinal ilu_GetILUMajorVersion (void)
{
  return (_ilu_ilu_major_version);
}

ilu_cardinal ilu_GetILUMinorVersion (void)
{
  return (_ilu_ilu_minor_version);
}

void ilu_SetDefaultServer(ilu_Server s)
{ DefaultServer = s;
}

ilu_Server ilu_GetDefaultServer(void)
{ return DefaultServer;
}

static ILU_ERRS((BadContactInfo, BadSBH))
     _ilu_SBHtoURL (char *sbh, char *mstid, char **url)
{
  ilu_boolean stat;
  ilu_Error err;
  char *cinfo = NIL;
  char *oid = NIL;
  char *pinfo = NIL;
  char *tinfo = NIL;

  if ((stat = ilu_ParseStringBindingHandle (sbh, &oid, &cinfo)))
    {
      if ((stat = _ilu_ParseConnectInfo (cinfo, &pinfo, &tinfo)))
	{
	  char host[1000];
	  ilu_cardinal port = 0;
	  char *buf;
	  
	  buf = ilu_must_malloc(2 * strlen(sbh));
	  
	  /* special case for TCP/IP connections */
	  if (strncmp (tinfo, "tcp_", 4) == 0
	      && sscanf (tinfo, "tcp_%[^_]_%lu", host, &port) == 2)
	    {
	      if (mstid != NIL)
		sprintf (buf, "ilu://%s:%lu|%s|%s/%s", host, port, pinfo, mstid, oid);
	      else
		sprintf (buf, "ilu://%s:%lu|%s/%s", host, port, pinfo, oid);
	    }
	  else
	    {
	      if (mstid != NIL)
		sprintf (buf, "ilu:/%s|%s|%s/%s", tinfo, pinfo, mstid, oid);
	      else
		sprintf (buf, "ilu:/%s|%s/%s", tinfo, pinfo, oid);
	    }
	  ilu_free (tinfo);
	  ilu_free (pinfo);
	  if (url != NIL)
	    *url = _ilu_Strdup(buf);
	  ilu_free(buf);
	  return (NIL);
	}
      else
	{
	  ILU_ERR_CONS1(BadContactInfo, &err, x, _ilu_Strdup(cinfo), 0);
	}
    }
  else
    {
      ILU_ERR_CONS1(BadSBH, &err, theSBH, _ilu_Strdup(sbh), 0);
    }
  ilu_free (oid);
  ilu_free (cinfo);
  return (err);
}

static char * urlstralloc (char *from, ilu_cardinal size)
{
  char *mem = (char *) ilu_must_malloc(size + 1);
  strncpy (mem, from, size);
  mem[size] = (char) 0;
  return (mem);
}

static ILU_ERRS((BadURL, BadTransportInfo))
     _ilu_URLtoSBH (char *url, char **sbh, char **mstid)
{
  ilu_Error err = NIL;

  if (sbh != NIL)
    *sbh = NIL;
  if (mstid != NIL)
    *mstid = NIL;

  if (strncmp (url, "ilu:/", 5) == 0)
    {
      /* ILU TCP URL */
      
      char *hostname = NIL;
      ilu_cardinal port = 0;
      char *pinfo = NIL;
      char *xmstid = NIL;
      char *tinfo = NIL;
      register char *p;
      char *p2;

      /* two possible ways to find tinfo, either ilu:/<tinfo>|...
	 or ilu://<ip-hostname>:<ip-port>|... (implicitly TCP) */

      if (url[5] == '/')
	{
	  /* special ILU TCP/IP URL */ 
	  p = url + 6;
	  if ((p2 = strchr (p, ':')) != NIL)
	    {
	      hostname = urlstralloc(p, p2 - p);

	      p = p2 + 1;
	      p2 = NIL;
	      if ((port = _ilu_atoi (p, &p2)) != 0)
		{
		  tinfo = ilu_must_malloc(strlen(hostname) + 20);
		  sprintf (tinfo, "tcp_%s_%lu", hostname, port);
		}
	      ilu_free (hostname);
	    }
	}
      else
	{
	  /* normal generic ILU URL */
	  p = url + 5;
	  if ((p2 = strpbrk (p, "|/")) != NIL)
	    tinfo = urlstralloc (p, p2 - p);
	}
      if (tinfo != NIL && (p2 != NIL))
	{
	  if (*p2 == '|')	/* optional pinfo there? */
	    {
	      p = p2 + 1;
	      if ((p2 = strpbrk(p, "|/")) != NIL)
		{
		  pinfo = urlstralloc (p, p2 - p);
		}
	    }
	  if (p2 != NIL && (*p2 == '|'))	/* mstid there? */
	    {
	      p = p2 + 1;
	      if ((p2 = strchr (p, '/')) != NIL)
		{
		  xmstid = urlstralloc (p, p2 - p);
		}
	    }
	  if (p2 != NIL && (*p2 == '/'))
	    {
	      char buf[2000];

	      sprintf (buf, "%s@%s|%s", p2 + 1, (pinfo == NIL) ? "sunrpc_2_0_0" : pinfo, tinfo);
	      if (sbh != NIL)
		*sbh = _ilu_Strdup (buf);
	      if (mstid != NIL)
		*mstid = xmstid;
	    }
	  if (pinfo == NIL)
	    ilu_free (pinfo);
	}
      else
	{
	  ILU_ERR_CONS1(BadTransportInfo, &err, x, _ilu_Strdup(url), 0);
	}
      if (tinfo != NIL)
	ilu_free (tinfo);
    }
  if (*sbh == NIL)
    {
      ILU_ERR_CONS1(BadURL, &err, theURL, _ilu_Strdup(url), 0);
    }
  return (err);
}

ILU_ERRS((BadSBH, BadURL, BadTransportInfo))
     ilu_ParseSBH (ilu_string sbh, ilu_string *oid, ilu_string *connectInfo, ilu_string *mstid)
{
  char *realsbh = NIL;
  ilu_Error err = NIL;

  if (strncmp (sbh, "ilu:/", 5) == 0)
    {
      if ((err = _ilu_URLtoSBH (sbh, &realsbh, mstid)) != NIL)
	return (err);
    }
  else
    realsbh = sbh;
  if (NOT ilu_ParseStringBindingHandle (realsbh, oid, connectInfo))
    {
      ILU_ERR_CONS1(BadSBH, &err, theSBH, _ilu_Strdup(realsbh), 0);
    }
  if (realsbh != sbh)
    ilu_free (realsbh);
  return (err);
}

ilu_boolean ilu_ParseStringBindingHandle (ilu_string sbh,
					  ilu_string *oid,
					  ilu_string *connectInfo)
{
  ilu_integer count;

/*
  DEBUG(OBJECT_DEBUG,
	(stderr, "ilu_ParseStringBindingHandle:  SBH <%s>.\n", sbh));
*/

  if (sbh == NIL)
    return (FALSE);

  for (count = 0;  sbh[count] != '\0';  count += 1) {
      if (sbh[count] == '@') {
	  if (sbh[count+1] == '@')
	    count += 1;
	  else
	    break;
	}
    }

  if (sbh[count] == '@')
    count += 1;

  for (;  sbh[count] != '\0';  count += 1) {
      if (sbh[count] == '@') {
	  if (sbh[count + 1] == '@')
	    count += 1;
	  else
	    break;
	}
    }

  if (oid != NIL) {
      if (count > 0) {
	  *oid = (ilu_string) ilu_malloc (count + 1);
	  strncpy (*oid, sbh, count);
	  (*oid)[count] = '\0';
	}
      else
	*oid = NIL;
    }

  if (sbh[count] == '@')
    count += 1;

  if (connectInfo != NIL) {
      if (sbh[count] != '\0') {
	  *connectInfo = (ilu_string) ilu_malloc (strlen (sbh + count) + 1);
	  strcpy(*connectInfo, sbh + count);
	}
      else
        *connectInfo = NIL;
    }


  DEBUG(OBJECT_DEBUG,
	(stderr,
	 "ilu_ParseStringBindingHandle:  parts are <%s> and <%s>\n",
	 (oid == NIL) ? "*" : *oid,
	 (connectInfo == NIL) ? "*" : *connectInfo));

  return (TRUE);
}

ilu_boolean ilu_ParseOID (ilu_string oid, ilu_string *instH,
					  ilu_string *serverID)
{
  ilu_integer count;
/*
  DEBUG(OBJECT_DEBUG, (stderr, "ilu_ParseOID:  OID <%s>.\n", oid));
*/
  if (oid == NIL)
    return (FALSE);

  for (count = 0;  oid[count] != '\0';  count += 1) {
      if (oid[count] == '@') {
	  if (oid[count+1] == '@')
	    count += 1;
	  else
	    break;
	}
    }

  if (instH != NIL) {
      if (count > 0) {
	  *instH = (ilu_string) ilu_malloc (count + 1);
	  strncpy (*instH, oid, count);
	  (*instH)[count] = '\0';
	}
      else
	*instH = NIL;
    }

  if (oid[count] == '@')
    count += 1;

  if (serverID != NIL) {
      if (oid[count] != '\0') {
	  *serverID = (ilu_string) ilu_malloc (strlen (oid + count) + 1);
	  strcpy(*serverID, oid + count);
	}
      else
        *serverID = NIL;
    }

  DEBUG(OBJECT_DEBUG, (stderr, "ilu_ParseOID:  parts are <%s> and <%s>\n",
		       (  instH  == NIL) ? "*" : *instH,
		       (serverID == NIL) ? "*" : *serverID));

  return (TRUE);
}

/*L1, L2, Main unconstrained*/

/* This procedure is used internally for creating both true and surrogate objects.  For true objects, this procedure is called with non-NIL ih, server, cl, and lspo; for surrogate objects, this procedure is called with non-NIL ih, server, and sbh.  Storage for ih and sbh is owned by the object, for mstid is owned by the caller. */
static ilu_Object CreateObject (ilu_string ih, ilu_Server server,
				ilu_Class class, ilu_refany lspo,
				ilu_string sbh, ilu_string mstid)
{
  ilu_Object new = (ilu_Object) ilu_malloc(sizeof(struct _ilu_Object_s));
  ilu_Alarmette_s gco = {NIL, NIL, FALSE, {0, 0}};
  
  object_oid(new)	= _ilu_Strcat3(ih, "@", server_id(server));
  object_ih(new)	= ih;
  object_server(new)	= server;
  object_timeout(new)	= 1200;
  object_class(new)	= class;
  object_mstid(new)	= _ilu_Strdup(mstid);
  object_sbh(new)	= sbh;
  object_lspo(new)	= lspo;
  object_holds(new)	= 0;
  object_intNoted(new)	= FALSE;
  object_notifying(new)	= FALSE;
  object_known(new)	= FALSE;
  object_gco(new)	= gco;
  object_lastRemote(new) = 0;
  object_gclist(new)	= NIL;
  return (new);
}

/*Destroy an object created but not yet put into the server's object table. */
static void UncreateObject(ilu_Object o)
{
  ilu_free(o->ob_oid);
  ilu_free(o->ob_ih);
  if (o->ob_mstid != NIL)
    ilu_free(o->ob_mstid);
  if (o->ob_sbh != NIL)
    ilu_free(o->ob_sbh);
  ilu_free(o);
}

/*L1 >= {s}*/

ilu_boolean _ilu_Addable(ilu_Server s, ilu_Class t, ilu_Object *h)
{
  int i, l;
  if (!class_singleton(t))
      return TRUE;
  if (  (*h = (ilu_Object) _ilu_hash_FindInTable(server_singles(s), t))
	!= NIL)
      return FALSE;
  l = class_superclass_count(t);
  for (i = 0; i < l; i++)
      if (!_ilu_Addable(s, class_superclass(t, i), h))
          return FALSE;
  return TRUE;
}

/*
 * Bug: if the ancestors don't form a tree, some ancestors will be
 * visited more than once, and the later visits will fail in
 * _ilu_hash_AddToTable.
 */
void 
_ilu_AddSingleton(ilu_Server s, ilu_Class t, ilu_Object o)
{
  int             i, l;
  if (!class_singleton(t))
    return;
  _ilu_Assert(_ilu_hash_AddToTable(server_singles(s), t, o),
	      "AddSingleton");
  l = class_superclass_count(t);
  for (i = 0; i < l; i++)
    _ilu_AddSingleton(s, class_superclass(t, i), o);
  return;
}

/*L1 >= {the object's server};
  L1 >= {gcmu} if cl collectible*/
ilu_Object ilu_FindOrCreateTrueObject (ilu_string ih, ilu_Server server,
					ilu_Class cl, ilu_refany lspo)
{ 
  ilu_Object new;
  if (server == NIL) {
      server = DefaultServer;
      if (server == NIL)
        return (NIL);
    }
  _ilu_Assert(ih != NIL, "CreateTrueObject: ih==NIL");
  _ilu_Assert(lspo != NIL, "CreateTrueObject: lspo==NIL");
  _ilu_Assert(cl != NIL, "CreateTrueObject: class==NIL");
  _ilu_HoldMutex(server_lock(server));
  new = (ilu_Object) _ilu_hash_FindInTable(server_objs(server), ih);
  if (new != NIL) {
      if (new->ob_class != cl) {
          DEBUG(OBJECT_DEBUG,
		(stderr,
		 "FindOrCreateTrueObject: type of %s is %s, not %s",
		 new->ob_oid, new->ob_class->cl_unique_id, cl->cl_unique_id));
          return (NIL);
        }
      return (new);
    }
  if (server->sr_closing == TRUE) {
      DEBUG(OBJECT_DEBUG,
	    (stderr,
	     "ilu_CreateTrueObject: invoked on closing server %s.\n",
	     server->sr_id));
      return (NIL);
    }
  if (!_ilu_Addable(server, cl, &new)) {
      DEBUG(OBJECT_DEBUG,
	    (stderr,
	     "%s %s %s of type %s in server %s because %s %s.\n",
	     "ilu_CreateTrueObject: can't create another",
	     "singleton object", ih, cl->cl_unique_id, server->sr_id, new->ob_ih,
	     "already exists"));
      return (NIL);
    }
  new = CreateObject(ih, server, cl, lspo, NIL, cl->cl_unique_id);
  _ilu_Assert(_ilu_hash_AddToTable(server_objs(server), ih, new),
	      "FindOrCreateTrueObject: AddToTable failed");
  _ilu_AddSingleton(server, cl, new);
  if (class_collectible(cl))
    _ilu_StartGCingTrueObj(new);
  return new;
}

/*before: L1 = {};
  after:  result!=NIL => Inside(result's server, cl);
  after:  result==NIL => L1 = {};
  forall conn: (L2 >= {conn.iomu}) => (L2 >= {conn.callmu});
  Main otherwise unconstrained*/
ilu_Object ilu_ObjectOfSBH (ilu_string sbh, ilu_string mstid_parm,
			    ilu_Class static_type)
{
  ilu_string connectInfo = NIL, oid = NIL;
  ilu_string ih = NIL, serverID = NIL;
  ilu_Object o = NIL;
  ilu_boolean tidfound = FALSE;
  ilu_Class c2;
  ilu_Server s;
  ilu_string mstid = NIL;
  ilu_Error err;

  _ilu_AutoSetDebugLevel();
  
  _ilu_Assert(sbh != NIL, "ObjectOfSBH: sbh==NIL");
  ASSERT(static_type!=NIL, buf,
	 (buf, "ObjectOfSBH: static_type==NIL, sbh=%s", sbh));

  if ((err = ilu_ParseSBH (sbh, &oid, &connectInfo, &mstid)) != NIL
      || oid == NIL || connectInfo == NIL) {
      DEBUG(OBJECT_DEBUG,
	    (stderr, "ObjectOfSBH: SBH parse of %s unsatisfactory.\n",
	     sbh));
      ilu_FreeError (err);
      return o;
    }
  if (!ilu_ParseOID (oid, &ih, &serverID)
      || ih == NIL || serverID == NIL) {
      DEBUG(OBJECT_DEBUG,
	    (stderr, "ObjectOfSBH: OID parse of %s unsatisfactory.\n",
	     oid));
      ilu_free(oid);
      ilu_free(connectInfo);
      ilu_free(mstid);
      return o;
    }
  if (mstid == NIL)
    mstid = mstid_parm;

  if (mstid != NIL) {
      c2 = ilu_FindClassFromID(mstid);
      if (c2 == NIL)
          c2 = _ilu_FindMSKA(mstid);
      c2 = (tidfound = (c2 != NIL)) ? c2 : static_type;
    }
  else c2 = static_type;
  s = _ilu_FindServer(serverID, TRUE, connectInfo, &err);
  ILU_MUST_BE_SUCCESS(err);
  ASSERT(ih != NIL, buf, (buf, "ObjectOfSBH: ih==NIL, sbh=%s", sbh));
  ASSERT(s != NIL, buf, (buf, "ObjectOfSBH: s==NIL, sid=%s", serverID));
  ilu_EnterServer(s, static_type);
      
  if (s->sr_objs == NIL) {
      DEBUG(CONNECTION_DEBUG,
	    (stderr, "ilu_ObjectOfSBH:  %s (id=%s, true=%d).\n",
	     "asked for object in closed server", serverID,
	     server_is_true(s)));
      ilu_ExitServer(s, static_type);
      ilu_free(oid);
      ilu_free(ih);
      ilu_free(serverID);
      ilu_free(connectInfo);
      return NIL;
    }
      
  DEBUG(CONNECTION_DEBUG,
	    (stderr,
	     "ilu_ObjectOfSBH:  Server for id %s, ci %s is %p.\n",
	     serverID, connectInfo, s));

  ilu_free(connectInfo);
  ilu_free(serverID);
  o = _ilu_FindObjectInServer(ih, s);

  if (o == NIL)
    {
      ilu_Object o2;
      if (server_is_true(s)) {
          ilu_ExitServer(s, static_type);
          ilu_free(ih);
        }
      else if (s->sr_closing == TRUE) {
          ilu_ExitServer(s, static_type);
          DEBUG(OBJECT_DEBUG,
		(stderr, "%s (oid %s) for closing server %s.\n",
		 "ilu_ObjectOfSBH:  refusing to create new surrogate",
		 oid, s->sr_id));
          ilu_free(ih);
        }
      else if (!_ilu_Addable(s, c2, &o2)) {
          ilu_ExitServer(s, static_type);
          DEBUG(OBJECT_DEBUG,
		(stderr,
		 "%s %s of type %s in server %s because %s %s.\n",
		 "ilu_ObjectOfSBH:  won't create new singleton", ih,
		 c2->cl_unique_id, s->sr_id, o2->ob_ih, "already exists"));
          ilu_free(ih);
        }
      else
	{
          o = CreateObject (ih, s, c2, NIL, _ilu_Strdup(sbh), mstid);
          DEBUG(OBJECT_DEBUG,
		    (stderr,
		     "ilu_ObjectOfSBH:  Couldn't find obj for oid <%s> on"
		     " server %p, created new surrogate object %p.\n",
		     oid, s, o));
          if (mstid==NIL || !tidfound) {
              ilu_ExitServer(s, static_type);
              c2 = _ilu_FindClassViaRPC(o);
              if (c2 != NIL) {
                  object_class(o) = c2;
                  ilu_EnterServer(s, static_type);
                }
              else {
                  DEBUG(OBJECT_DEBUG,
			(stderr,
			 "ilu_ObjectOfSBH:  Couldn't determine type for"
			 " object %p, given mstid \"%s\".\n", o, mstid));
                  UncreateObject(o);
                  o = NIL;
                }
            }
          if (o != NIL) {
              if (s->sr_closing == TRUE) {
                  DEBUG(OBJECT_DEBUG,
			(stderr,
			 "ilu_ObjectOfSBH:  %s (oid %s) %s %s.\n",
			 "refusing to create new surrogate",
			 oid, "for closing server", s->sr_id));
                  UncreateObject(o);
                  ilu_ExitServer(s, static_type);
                  o = NIL;
                }
              else if (NIL != (o2 = (ilu_Object) _ilu_hash_FindInTable
						  (server_objs(s), ih) ) )
                {
                  UncreateObject(o);
                  o = o2;
                }
              else if (!_ilu_Addable(s, object_class(o), &o2)) {
                  DEBUG(OBJECT_DEBUG,
			(stderr,
			 "%s %s of type %s in server %s because %s %s.\n",
			 "ilu_ObjectOfSBH:  won't create new singleton",
			 ih, object_class(o)->cl_unique_id, s->sr_id, o2->ob_ih,
			 "already exists"));
                  UncreateObject(o);
                  ilu_ExitServer(s, static_type);
                  o = NIL;
                }
              else {
                  _ilu_Assert(_ilu_hash_AddToTable(server_objs(s), ih, o),
			      "ObjectOfSBH: add to cache (2) failed");
                  _ilu_AddSingleton(s, object_class(o), o);
                }
            }
	}
    }
  else
      ilu_free(ih);
      
  ilu_free(oid);

  DEBUG(OBJECT_DEBUG, (stderr,
		       "ilu_ObjectOfSBH:  Object of <%s> is %p.\n",
		       sbh, o));
  if (o==NIL)
      {}
  else if (o->ob_mstid==NIL && mstid!=NIL)
      o->ob_mstid = _ilu_Strdup(mstid);
  else if (o->ob_mstid==NIL || mstid==NIL)
      {}
  else if (strcmp(o->ob_mstid, mstid) != 0) {
      DEBUG(OBJECT_DEBUG,
	    (stderr,
	     "ObjectOfSBH: Existing object %s has mstid %s, not %s.\n",
	     o->ob_oid, o->ob_mstid, mstid));
    }
  if (o!=NIL && !ilu_IsSubObjectType(o->ob_class, static_type)) {
      DEBUG(OBJECT_DEBUG,
	    (stderr,
	     "%s %s has type %s:%s, not a subtype of %s:%s.\n",
	     "ilu_ObjectOfSBH: Existing object", o->ob_oid,
	     o->ob_class->cl_unique_id, o->ob_class->cl_name,
	     static_type->cl_unique_id, static_type->cl_name));
    }

  return (o);
}

/*L1 >= {obj's server}; L1_sup < prmu*/
ilu_string ilu_SBHOfObject (ilu_Object obj)
{
  ilu_string      ans, pinfo;
  ilu_Server      s;

  if (obj == NIL)
    return (NIL);

  s = object_server(obj);
  _ilu_HoldMutex(server_lock(s));
  if (object_sbh(obj) != NIL) {
  } else if (object_is_true(obj)) {
    if (server_default_port(s) != NIL) {
      ilu_Port        port = server_default_port(s);
      pinfo = protocol_form_handle(port_protocol(port),
				   obj);
      if (pinfo != NIL) {
	object_sbh(obj) = _ilu_Strcat5(object_oid(obj), "@",
				       pinfo, "|",
				       port_tinfo(port));
	ilu_free(pinfo);
      }
    }
  } else
    object_sbh(obj) = _ilu_Strcat5(object_oid(obj), "@",
				   s->sr_pinfo, "|",
				   s->sr_tinfo);
  ans = object_sbh(obj);
  return (ans);
}

/*L1, L2, Main unconstrained*/

ilu_string ilu_MstidOfObject (ilu_Object obj)
{
  return (obj->ob_mstid);
}

ilu_Class ilu_ClassOfObject (ilu_Object obj)
{
  if (obj == NIL)
    return (NIL);
  else
    return (object_class(obj));
}

ilu_boolean ilu_TrueInstanceP (ilu_Object obj)
{
  ilu_boolean s = FALSE;

  if (obj != NIL)
    {
      s = object_is_true(obj);
    }
  return (s);
}

ilu_Server ilu_ServerOfObject (ilu_Object obj)
{
  if (obj == NIL)
    return (NIL);
  else
    return (object_server(obj));
}

ilu_string ilu_IhOfObject (ilu_Object obj)
{
  if (obj == NIL)
    return (NIL);
  else
    return (object_ih(obj));
}

/*L1 >= {cmu, s}*/
ILU_ERRS((BadDataStructure, KernelBroken))
_ilu_DestroyObject(ilu_Server s, ilu_Object obj)
{
  ilu_Class       c = object_class(obj);
  ilu_Error       err;
  ASSERT(obj->ob_oid != NIL, buf,
	 (buf, "object.c:DestroyObject: %p->ob_oid == NIL", obj));
  ASSERT(c != NIL, buf,
	 (buf, "object.c:DestroyObject: class(%s) == NIL",
	  obj->ob_oid));
  ASSERT(obj->ob_ih != NIL, buf,
    (buf, "object.c:DestroyObject: ih(%s) == NIL", obj->ob_oid));
  _ilu_HoldMutex(ilu_cmu);
  _ilu_HoldMutex(server_lock(s));
  if (object_collectible(obj) && server_is_true(s))
    _ilu_StopGCingTrueObj(obj);
  err = _ilu_ServerRemoveObject(s, obj);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE2(BadDataStructure, KernelBroken) return err;
  } ILU_ERR_ENDSWITCH;
  FREETOKEN(object_ih(obj));
  FREETOKEN(object_oid(obj));
  if (object_sbh(obj) != NIL)
    FREETOKEN(object_sbh(obj));
  ilu_free(obj);
  return ILU_NO_ERR;
}

/*L1 >= {s}*/
ilu_refany ilu_GetLanguageSpecificObject (ilu_Object obj)
{
  ilu_refany s = NIL;
  if (obj != NIL)
    {
      _ilu_HoldMutex(server_lock(object_server(obj)));
      s = object_lspo(obj);
    }
  return (s);
}

/*Inside(obj's server, st); L1 no higher than required by that*/
static 
ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed))
IUpdate(ilu_Object obj, ilu_Class st)
{
  ilu_Server      s = object_server(obj);
  ilu_Error       ans = ILU_NO_ERR;
  if (obj->ob_notifying)
    return ILU_NO_ERR;
  if (object_collectible(obj) && !object_is_true(obj)) {
    int             should;
    should = (obj->ob_lspo != NIL) || (obj->ob_holds != 0);
    obj->ob_notifying = TRUE;
    while (should != obj->ob_known && ans == ILU_NO_ERR) {
      ilu_ExitServer(s, st);
      if (should)
	ans = _ilu_RegisterGCInterest(obj);
      else
	ans = _ilu_UnregisterGCInterest(obj);
      ilu_EnterServer(s, st);
      if (ans == ILU_NO_ERR)
	obj->ob_known = should;
      should = (obj->ob_lspo != NIL) || (obj->ob_holds != 0);
    }
    obj->ob_notifying = FALSE;
  }
  if (ans != ILU_NO_ERR)
    return ans;
  if (obj->ob_lspo != NIL || obj->ob_holds != 0)
    return ILU_NO_ERR;
  if (object_is_true(obj) && object_collectible(obj)) {
    if (obj->ob_gclist != NIL && _ilu_vector_size(obj->ob_gclist) != 0)
      return ILU_NO_ERR;
    if (ilu_CoarseTime_Now() < obj->ob_lastRemote + obj->ob_timeout)
      return ILU_NO_ERR;
  }
  DEBUG(OBJECT_DEBUG,
    (stderr, "IUpdate:  deleting uninteresting object %s (%p).\n",
     obj->ob_oid, obj));
  return _ilu_DestroyObject(object_server(obj), obj);
}

/*Inside(obj's server, obj's type)*/

ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed)) _ilu_VIUpdate(ilu_Object obj)
{
  int vi =   (obj->ob_holds != 0)
	  || (object_collectible(obj) &&
	      (object_is_true(obj)
	      ? (   (obj->ob_gclist != NIL &&
	             _ilu_vector_size(obj->ob_gclist) != 0)
	         || (ilu_CoarseTime_Now() <
	             obj->ob_lastRemote + obj->ob_timeout) )
	      : (obj->ob_notifying && !obj->ob_known) ));
  if (vi != obj->ob_intNoted) {
      obj->ob_intNoted = vi;
      if (theNoter != NULLFN)
        (*theNoter)(obj, vi);
    }
  return IUpdate(obj, obj->ob_class);
}

void ilu_RegisterLanguageSpecificObject (ilu_Object obj, ilu_refany lso)
{
  ilu_SetLSO(obj, obj->ob_class, lso);
  return;
}

/*Inside(obj's server, t); L1 no higher than required by that*/
void ilu_SetLSO (ilu_Object obj, ilu_Class t, ilu_refany lso)
{
  ilu_Error err;
  obj->ob_lspo = lso;
  err = IUpdate(obj, t);
  ILU_MUST_BE_SUCCESS(err);
}

void 
ilu_DHolds(ilu_Object obj, ilu_integer dholds)
{
  obj->ob_holds += dholds;
  return;
}

ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed))
ilu_DeltaHolds(ilu_Object obj, ilu_integer dholds)
{
  return _ilu_DeltaHolds(obj, dholds);
}

ILU_ERRS((BadDataStructure, KernelBroken, GcRegFailed))
_ilu_DeltaHolds(ilu_Object obj, ilu_integer dholds)
{
  obj->ob_holds += dholds;
  return _ilu_VIUpdate(obj);
}

/* a bit of scaffolding code for debugging the class record */
/*L1, L2, Main unconstrained; for calling from debugger*/
#ifdef NOPE
static void dump_class_record(ilu_Class class)
{
	printf("dumping class record:\n");
	printf("class name: %s\n",class->cl_name);
	printf("class brand: %s\n",class->cl_brand ? class->cl_brand : "");
	printf("class id: %s\n",class->cl_unique_id ? class->cl_unique_id : "");
	printf("class singleton == \"%s\"",
	       class->cl_singleton ? class->cl_singleton : "(not a singleton)");
	printf("class collectible == 0x%x\n",class->cl_collectible);
	printf("class authentication: %s\n",
	       class->cl_authentication ? class->cl_authentication : "");
	printf("class methods table: 0x%x\n",class->cl_methods);
	printf("class method count: %d\n",class->cl_method_count);
	printf("class superclass record: 0x%x\n",class->superclass);
	printf("class superclass name: 0x%x (%s)\n",class->superclassname,
	       class->superclassname ? class->superclassname : "");
}
#endif

/*L1, L2, Main unconstrained*/

ilu_string _ilu_Strdup (const ilu_string str)
{
  ilu_string p;

  if (str == NIL)
    return (NIL);
  else if ((p = ilu_malloc(strlen(str) + 1)) == NIL)
    return (NIL);
  else
    {
      strcpy(p, str);
      return(p);
    }
}

ilu_string _ilu_Strcat3(const ilu_string s1, const ilu_string s2,
			const ilu_string s3)
{
  int l1 = (s1==NIL) ? 0 : strlen(s1);
  int l2 = (s2==NIL) ? 0 : strlen(s2);
  int l3 = (s3==NIL) ? 0 : strlen(s3);
  ilu_string t = ilu_malloc(1 + l1 + l2 + l3);
  ilu_string u = t;
  if (t==NIL)
    return t;
  if (s1 != NIL) {
    strcpy(u, s1);
    u += l1;
  }
  if (s2 != NIL) {
    strcpy(u, s2);
    u += l2;
  }
  if (s3 != NIL) {
    strcpy(u, s3);
    u += l3;
  }
  *u = 0;
  return t;
}

ilu_string _ilu_Strcat5(const ilu_string s1, const ilu_string s2,
			const ilu_string s3, const ilu_string s4,
			const ilu_string s5)
{
  int l1 = (s1==NIL) ? 0 : strlen(s1);
  int l2 = (s2==NIL) ? 0 : strlen(s2);
  int l3 = (s3==NIL) ? 0 : strlen(s3);
  int l4 = (s4==NIL) ? 0 : strlen(s4);
  int l5 = (s5==NIL) ? 0 : strlen(s5);
  ilu_string t = ilu_malloc(1 + l1 + l2 + l3 + l4 + l5);
  ilu_string u = t;
  if (t==NIL)
    return t;
  if (s1 != NIL) {
    strcpy(u, s1);
    u += l1;
  }
  if (s2 != NIL) {
    strcpy(u, s2);
    u += l2;
  }
  if (s3 != NIL) {
    strcpy(u, s3);
    u += l3;
  }
  if (s4 != NIL) {
    strcpy(u, s4);
    u += l4;
  }
  if (s5 != NIL) {
    strcpy(u, s5);
    u += l5;
  }
  *u = 0;
  return t;
}

/*Main Invariant holds; L2 otherwise unconstrained*/
ilu_boolean ilu_PingObject (ilu_Object o)
{
  ilu_Call call;
  ilu_cardinal estatus = 0;
  ilu_ProtocolException internal;
  ilu_cardinal reqSize;
  ilu_Class pclass = object_class(o);
  ilu_Server s = object_server(o);
  ilu_boolean status;

  if (class_singleton(pclass))
    return (ilu_FALSE);

  DEBUG(OBJECT_DEBUG,
        (stderr, "ilu_PingObject:  object <%s>...\n", object_oid(o)));

  if ((call = ilu_BeginCall(s)) == NIL)
    {
      DEBUG(OBJECT_DEBUG,
	    (stderr, "ilu_PingObject:  ilu_BeginCall failed.\n"));
      return (ilu_FALSE);
    }
  _ilu_AcquireMutex(server_lock(s));
  reqSize = ilu_SizeOfObjectID(call, o, TRUE, _ilu_rootClass);
  _ilu_ReleaseMutex(server_lock(s));
  if (ilu_BeginRequest(call, _ilu_rootClass, _ilu_PingMethod, reqSize))
    {
      ilu_EnterServer(s, pclass);
      ilu_OutputObjectID(call, o, TRUE, _ilu_rootClass);
      ilu_FinishRequest(call);
      internal = ilu_GetReply(call, &estatus);
      status = (internal==ilu_ProtocolException_Success AND estatus==0);
      DEBUG(OBJECT_DEBUG,
	    (stderr, "ilu_PingObject:  returns %u and %lu => %s.\n",
	     internal, estatus, status ? "Good object" : "Bad object"));
    }
  else
    {
      DEBUG(OBJECT_DEBUG,
	    (stderr, "ilu_PingObject:  ilu_BeginRequest failed.\n"));
      status = ilu_FALSE;
    }
  ilu_FinishCall(call);
  return status;
}

/*L2    >=    {conn's callmu, iomu} before,
  L2 disjoint {conn's callmu, iomu} after*/
void 
_ilu_HandlePing(ilu_Call call)
{
  ilu_Object      disc;
  ilu_Error       err;

  ilu_InputObjectID(call, &disc, TRUE, _ilu_rootClass);
  if (disc != NIL) {
    err = ilu_DeltaHolds(disc, 1);
    ILU_MUST_BE_SUCCESS(err);
    ilu_ExitServer(object_server(disc), _ilu_rootClass);
  }
  ilu_RequestRead(call);

  if ((disc != NIL) AND ilu_TrueInstanceP(disc)) {
    if (ilu_BeginReply(call, FALSE, 0))
      ilu_FinishReply(call);
  } else {
    if (ilu_BeginException(call, 0,
			   ilu_ProtocolException_GarbageArguments))
      ilu_FinishException(call);
  }

  if (disc != NIL) {
    ilu_Server      s = object_server(disc);
    ilu_Class       cl = object_class(disc);
    ilu_EnterServer(s, cl);
    err = ilu_DeltaHolds(disc, -1);
    ILU_MUST_BE_SUCCESS(err);
    ilu_ExitServer(s, cl);
  }
  return;
}
