/*
Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: sunrpc.c,v 1.112 1995/03/02 23:50:16 janssen Exp $ */
/* Last edited by Mike Spreitzer December 6, 1994 3:41 pm PST */

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluntrnl.h"
#include "packet.h"
#include "sunrpc.h"

#include "call.h"
#include "protocol.h"
#include "connect.h"
#include "transprt.h"
#include "port.h"
#include "object.h"
#include "type.h"
#include "method.h"
#include "mooring.h"
#include "authschm.h"

#ifdef USE_SUNRPC_UNIX_AUTHENTICATION

#include "iluaunix.h"

static ilu_AuthenticationScheme SunRPCUnixAuthenticationScheme = NIL;

#endif /* def USE_SUNRPC_UNIX_AUTHENTICATION */

#define ILU_SUNRPC_PROGRAM_NUMBER	0x31000400

/*************************************************************/
/*		Implement XDR PACKET			     */
/*************************************************************/

/*L1, Main unconstrained*/
/*L2 >= call's conn's iomu*/

#define ODD(x)		(((x)&0x1)!=0)
#define EVEN(x)		(((x)&0x1)==0)  
#define PADDED_SIZE(x)	((((unsigned) (x))+3) & (~0x3))
  
static unsigned long endianNessTest = 0x01020304;
#ifdef BIG_ENDIAN
#undef BIG_ENDIAN
#endif
#define BIG_ENDIAN	((((unsigned char *)(&endianNessTest))[0])==0x01)

static ilu_boolean _xdr_get_u_short (PACKET p, ilu_shortcardinal *l)
{
  _ilu_Assert (0, "Attempt to input 2 byte short in Sun RPC!");
  return (ilu_FALSE);
}

static ilu_boolean _xdr_get_u_long (PACKET p, ilu_cardinal *l)
{
  if (packet_allocation(p) - packet_size(p) < 4)
    return(ilu_FALSE);
  *l = (p->next[0] << 24) + (p->next[1] << 16) + (p->next[2] << 8) + p->next[3];
  p->next += 4;
  return (ilu_TRUE);
}

static ilu_boolean _xdr_get_long (PACKET p, ilu_integer *l)
{
  return (_xdr_get_u_long (p, (ilu_cardinal *) l));
}

static ilu_boolean _xdr_get_short (PACKET p, short *l)
{
  return (_xdr_get_u_short(p, (ilu_shortcardinal *) l));
}

static ilu_boolean _xdr_get_byte (PACKET p, char *l)
{
  _ilu_Assert (0, "Attempt to input 1 byte octet in Sun RPC!");
  return (ilu_FALSE);
}

static ilu_boolean _xdr_get_u_byte (PACKET p, ilu_byte *l)
{
  _ilu_Assert (0, "Attempt to input 1 byte octet in Sun RPC!");
  return (ilu_FALSE);
}

static ilu_boolean _xdr_get_float(PACKET p, float *l)
{
  return (_xdr_get_u_long (p, (ilu_cardinal *) l));
}

static ilu_boolean _xdr_get_double (PACKET p, double *l)
{
  return (_xdr_get_u_long (p, ((ilu_cardinal *) l) + (BIG_ENDIAN ? 0 : 1))
	  && _xdr_get_u_long(p, ((ilu_cardinal *) l) + (BIG_ENDIAN ? 1 : 0)));
}

static ilu_cardinal SunRPCMaxStringSize = 100000;

static ilu_boolean _xdr_get_bytes(PACKET p, ilu_byte **b, ilu_cardinal *l, ilu_cardinal limit)
{
  ilu_cardinal l2;
  register ilu_byte *p1, *p2;
  register ilu_cardinal c;

  if (packet_allocation(p) - packet_size(p) < 4
      || (!_xdr_get_u_long (p, &l2))
      || packet_allocation(p) - packet_size(p) < PADDED_SIZE(l2))
    return (ilu_FALSE);

  *l = 0;
  if (l2 > SunRPCMaxStringSize OR (limit > 0 AND l2 > limit))
    {
      fprintf (stderr,
	       "%s %lu, which exceeds SunRPCMaxStringSize value of %lu or call limit of %lu.\n",
	       "Attempt to read string of length", l2,
	       SunRPCMaxStringSize, limit);
      return (ilu_FALSE);
    }

  {
    register ilu_integer i;

    if (*b == NIL)
      *b = (ilu_byte *) ilu_malloc(l2 + 1);
    for (p1 = *b, p2 = p->next, c = l2;  c > 0;  c--)
      *p1++ = *p2++;
    for (i = PADDED_SIZE(l2) - l2;  i > 0;  i--)
      p2++;	/* eat padding characters */
    *p1 = '\0';
    p->next = p2;
    *l = l2;
    return (ilu_TRUE);
  }
}

static ilu_boolean _xdr_get_opaque (PACKET p, ilu_byte **b, ilu_cardinal len)
{
  register ilu_byte *p1, *p2;
  register ilu_cardinal c;
  register ilu_integer i;

  if (packet_allocation(p) - packet_size(p) < PADDED_SIZE(len))
    return (ilu_FALSE);

  if (*b == NIL)
    *b = (ilu_byte *) ilu_malloc(len);
  for (p1 = *b, p2 = p->next, c = len;  c > 0;  c--)
    *p1++ = *p2++;
  for (i = PADDED_SIZE(len) - len;  i > 0;  i--)
    p2++;	/* eat padding characters */
  p->next = p2;
    return (ilu_TRUE);
}

static void _xdr_destroy (PACKET p)
{
  if (p->buf != NIL && p->count > 0)
    ilu_free(p->buf);
  ilu_free(p);
}

static PACKET _xdr_CreatePacket (ilu_bytes buf, ilu_cardinal size)
{
  PACKET p = (PACKET) ilu_malloc (sizeof(struct ilu_packet_s));
  if (buf != NIL)
    p->buf = buf;
  else if (size > 0)
    p->buf = (ilu_byte *) ilu_malloc(size);
  else
    p->buf = NIL;
  p->count = size;
  p->next = p->buf;

  p->put_long = NULLFN;
  p->put_byte = NULLFN;
  p->put_short = NULLFN;
  p->put_u_long = NULLFN;
  p->put_u_byte = NULLFN;
  p->put_u_short = NULLFN;
  p->put_float = NULLFN;
  p->put_double = NULLFN;
  p->put_bytes = NULLFN;
  p->put_opaque = NULLFN;

  p->get_long = _xdr_get_long;
  p->get_byte = _xdr_get_byte;
  p->get_short = _xdr_get_short;
  p->get_u_long = _xdr_get_u_long;
  p->get_u_byte = _xdr_get_u_byte;
  p->get_u_short = _xdr_get_u_short;
  p->get_float = _xdr_get_float;
  p->get_double = _xdr_get_double;
  p->get_bytes = _xdr_get_bytes;
  p->get_opaque = _xdr_get_opaque;

  p->destroy = _xdr_destroy;

  return (p);
}

/**********************************************************************
  End of packet implementation
***********************************************************************/

/*L1, L2, Main unconstrained*/

static ilu_refany _sunrpc_CreateDataBlock (void)
{
  struct SunRPC *new = (struct SunRPC *) ilu_malloc(sizeof(struct SunRPC));

  new->sun_in = NIL;
  return ((ilu_refany) new);
}

static void _sunrpc_FreeDataBlock (struct SunRPC *d)
{
  if (d->sun_in != NIL)
    {
      packet_destroy (d->sun_in);
      d->sun_in = NIL;
    }
  ilu_free(d);
}

/**********************************************************************
  Code to figure Sun RPC program # given ILU type code
***********************************************************************/

/*
 * Calculate the CRC32 value of a given buffer
 */
static unsigned long CRC32_of_buffer (register unsigned char *bufp, register int len)
{

#if 0
/*
 * Generate a FCS table for the HDLC FCS.
 *
 * after:  Karl Fox at Morning Star Technologies,
 * who borrowed from Drew D. Perkins at Carnegie Mellon University,
 * who liberally borrowed from Mohsen Banan and D. Hugh Redelmeier.
 *
 *==================================================
 *
 * The 32-bit HDLC generator polynomial:
 *
 *	x**32 + x**26 + x**23 + x**22 + x**16 + x**12 + x**11 + x**10 +
 *	x**8 + x**7 + x**5 + x**4 + x**2 + x**1 + x**0
 *
 *	(1110 1101 1011 1000 1000 0011 0010 0000)
 *	(0xEDB88320)
 *
 * The 32-bit HDLC remainder polynomial:
 *
 *	x**31 + x**30 + x**26 + x**25 + x**24 + x**18 + x**15 + x**14 +
 *	x**12 + x**11 + x**10 + x**8 + x**6 + x**5 + x**4 + x**3 + x**2 +
 *	x**1 + x**0
 *
 *	(1111 1110 1011 1011 0010 0000 1110 0011)
 *	(0xFEBB20E3)
 *
 */
# define P      0xEDB88320

main() {
  register unsigned int b, v;
  register int i;

  for (b = 0; ; )
    {
      if (b % 4 == 0)
	printf("\n");
      v = b;
      for (i = 8; i--; )
	v = v & 1 ? (v >> 1) ^ P : v >> 1;
      printf("\t0x%08x", v & 0xFFFFFFFF);
      if (++b == 256)
	break;
      printf(",");
    }
}
#endif /* 0 */

  static unsigned long fcstab[] = {
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
    0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
    0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
    0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
    0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
    0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
    0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
    0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
    0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
    0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
    0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
    0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
    0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
    0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
    0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
    0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
    0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
    0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
    0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
    0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
    0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
    0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
    0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
    0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
    0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
    0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
    0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
    0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
    0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
    0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
    0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
    0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
    0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
    0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
    0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
    0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
    0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
    0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
    0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
    0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
    0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
    0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
    0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
    0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
    0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
    0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
    0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
    0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
    0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
    0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
    0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
    0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
    0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
    0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
    0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
    0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
    0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
    0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
    0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
    0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
    0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
    0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
    };
  register unsigned long fcs = 0xFFFFFFFF;
  while (len--)
    fcs = (((fcs) >> 8) ^ fcstab[((fcs) ^ (*bufp++)) & 0xFF]);
  return (fcs);
}

/*L1 >= {prmu}*/

static HashTable ProgramNumberHashTable = NIL;
/* program number & version -> sunrpcinfo* */

static HashTable RegistryHashTable = NIL;
/* type unique_id -> sunrpcinfo* */

static sunrpcinfo *_sunrpc_AddClassInformation(ilu_Class);

static ilu_cardinal _sunrpc_HashSunRPC (sunrpcinfo *key, ilu_cardinal size)
{
  return ((key->sui_pnumber + key->sui_version) % size);
}

static ilu_boolean _sunrpc_CompareSunRPC (sunrpcinfo *key1, sunrpcinfo *key2)
{
  return ((key1->sui_pnumber == key2->sui_pnumber) AND (key1->sui_version == key2->sui_version));
}

typedef ilu_cardinal (*ILU_hash_proc)(ilu_refany, ilu_cardinal);
typedef ilu_boolean (*ILU_find_proc)(ilu_refany, ilu_refany);

static void _sunrpc_EnsureRegistries(void)
{
  if (RegistryHashTable == NIL)
    {
      RegistryHashTable = _ilu_hash_MakeNewTable (137, NULLFN, NULLFN);
      ProgramNumberHashTable =
	_ilu_hash_MakeNewTable (137,
				(ILU_hash_proc)_sunrpc_HashSunRPC,
				(ILU_find_proc)_sunrpc_CompareSunRPC);
      _sunrpc_AddClassInformation (_ilu_rootClass);
      _sunrpc_AddClassInformation (_ilu_GcCallbackClass);
    }
}

static sunrpcinfo * _sunrpc_AddClassInformation (ilu_Class class)
{
  char *next = NIL;
  sunrpcinfo *s = (sunrpcinfo *) ilu_malloc(sizeof(sunrpcinfo));

  _sunrpc_EnsureRegistries();

  s->sui_class = class;
  s->sui_type_id = class->cl_unique_id;
  if (class_singleton(class))
    {
      if (((unsigned long) (class->cl_singleton)) == 1)
	{
	  fprintf (stderr, "\
Fatal ILU error:  Stubs for type \"%s\" were generated\n\
by a pre-1.6.4-p8 stubber.  Please re-stub, re-compile, and re-link.\n",
		   class->cl_name);
	  exit(1);
	}
      if (NOT (strncmp(class->cl_singleton, "sunrpc_2_", 9) == 0 AND
	       (s->sui_pnumber = _ilu_atoi(class->cl_singleton + 9, &next),
		(s->sui_pnumber > 0 && (next != NIL) && (*next == '_'))) AND
	       (s->sui_version = _ilu_atoi(next + 1, NIL),
		(s->sui_version > 0))))
	{
	  fprintf (stderr, "(ILU:sunrpc)  Couldn't determine Sun RPC program number for ILU class \"%s\", given native Sun RPC info \"%s\".\n",
		   class->cl_name, class->cl_singleton);
	  return (NIL);
	}
    }
  else
    {
      s->sui_pnumber = ILU_SUNRPC_PROGRAM_NUMBER;
      s->sui_version = CRC32_of_buffer ((unsigned char *) class->cl_unique_id,
					strlen(class->cl_unique_id));
    }
  _ilu_Assert(_ilu_hash_AddToTable(RegistryHashTable,
				   s->sui_type_id, s),
	      "SunRPC AddToTable RegistryHashTable");
  _ilu_Assert(_ilu_hash_AddToTable(ProgramNumberHashTable, s, s),
	      "SunRPC AddToTable ProgramNumberHashTable");
  return (s);
}

static sunrpcinfo *_sunrpc_SunRPCInformationForClass (ilu_Class class)
{
  sunrpcinfo *s = NIL;

  _sunrpc_EnsureRegistries();

  s = (sunrpcinfo *) _ilu_hash_FindInTable (RegistryHashTable,
					    class->cl_unique_id);
  if (s == NIL)
    s = _sunrpc_AddClassInformation (class);

  DEBUG(SUNRPC_DEBUG,
	(stderr, "%s \"%s:%s\", pnumber is 0x%lx, version is %lu.\n",
	 "_sunrpc_SunRPCInformationForClass:  Class",
	 class->cl_name, class->cl_unique_id,
	 (s == NIL) ? 0 : s->sui_pnumber,
	 (s == NIL) ? 0 : s->sui_version));

  return (s);
}

static void AddClass (ilu_Class c, ilu_refany junk)
{
  _sunrpc_SunRPCInformationForClass (c);
}

static sunrpcinfo *_sunrpc_ClassFromProgramNumber (ilu_cardinal pnumber,
						   ilu_cardinal version,
						   ilu_boolean try_adds)
{
  sunrpcinfo dummy;
  sunrpcinfo *s = NIL;

  _sunrpc_EnsureRegistries();

  dummy.sui_pnumber = pnumber;
  dummy.sui_version = version;
  s = (sunrpcinfo *) _ilu_hash_FindInTable (ProgramNumberHashTable,
					      (ilu_refany) &dummy);
  if (s == NIL)
    {
      if (try_adds)
	{
	  _ilu_EnumerateClasses (AddClass, NIL);
	  return _sunrpc_ClassFromProgramNumber (pnumber, version, ilu_FALSE);
	}
      else
	{
	  fprintf (stderr, "(ILU:_sunrpc_ClassFromProgramNumber):  Couldn't find class for program number 0x%lx, version %lu\n",
		   pnumber, version);
	}
    }
  else
    {
      DEBUG(SUNRPC_DEBUG,
	    (stderr, "%s \"%s:%s\", pnumber is 0x%lx, version is %lu.\n",
	     "_sunrpc_ClassFromProgramNumber:  Class",
	     s->sui_class->cl_name, s->sui_class->cl_unique_id,
	     s->sui_pnumber, s->sui_version));
    }

  return (s);
}

/*======================================================================*/
/*======================== Basic I/O code ==============================*/
/*======================================================================*/

#define INPUT_ERROR		1
#define OUTPUT_ERROR		2

/*L1, L2, Main unconstrained (this is only for calling from debugger)*/
static ilu_cardinal ilu_sunrpc_SetMaxStringSize (ilu_cardinal size)
{
  ilu_cardinal old_size = SunRPCMaxStringSize;
  if (size > 0)
    SunRPCMaxStringSize = size;
  return (old_size);  
}

/*L2 >= {call's connection's iomu}*/
/*L1, Main unconstrained*/

/* ==================== cardinal ==================== */

static ilu_boolean _sunrpc_OutputCardinal (ilu_Call call, ilu_cardinal l)
{
  register unsigned char * buf;
  ilu_Error err = NIL;

  buf = transport_get_output_buffer(connection_transport(call_connection(call)), 4, &err);
  if (err != NIL)
    {
      fprintf (stderr, "(ILU)_sunrpc_OutputCardinal:  Error \"%s: %s\" signalled on attempt to fetch 4 bytes of buffer.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
      ilu_FreeError(err);
    }
  else
    {
      *buf++ = (l >> 24) & 0xFF;
      *buf++ = (l >> 16) & 0xFF;
      *buf++ = (l >> 8) & 0xFF;
      *buf++ = l & 0xFF;
      return (ilu_TRUE);
    }
  return (err == NIL);
}

static ilu_boolean _sunrpc_InputCardinal (ilu_Call call, ilu_cardinal *i)
{
  return (packet_get_u_long (SUNRPC(call_connection(call))->sun_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfCardinal (ilu_Call call, ilu_cardinal i)
{
  return (4);
}

/* ==================== integer ==================== */

static ilu_boolean _sunrpc_OutputInteger (ilu_Call call, ilu_integer i)
{
  return _sunrpc_OutputCardinal(call, (ilu_cardinal) i);
}

static ilu_boolean _sunrpc_InputInteger (ilu_Call call, ilu_integer *i)
{
  return (packet_get_long (SUNRPC(call_connection(call))->sun_in, i));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfInteger (ilu_Call call, ilu_integer i)
{
  return (4);
}

/* ==================== short integer ==================== */

static ilu_boolean _sunrpc_OutputShortInteger (ilu_Call call, ilu_shortinteger i)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) i));
}

static ilu_boolean _sunrpc_InputShortInteger (ilu_Call call, ilu_shortinteger *i)
{
  ilu_integer l;
  ilu_boolean stat;

  stat = packet_get_long (SUNRPC(call_connection(call))->sun_in, &l);
  if (stat)
    *i = l;
  return (stat);
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfShortInteger (ilu_Call call, ilu_shortinteger i)
{
  return (4);
}

/* ==================== long integer ==================== */

static ilu_boolean _sunrpc_OutputLongInteger (ilu_Call call, ilu_longinteger i)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) i.high) &&
	  _sunrpc_OutputCardinal(call, (ilu_cardinal) i.low));
}

static ilu_boolean _sunrpc_InputLongInteger (ilu_Call call, ilu_longinteger *i)
{
  ilu_boolean stat = ilu_TRUE;
  PACKET pkt = SUNRPC(call_connection(call))->sun_in;
  stat = (packet_get_long (pkt, &i->high));
  return (stat && (packet_get_u_long (pkt, &i->low)));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfLongInteger (ilu_Call call, ilu_longinteger i)
{
  return (8);
}

/* ==================== short cardinal ==================== */

static ilu_boolean _sunrpc_OutputShortCardinal (ilu_Call call, ilu_shortcardinal i)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) i));
}

static ilu_boolean _sunrpc_InputShortCardinal (ilu_Call call, ilu_shortcardinal *i)
{
  ilu_cardinal l;
  ilu_boolean stat;

  stat = packet_get_u_long (SUNRPC(call_connection(call))->sun_in, &l);
  if (stat)
    *i = l & 0xFFFF;
  return (stat);  
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfShortCardinal (ilu_Call call, ilu_shortcardinal i)
{
  return (4);
}

/* ==================== long cardinal ==================== */

static ilu_boolean _sunrpc_OutputLongCardinal (ilu_Call call, ilu_longcardinal i)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) i.high) &&
	  _sunrpc_OutputCardinal(call, (ilu_cardinal) i.low));
}

static ilu_boolean _sunrpc_InputLongCardinal (ilu_Call call, ilu_longcardinal *i)
{
  ilu_boolean stat;
  PACKET pkt = SUNRPC(call_connection(call))->sun_in;

  stat = (packet_get_u_long (pkt, &i->high));
  return (stat && (packet_get_u_long (pkt, &i->low)));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfLongCardinal (ilu_Call call, ilu_longcardinal i)
{
  return (8);
}

/* ==================== enumeration ==================== */

static ilu_boolean _sunrpc_OutputEnumeration (ilu_Call call, ilu_shortcardinal i)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) i));
}

static ilu_boolean _sunrpc_InputEnumeration (ilu_Call call, ilu_shortcardinal *i)
{
  ilu_cardinal i2;
  ilu_cardinal stat;

  stat = packet_get_u_long (SUNRPC(call_connection(call))->sun_in, &i2);
  *i = i2;
  return(stat);
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfEnumeration (ilu_Call call, ilu_shortcardinal i)
{
  return (4);
}

/* ==================== real ==================== */

static ilu_boolean _sunrpc_OutputReal (ilu_Call call, double d)
{
  double l2 = d;

  return (_sunrpc_OutputCardinal(call, ((ilu_cardinal *) &l2)[BIG_ENDIAN ? 0 : 1]) &&
	  _sunrpc_OutputCardinal(call, ((ilu_cardinal *) &l2)[BIG_ENDIAN ? 1 : 0]));
}

static ilu_boolean _sunrpc_InputReal (ilu_Call call, double *d)
{
  return (packet_get_double (SUNRPC(call_connection(call))->sun_in, d));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfReal (ilu_Call call, double d)
{
  return (8);
}

/* ==================== long real ==================== */

static ilu_boolean _sunrpc_OutputLongReal (ilu_Call call, ilu_longreal d)
{
  register unsigned char * buf;
  ilu_Error err = NIL;

  buf = transport_get_output_buffer(connection_transport(call_connection(call)), 16, &err);
  if (err != NIL)
    {
      fprintf (stderr, "(ILU)_sunrpc_OutputLongReal:  Error \"%s: %s\" signalled on attempt to fetch 16 bytes of buffer.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
      ilu_FreeError(err);
      return (ilu_FALSE);
    }
  else
    {
      register int index;
      for (index = 0;  index < 16;  index++)
	*buf++ = d.b[index];
      return (ilu_TRUE);
    }
}

static ilu_boolean _sunrpc_InputLongReal (ilu_Call call, ilu_longreal *d)
{
  unsigned char *ref = d->b;
  return(packet_get_opaque (SUNRPC(call_connection(call))->sun_in, &ref, 16));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfLongReal (ilu_Call call, ilu_longreal d)
{
  return (16);
}

/* ==================== short real ==================== */

static ilu_boolean _sunrpc_OutputShortReal (ilu_Call call, float f)
{
  float f2;

  f2 = f;
  return (_sunrpc_OutputCardinal(call, *((ilu_cardinal *) &f2)));
}

static ilu_boolean _sunrpc_InputShortReal (ilu_Call call, float *f)
{
  return (packet_get_float (SUNRPC(call_connection(call))->sun_in, f));
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfShortReal (ilu_Call call, float d)
{
  return (4);
}

/* ==================== bytes ==================== */

static ilu_boolean _sunrpc_OutputBytes (ilu_Call call, ilu_bytes s, ilu_cardinal len, ilu_cardinal limit)
{
  ilu_Error err = NIL;
  ilu_cardinal size;

  if (limit > 0)
    size = MIN(len, limit);
  else
    size = len;

  if (_sunrpc_OutputCardinal(call, size))
    {
      err = transport_write_bytes(call_transport(call), s, PADDED_SIZE(size));
      if (err != NIL)
	{
	  fprintf (stderr, "(ILU)_sunrpc_OutputBytes:  Error \"%s: %s\" on output of string %p, len %lu, limit %lu.\n",
		   ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err), s, len, limit);
	  ilu_FreeError(err);
	}
      return (err == NIL);
    }
  else
    return (ilu_FALSE);
}

static ilu_boolean _sunrpc_InputBytes (ilu_Call call, ilu_bytes *s, ilu_cardinal *len, ilu_cardinal limit)
{
  *s = NIL;
  return (packet_get_bytes (SUNRPC(call_connection(call))->sun_in, s, len, limit));
}

  /*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfBytes (ilu_Call call, ilu_bytes i, ilu_cardinal l, ilu_cardinal limit)
{
  if (((limit > 0) && (l > limit)) || l > SunRPCMaxStringSize)
    {
      fprintf (stderr, "_sunrpc_SizeOfBytes:  Attempt to pass bytestring of length %lu, which exceeds either SunRPCMaxStringSize of %lu, or limit on this sequence type of %lu.\n",
	       l, SunRPCMaxStringSize, limit);
      return (0);
    }
  return (4 + PADDED_SIZE(l));
}

/* ==================== string ==================== */

static ilu_boolean _sunrpc_OutputString (ilu_Call call, ilu_string s, ilu_cardinal len, ilu_cardinal limit)
{
  return (_sunrpc_OutputBytes(call, (ilu_bytes) s, len, limit));
}

static ilu_boolean _sunrpc_InputString (ilu_Call call, ilu_string *s, ilu_cardinal *len, ilu_cardinal limit)
{
  *s = NIL;
  if (!packet_get_bytes (SUNRPC(call_connection(call))->sun_in,
	(ilu_byte **) s, len, limit))
    return (ilu_FALSE);
  return (ilu_TRUE);
}

  /*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfString (ilu_Call call, ilu_string i, ilu_cardinal l, ilu_cardinal limit)
{
  if (l > SunRPCMaxStringSize)
    {
      fprintf (stderr, "_sunrpc_SizeOfString:  Attempt to pass string of length %lu, which exceeds either SunRPCMaxStringSize value of %lu, or limit on this sequence type of %lu.\n",
	       l, SunRPCMaxStringSize, limit);
      return (0);
    }

  return (4 + PADDED_SIZE(l));
}

/* ==================== byte ==================== */

static ilu_boolean _sunrpc_OutputByte (ilu_Call call, ilu_byte b)
{
  return (_sunrpc_OutputCardinal(call, (ilu_cardinal) b));
}

static ilu_boolean _sunrpc_InputByte (ilu_Call call, ilu_byte *b)
{
  ilu_cardinal l;
  ilu_boolean stat;

  stat = packet_get_u_long (SUNRPC(call_connection(call))->sun_in, &l);
  if (stat)
    *b = l & 0xFF;
  return (stat);
}

/*ARGSUSED*/
static ilu_cardinal _sunrpc_SizeOfByte (ilu_Call call, ilu_byte i)
{
  return (4);
}

/* ==================== opaque ==================== */

static ilu_boolean _sunrpc_OutputOpaque (ilu_Call call, ilu_bytes o, ilu_cardinal len)
{
  ilu_Error err;

  err = transport_write_bytes(call_transport(call), o, PADDED_SIZE(len));
  if (err != NIL)
    {
      fprintf (stderr, "(ILU)_sunrpc_OutputBytes:  Error \"%s: %s\" on output of string %p, len %lu.\n",
	       ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err), o, len);
      ilu_FreeError(err);
    }
  return (err == NIL);
}

static ilu_boolean _sunrpc_InputOpaque (ilu_Call call, ilu_bytes *o, ilu_cardinal len)
{
  return(packet_get_opaque (SUNRPC(call_connection(call))->sun_in, o, len));
}

static ilu_cardinal _sunrpc_SizeOfOpaque (ilu_Call call, ilu_bytes o, ilu_cardinal len)
{
  return (PADDED_SIZE(len));
}

static ilu_boolean _sunrpc_OutputSequence (ilu_Call c, ilu_cardinal sequenceLength, ilu_cardinal limit)
{
  return (_sunrpc_OutputCardinal (c, sequenceLength));
}

static ilu_boolean _sunrpc_OutputSequenceMark (ilu_Call c, ilu_cardinal extent)
{
  return (ilu_TRUE);
}

static ilu_boolean _sunrpc_InputSequenceMark (ilu_Call c, ilu_cardinal extent)
{
  return (ilu_TRUE);
}

static ilu_boolean _sunrpc_InputSequence (ilu_Call c, ilu_cardinal *sequenceLength, ilu_cardinal limit)
{
  ilu_boolean stat;
  ilu_cardinal len;

  stat = _sunrpc_InputCardinal (c, &len);
  *sequenceLength = len;
  return(stat);
}

static ilu_boolean _sunrpc_EndSequence (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _sunrpc_SizeOfSequence (ilu_Call c, ilu_cardinal length, ilu_cardinal limit)
{
  return (_sunrpc_SizeOfCardinal(c, length));
}

static ilu_boolean _sunrpc_OutputUnion (ilu_Call c, ilu_shortcardinal typeIndex)
{
  return (_sunrpc_OutputCardinal (c, (ilu_cardinal) typeIndex));
}

static ilu_boolean _sunrpc_InputUnion (ilu_Call c, ilu_shortcardinal *typeIndex)
{
  ilu_boolean stat;
  ilu_cardinal i;

  stat = _sunrpc_InputCardinal (c, &i);
  *typeIndex = i;
  return(stat);
}

static ilu_boolean _sunrpc_EndUnion (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _sunrpc_SizeOfUnion (ilu_Call c, ilu_shortcardinal typeIndex)
{
  return (_sunrpc_SizeOfCardinal(c, typeIndex));
}

static ilu_boolean _sunrpc_OutputArray (ilu_Call c, ilu_cardinal len)
{
  return(ilu_TRUE);
}

static ilu_boolean _sunrpc_InputArray (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _sunrpc_EndArray (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _sunrpc_SizeOfArray (ilu_Call c, ilu_cardinal len)
{
  return(0);
}

static ilu_boolean _sunrpc_OutputRecord (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _sunrpc_InputRecord (ilu_Call c)
{
  return(ilu_TRUE);
}

static ilu_boolean _sunrpc_EndRecord (ilu_Call c)
{
  return (ilu_TRUE);
}

static ilu_cardinal _sunrpc_SizeOfRecord (ilu_Call c)
{
  return(0);
}

/*======================================================================
**======================================================================
**====================  Non-I/O code ===================================
**======================================================================
**====================================================================*/

/*L2, Main unconstrained*/
/*L1_sup < prmu*/
static ilu_string _sunrpc_FormProtocolHandle (ilu_Object obj)
{
  char buf[100];
  sunrpcinfo *s = NIL;
  ilu_Class class = object_class(obj);

  _ilu_AcquireMutex(ilu_prmu);
  s = _sunrpc_SunRPCInformationForClass (class);
  if (s == NIL)
    {
      _ilu_ReleaseMutex(ilu_prmu);
      fprintf (stderr, "%s program#/version for class %s.\n",
	       "_sunrpc_FormProtocolHandle:  Can't figure", class->cl_name);
      return (NIL);
    }
  else
    {
      sprintf (buf, "sunrpc_2_0x%lx_%lu", s->sui_pnumber, s->sui_version);
      _ilu_ReleaseMutex(ilu_prmu);
      return (_ilu_Strdup(buf));
    }
}
     
/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static ilu_bytes _sunrpc_ReadPacket (ilu_Transport bs,
			      ilu_cardinal *packetLength,
			      ilu_PacketType *type, ilu_cardinal *sn)
{
  ilu_bytes packet;
  ilu_cardinal serialNumber, packetType, size;

  if (!transport_read_message(bs, &packet, &size)) {
    DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	  (stderr, "%s:  error on transport_read_message\n",
	   "_sunrpc_ReadPacket"));
    return (NIL);
  };

  serialNumber =  (packet[0] << 24) | (packet[1] << 16)
		| (packet[2] <<  8) |  packet[3];
  packetType =    (packet[4] << 24) | (packet[5] << 16)
		| (packet[6] <<  8) |  packet[7];

  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr, "%s %lu, SN %lu, type %lu (%s).\n",
	 "_sunrpc_ReadPacket:  reading packet of length", size,
	 serialNumber, packetType,
	 ((packetType == ilu_PacketType_Request) ? "request"
	  : (packetType == ilu_PacketType_Reply) ? "reply"
	  : "unknown type")));

  *packetLength = size;
  *type = packetType;
  *sn = serialNumber;
  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr, "_sunrpc_ReadPacket:  returning packet SN %lu\n",
	 serialNumber));
  return (packet);  
}

static ilu_boolean _sunrpc_InterpretPipeRequest (ilu_Call call, ilu_bytes packet,
					  ilu_cardinal len)
{
  struct SunRPC *c = SUNRPC(call_connection(call));
  ilu_cardinal rpcVersion, credSize, verifierSize, programNumber;
  ilu_cardinal programVersion, methodID;
  ilu_cardinal verifierType, credType, serialNumber, packetType;
  ilu_bytes credBuf, verifierBuf;

  c->sun_in = _xdr_CreatePacket (packet, len);

  _sunrpc_InputCardinal (call, &serialNumber);
  _sunrpc_InputCardinal (call, &packetType);
  _sunrpc_InputCardinal (call, &rpcVersion);
  _sunrpc_InputCardinal (call, &programNumber);
  _sunrpc_InputCardinal (call, &programVersion);
  _sunrpc_InputCardinal (call, &methodID);
  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr,
	 "%s:  (call SN %lu) prognum 0x%lx version %lu, method ID = %lu\n",
	 "_sunrpc_InterpretPipeRequest", call->ca_SN,
	 programNumber, programVersion, methodID));

  _sunrpc_InputCardinal (call, &credType);
  _sunrpc_InputBytes (call, &credBuf, &credSize, 0);
  _sunrpc_InputCardinal (call, &verifierType);
  _sunrpc_InputBytes (call, &verifierBuf, &verifierSize, 0);
  DEBUG(AUTHENTICATION_DEBUG,
	(stderr, "_sunrpc_InterpretPipeRequest:  (call SN %lu) credBuf is %p, credSize is %lu, verifierBuf is %p, verifierSize is %lu\n",
	 call->ca_SN, credBuf, credSize, verifierBuf, verifierSize));
  
  /* not used at the moment */
  FREETOKEN(credBuf);
  FREETOKEN(verifierBuf);
  return (ilu_TRUE);
}

/*L1_sup < prmu*/
static ilu_Passport ReadCredentialsAndVerifier (ilu_Call call)
{
  ilu_bytes credBuf = NIL, verifierBuf = NIL;
  ilu_cardinal credSize, credType, verifierSize, verifierType, stamp;
  ilu_Passport passport = NIL;

  _sunrpc_InputCardinal (call, &credType);
  if (credType == 0)	/* no authentication info */
    {
      _sunrpc_InputBytes (call, &credBuf, &credSize, 0);
      _sunrpc_InputCardinal (call, &verifierType);
      _sunrpc_InputBytes (call, &verifierBuf, &verifierSize, 0);
    }

#ifdef USE_SUNRPC_UNIX_AUTHENTICATION

  else if (credType == 1)	/* Sun RPC UNIX authentication */
    {
      struct _ilu_AS_SunRPCUnix_UserID_s uid;
      ilu_Credential cred = NIL;
      ilu_cardinal gid;
      ilu_Error err;

      if (SunRPCUnixAuthenticationScheme == NIL)
	{
	  _ilu_AcquireMutex(ilu_prmu);
	  if (SunRPCUnixAuthenticationScheme == NIL)
	    {
	      SunRPCUnixAuthenticationScheme = ilu_FindAuthenticationSchemeByName ("sunrpc-unix");
	    }
	  _ilu_ReleaseMutex(ilu_prmu);
	}

      passport = ilu_CreatePassport();

      _sunrpc_InputCardinal (call, &credSize);
      _sunrpc_InputCardinal (call, &stamp);
      uid.au_magic = ILU_AS_SUNRPCUNIX_MAGIC;
      uid.au_hostname = NULL;
      _sunrpc_InputString (call, &uid.au_hostname, &stamp, 255);
      _sunrpc_InputCardinal (call, &uid.au_UID);
      _sunrpc_InputCardinal (call, &gid);
      uid.au_GID = gid;
      _sunrpc_InputCardinal (call, &uid.au_ngids);
      uid.au_gids = (gid_t *) ilu_must_malloc(sizeof(gid_t) * uid.au_ngids);
      for (stamp = 0;  stamp < uid.au_ngids;  stamp++)
	{
	  _sunrpc_InputCardinal(call, &gid);
	  uid.au_gids[stamp] = gid;
	}
      DEBUG(AUTHENTICATION_DEBUG,
	    (stderr, "Sun RPC UNIX auth:  uid=%d,gid=%d,hostname=\"%s\",%d groups\n",
	     uid.au_UID, uid.au_GID, uid.au_hostname, uid.au_ngids));
      err = authscheme_create_cred(SunRPCUnixAuthenticationScheme,(void *) &uid,&cred);
      ILU_ERR_SWITCH(err) {
	ILU_SUCCESS_CASE
	  ilu_AddCredential (passport, cred);
	ILU_ERR_ELSE
	  {
	    DEBUG(AUTHENTICATION_DEBUG,
		  (stderr, "_sunrpc_InterpretRequest:  couldn't create credentials from Sun RPC UNIX authentication credentials:  %s\n", ILU_ERR_NAME(err)));
	  }
      } ILU_ERR_ENDSWITCH;
      /* discard verifier, if any */
      _sunrpc_InputCardinal (call, &verifierType);
      _sunrpc_InputBytes (call, &verifierBuf, &verifierSize, 0);
    }

#endif  /* def USE_SUNRPC_UNIX_AUTHENTICATION */

  else
    {
      _sunrpc_InputBytes (call, &credBuf, &credSize, 0);
      _sunrpc_InputCardinal (call, &verifierType);
      _sunrpc_InputBytes (call, &verifierBuf, &verifierSize, 0);
      DEBUG(AUTHENTICATION_DEBUG,
	    (stderr,
	     "%s %lu%s %p, credSize is %lu, verfBuf is %p, verfSize is %lu\n",
	     "_sunrpc_InterpretRequest:  (call SN", call->ca_SN,
	     ") credBuf is", credBuf, credSize,
	     verifierBuf, verifierSize));
    }
  
  FREETOKEN(credBuf);
  FREETOKEN(verifierBuf);

  return passport;
}

static ilu_boolean _sunrpc_InterpretRequest (ilu_Call call, ilu_bytes packet,
				      ilu_cardinal len)
{
  struct SunRPC *c = SUNRPC(call_connection(call));
  sunrpcinfo *s;
  ilu_cardinal rpcVersion, serialNumber, packetType;
  ilu_cardinal programNumber, programVersion, methodID;

  c->sun_in = _xdr_CreatePacket (packet, len);

  _sunrpc_InputCardinal (call, &serialNumber);
  _sunrpc_InputCardinal (call, &packetType);
  _sunrpc_InputCardinal (call, &rpcVersion);
  _sunrpc_InputCardinal (call, &programNumber);
  _sunrpc_InputCardinal (call, &programVersion);
  _sunrpc_InputCardinal (call, &methodID);
  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr,
	 "%s  (call SN %lu) prognum 0x%lx (version %lu), method ID = %lu\n",
	 "_sunrpc_InterpretRequest:", call->ca_SN, programNumber,
	 programVersion, methodID));
  _ilu_Assert(serialNumber == call->ca_SN,
	      "sunrpc_InterpretRequest: serial number snafu");

  call->ca_caller = ReadCredentialsAndVerifier (call);

  _ilu_AcquireMutex(ilu_prmu);
  s = _sunrpc_ClassFromProgramNumber(programNumber, programVersion, ilu_TRUE);
  call_intro_type(call)
    = (s == NIL) ? NIL
		  : (s->sui_class == NIL) ? ilu_FindClassFromID(s->sui_type_id)
				       : s->sui_class;
  _ilu_ReleaseMutex(ilu_prmu);
  if (s == NIL OR call_intro_type(call) == NIL)
    {
      DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	    (stderr, "%s  (call %lu) %s 0x%lx, version %lu\n",
	     "_sunrpc_InterpretRequest:", call->ca_SN,
	     "Can't find ilu_Class with pn",
	     programNumber, programVersion));
      return (ilu_FALSE);
    }

  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr,
	 "_sunrpc_InterpretRequest:  (call SN %lu) intro_type is %s:%s\n",
	 call->ca_SN, call->ca_intro_type->cl_name,
	 call->ca_intro_type->cl_unique_id));

  call->ca_method = ilu_FindMethodByID (call_intro_type(call), methodID);
  
  if (call->ca_method == NIL)
    {
      DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	    (stderr,
	     "%s  (call %lu) %s \"%s\" (pn 0x%lx) with methodID = %lu.\n",
	     "_sunrpc_InterpretRequest:", call->ca_SN,
	     "Can't find method on class", call_intro_type(call)->cl_name,
	     programNumber, methodID));
      return (ilu_FALSE);
    }
  else
    {
      DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	    (stderr, "%s %lu is %p (%s).\n",
	     "_sunrpc_InterpretRequest:  record for method", 
	     methodID, call->ca_method, call->ca_method->me_name));
    }
  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr, "_sunrpc_InterpretRequest:  returning TRUE\n"));
  return (ilu_TRUE);
}

/*L1, Main unconstrained*/
/*L2 >= {conn's iomu}*/

static ilu_ProtocolException _sunrpc_InterpretReply (ilu_Call call,
					      ilu_bytes packet,
					      ilu_cardinal len,
					      ilu_cardinal *estatus)
{
  struct SunRPC *c = SUNRPC(call->ca_connection);
  ilu_cardinal verifierSize, replyStatus, serialNumber, packetType;
  ilu_bytes verifierBuf;
  ilu_cardinal authenticationType;
  static ilu_ProtocolException replyStatusExceptions[] = {
    ilu_ProtocolException_Success,
    ilu_ProtocolException_NoSuchClassAtServer,
    ilu_ProtocolException_ClassVersionMismatch,
    ilu_ProtocolException_NoSuchMethodOnClass,
    ilu_ProtocolException_GarbageArguments
    };
  static ilu_string acceptErrors[] = {
    "Success",
    "program unavailable",
    "program version mismatch",
    "procedure unavailable",
    "garbage arguments",
  };

  c->sun_in = _xdr_CreatePacket (packet, len);

  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr, "_sunrpc_InterpretReply:  SN %lu, %lu bytes\n",
	 call->ca_SN, len));

  _sunrpc_InputCardinal (call, &serialNumber);
  _sunrpc_InputCardinal (call, &packetType);
  _sunrpc_InputCardinal (call, &replyStatus);

  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	(stderr, "_sunrpc_InterpretReply:  replyStatus is %lu\n",
	 replyStatus));

  if (replyStatus == 0)		/* MSG_ACCEPTED */
    {
      _sunrpc_InputCardinal (call, &authenticationType);
      _sunrpc_InputBytes (call, &verifierBuf, &verifierSize, 0);
      FREETOKEN(verifierBuf);
      _sunrpc_InputCardinal (call, &replyStatus);

      DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	    (stderr,
	     "_sunrpc_InterpretReply:  *real* replyStatus is %lu\n",
	     replyStatus));

      if (replyStatus > 0)		/* *really* accepted */
	{
	  DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
		(stderr,
		 "Sun RPC call %lu signals protocol error %lu (%s).\n",
		 call->ca_SN, replyStatus,
		 (replyStatus < 5) ? acceptErrors[replyStatus]
				   : "unknown error"));
	}
      else if (call->ca_method->me_exceptionCount > 0)
	_sunrpc_InputCardinal (call, estatus);
      else
	*estatus = 0;

      if (replyStatus < 5)
	return (replyStatusExceptions[replyStatus]);
      else
	return(ilu_ProtocolException_Unknown);
    }
  else if (replyStatus == 1)	/* MSG_REJECTED */
    {
      DEBUG((INCOMING_DEBUG | SUNRPC_DEBUG),
	    (stderr, "Sun RPC call %lu signals \"Message Rejected\".\n",
	     call->ca_SN));
      return(ilu_ProtocolException_RequestRejected);
    }
  return(ilu_ProtocolException_Unknown);
}

/*L1_sup < prmu*/

static ilu_boolean GetSunRPCProgramNumberAndVersion (ilu_Class pclass,
			ilu_cardinal *pnumber, ilu_cardinal *version)
{
  sunrpcinfo *s = NIL;
  _ilu_AcquireMutex(ilu_prmu);
  s = _sunrpc_SunRPCInformationForClass (pclass);
	/* probably consult external DB */
  if (s != NIL) {
      *pnumber = s->sui_pnumber;
      *version = s->sui_version;
    }
  else
    {
      fprintf (stderr, "ILU: (GetSunRPCProgramNumberAndVersion):  Can't get program number information for class \"%s\".\n", pclass->cl_name);
    }
  _ilu_ReleaseMutex(ilu_prmu);
  return (s != NIL);
}

/*L1 >= {prmu}*/
#ifdef _IS_POSIX

#include "os/posix.h"

static int nameset = 0;
static char hostname[255];
static int hostnamelen, hostnamelenr, credlen;
static int gid, uid, ngids;
static gid_t gids[100];
#ifdef _IS_BSD
#include <netinet/in.h>
#include <netdb.h>
#endif /* def _IS_BSD */
#endif /* _IS_POSIX */

static int credlenr;

/*L1_sup < prmu*/

static ilu_cardinal FigureCredentialSize (ilu_Call call)
{
#ifdef _IS_POSIX
  if (!nameset) {

      _ilu_AcquireMutex(ilu_prmu);
      if (!nameset) {
          nameset = 1;
	  strcpy (hostname, _ilu_Hostname());

#ifdef _IS_BSD
	  {
	    struct hostent *he;
	    struct in_addr *hea;

	    he = gethostbyname(hostname);
	    hea = (struct in_addr *) (he->h_addr_list[0]);
	    strcpy (hostname, (ilu_string) inet_ntoa(*hea));
	  }
#endif /* def _IS_BSD */

          hostnamelen = strlen(hostname);
          hostnamelenr = 4 * ((hostnamelen+3)/4);
          gid = getgid();
          uid = geteuid();
          ngids = getgroups(sizeof(gids)/sizeof(gid), gids);
          _ilu_Assert(ngids >= 0, "getgroups() failed");
          credlen =   4/*stamp*/ + 4+hostnamelenr/*machinename*/
                    + 4/*uid*/ + 4/*gid*/ + 4+4*ngids/*gids*/;
          credlenr = 4 * ((credlen+3)/4);
        }
      _ilu_ReleaseMutex(ilu_prmu);
    }
#else
  credlenr = 4 * 2;
#endif
}

static void WriteCredentials (ilu_Call call)
{
#ifdef _IS_POSIX
  _sunrpc_OutputCardinal (call, 1);	/* UNIX credentials follow */
  _sunrpc_OutputCardinal (call, credlen); /* credentials length */
  _sunrpc_OutputCardinal (call, 1);	/* stamp */
  _sunrpc_OutputString (call, hostname, hostnamelen, 255);
  _sunrpc_OutputCardinal (call, uid);
  _sunrpc_OutputCardinal (call, gid);
  _sunrpc_OutputCardinal (call, ngids);
  {
    int i;
    for (i=0; i<ngids; i++)
      _sunrpc_OutputCardinal (call, gids[i]);
  }
#else
  _sunrpc_OutputCardinal (call, 0);	/* credentials:  AUTH_NULL */
  _sunrpc_OutputCardinal (call, 0);	/* 		 0 bytes */
#endif

  _sunrpc_OutputCardinal (call, 0);	/* verifier:  AUTH_NULL */
  _sunrpc_OutputCardinal (call, 0);	/*            0 bytes */
}

/*L1_sup < prmu*/

static ilu_boolean _sunrpc_BeginRequest (ilu_Call call, ilu_Class pclass,
					 ilu_Method method,
					 ilu_cardinal argSize)
{
  ilu_cardinal packetSize;
  ilu_cardinal pnumber=0, version=0;

  call->ca_method = method;

#if 0
  if (call_object(call)==NIL) { /* no call object, must be pipe req */
#else
  if (0) {
#endif
      DEBUG(CALL_DEBUG,
	    (stderr,
	     "%s %p (sn %lu), argSize %lu, class %s (%s), meth %s (%lu)\n",
	     "_sunrpc_BeginRequest:  pipe call", call,
	     call_serial_number(call), argSize, class_name(pclass),
	     class_unique_id(pclass), method_name(method),
	     method_id(method)));
    }
  else {
      DEBUG(CALL_DEBUG,
	    (stderr,
	     "%s %p (sn %lu), argSize %lu, class %s (%s), meth %s (%lu)\n",
	     "_sunrpc_BeginRequest:  call", call,
	     call_serial_number(call), argSize, class_name(pclass),
	     class_unique_id(pclass), method_name(method),
	     method_id(method)));

      if (!GetSunRPCProgramNumberAndVersion (pclass, &pnumber, &version))
        {
	  fprintf (stderr, "%s %s of class \"%s\" on call %p.\n",
		   "_sunrpc_BeginRequest:  Can't determine",
		   "program-number/version",
		   pclass->cl_name,  call);
          return (ilu_FALSE);
        }
    }
  DEBUG(CALL_DEBUG,
	(stderr,
	 "%s %p (sn %lu), argSize %lu, pnum/vers 0x%lx/%lu, method id %lu\n",
	 "_sunrpc_BeginRequest:  call", call, call_serial_number(call),
	 argSize, pnumber, version, method_id(method)));

  credlenr = FigureCredentialSize (call);

  packetSize = argSize
    + (4*10)		/* for fields of header */
    + credlenr		/* for UNIX credentials */;

  transport_flush_output(connection_transport(call_connection(call)));

  _sunrpc_OutputCardinal (call, call->ca_SN);
  _sunrpc_OutputCardinal (call, 0);	/* message type == CALL */
  _sunrpc_OutputCardinal (call, 2);	/* Sun RPC version 2 */
  _sunrpc_OutputCardinal (call, pnumber);
  _sunrpc_OutputCardinal (call, version);
  _sunrpc_OutputCardinal (call, method->me_id);

  WriteCredentials (call);

  DEBUG(CALL_DEBUG,
	(stderr, "_sunrpc_BeginRequest:  request %lu begun (size %lu).\n",
	 call->ca_SN, packetSize));
  return (ilu_TRUE);
}

/*L1, Main unconstrained*/
/*L2 >= {conn's iomu}*/

static void _sunrpc_DropInput (ilu_Call call)
{
  ilu_Connection  conn = call_connection(call);
  struct SunRPC  *s = SUNRPC(conn);
  if (connection_closed(conn)) {
    _ilu_Assert(s == NIL,
		"protocol_data not NIL for closed connection");
    return;
  }
  if (s->sun_in != NIL)
    packet_destroy(s->sun_in);
  s->sun_in = NIL;
}

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static void _sunrpc_FlushBuffer (ilu_Call call)
{
  transport_flush_output(call_transport(call));
}

static void _sunrpc_DropRequest (ilu_Call call)
{
  transport_flush_output(connection_transport(call_connection(call)));
}

static          ilu_boolean
_sunrpc_SendPacket(ilu_Call call, ilu_boolean isReply)
{
  ilu_Error       err;

  ilu_Transport   bs = call_transport(call);
  err = transport_flush_output(bs);
  if (err != NIL) {
    fprintf(stderr,
	    "(ILU)_sunrpc_SendPacket:"
	    "  Error \"%s: %s\" in write of packet.\n",
	    ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
    ILU_ERR_SWITCH(err) {
      ILU_ERR_CASE(WriteError, ev) {
	fprintf(stderr,
		"(ILU)_sunrpc_SendPacket:"
		"  write error on packet, errno %u, \"%s\".\n",
		ev->ep_errno, ev->errstr);
	ilu_free(ev->buffer);	/* pretty useless by now */
      }
      ILU_ERR_ELSE
	fprintf(stderr,
		"(ILU)_sunrpc_SendPacket:"
		"  Error \"%s: %s\" in write of packet.\n",
		ILU_ERR_NAME(err), ILU_ERR_DESCRIPTION(err));
    } ILU_ERR_ENDSWITCH;
    ilu_FreeError(err);
    return (ilu_FALSE);
  }
  if (bs->tr_class->tc_timesout) {
    if (isReply) {
      err = _ilu_CacheCall(call, bs->tr_packet, bs->tr_size);
      ILU_ERR_SWITCH(err) {
	ILU_SUCCESS_CASE;
	ILU_ERR_CASE(TInfoTooLong, x)
	  fprintf(stderr, "(ILU)_sunrpc_SendPacket:"
		  "  cant cache reply because transport info"
		  " (%s) too long", x->tinfo);
      } ILU_ERR_ENDSWITCH;
      ilu_FreeError(err);
    } else
      ilu_free(bs->tr_packet);
    bs->tr_packet = NIL;
  }
  return (ilu_TRUE);
}

static ilu_boolean _sunrpc_FinishRequest (ilu_Call call)
{
  return (_sunrpc_SendPacket(call, ilu_FALSE));
}

static ilu_boolean _sunrpc_FinishReply (ilu_Call call)
{
  return (_sunrpc_SendPacket (call, ilu_TRUE));
}

static ilu_boolean _sunrpc_FinishException (ilu_Call call)
{
  return (_sunrpc_SendPacket(call, ilu_TRUE));
}

/*L1, Main unconstrained*/
/*L2 >= {call's conn's iomu}*/

static ilu_boolean _sunrpc_BeginReply (ilu_Call call,
				       ilu_boolean exceptions,
				       ilu_cardinal argSize)
{
  ilu_cardinal packetSize;

  DEBUG(PACKET_DEBUG,
	(stderr, "%s %lu, argSize %lu, exceptions %s, fd %d.\n",
	 "_sunrpc_BeginReply:  SN", call->ca_SN, argSize,
	 exceptions ? "TRUE" : "FALSE",
	 transport_file_descriptor(call_transport(call))));

  packetSize = argSize
    + (4*6)			/* for the basic header fields */
    + (exceptions ? 4 : 0);	/* possible extra word for excn code */

  transport_flush_output(call_transport(call));

  _sunrpc_OutputCardinal (call, call->ca_SN);
  _sunrpc_OutputCardinal (call, 1);		/* message type ==REPLY */
  _sunrpc_OutputCardinal (call, 0);		/* message accepted */
  _sunrpc_OutputCardinal (call, 0);		/* verifier:  AUTH_NULL */
  _sunrpc_OutputCardinal (call, 0);		/*            0 bytes */
  _sunrpc_OutputCardinal (call, 0);		/* successful execution */
  if (exceptions)
    _sunrpc_OutputCardinal (call, 0);		/* ret code is Success */

  DEBUG(PACKET_DEBUG,
	(stderr, "_sunrpc_BeginReply:  started reply %lu (size %lu).\n",
	 call->ca_SN, packetSize));
  return (ilu_TRUE);
}

static ilu_boolean _sunrpc_BeginException (ilu_Call call,
					   ilu_cardinal evalue,
					   ilu_cardinal argSize)
{
  ilu_cardinal packetSize;

  /* if "evalue" == 0, then argSize contains a protocol exception
     detail code. */

  if (evalue == 0)	/* signal protocol error */
    packetSize = (4*6);
  else
    packetSize = (4*7) + argSize;

  transport_flush_output(call_transport(call));

  _sunrpc_OutputCardinal (call, call->ca_SN);
  _sunrpc_OutputCardinal (call, 1);		/* message type ==REPLY */
  _sunrpc_OutputCardinal (call, 0);		/* message accepted */
  _sunrpc_OutputCardinal (call, 0);		/* verifier:  AUTH_NULL */
  _sunrpc_OutputCardinal (call, 0);		/*            0 bytes */
  _sunrpc_OutputCardinal (call, (evalue == 0) ? argSize : 0);
						/* successful execution */
  if (evalue > 0)
    _sunrpc_OutputCardinal (call, evalue);	/* exception value */

  DEBUG(PACKET_DEBUG,
	(stderr, "%s %d:  SN %lu, size %lu, evalue %lu.\n",
	 "_sunrpc_BeginException:  exception started to peer",
	 transport_file_descriptor(call_transport(call)),
	 call_serial_number(call), packetSize, evalue));
  return (ilu_TRUE);
}

/*L2, Main unconstrained*/
/*L1 >= {prmu}*/

static ilu_Protocol _sunrpc_NewSunRPC (void)
{
  ilu_Protocol new = (ilu_Protocol)
		     ilu_malloc(sizeof(struct _ilu_Protocol_s));

  new->pr_concurrent_requests = ilu_FALSE;

  new->pr_begin_request = _sunrpc_BeginRequest;
  new->pr_finish_request = _sunrpc_FinishRequest;
  new->pr_resend_request = _sunrpc_FinishRequest;
  new->pr_drop_request = _sunrpc_DropRequest;
  new->pr_begin_reply = _sunrpc_BeginReply;
  new->pr_finish_reply = _sunrpc_FinishReply;
  new->pr_begin_exception = _sunrpc_BeginException;
  new->pr_finish_exception = _sunrpc_FinishException;

  new->pr_output_integer = _sunrpc_OutputInteger;
  new->pr_input_integer = _sunrpc_InputInteger;
  new->pr_size_of_integer = _sunrpc_SizeOfInteger;

  new->pr_output_shortinteger = _sunrpc_OutputShortInteger;
  new->pr_input_shortinteger = _sunrpc_InputShortInteger;
  new->pr_size_of_shortinteger = _sunrpc_SizeOfShortInteger;

  new->pr_output_longinteger = _sunrpc_OutputLongInteger;
  new->pr_input_longinteger = _sunrpc_InputLongInteger;
  new->pr_size_of_longinteger = _sunrpc_SizeOfLongInteger;

  new->pr_output_cardinal = _sunrpc_OutputCardinal;
  new->pr_input_cardinal = _sunrpc_InputCardinal;
  new->pr_size_of_cardinal = _sunrpc_SizeOfCardinal;

  new->pr_output_shortcardinal = _sunrpc_OutputShortCardinal;
  new->pr_input_shortcardinal = _sunrpc_InputShortCardinal;
  new->pr_size_of_shortcardinal = _sunrpc_SizeOfShortCardinal;

  new->pr_output_longcardinal = _sunrpc_OutputLongCardinal;
  new->pr_input_longcardinal = _sunrpc_InputLongCardinal;
  new->pr_size_of_longcardinal = _sunrpc_SizeOfLongCardinal;

  new->pr_output_real = _sunrpc_OutputReal;
  new->pr_input_real = _sunrpc_InputReal;
  new->pr_size_of_real = _sunrpc_SizeOfReal;

  new->pr_output_shortreal = _sunrpc_OutputShortReal;
  new->pr_input_shortreal = _sunrpc_InputShortReal;
  new->pr_size_of_shortreal = _sunrpc_SizeOfShortReal;

  new->pr_output_longreal = _sunrpc_OutputLongReal;
  new->pr_input_longreal = _sunrpc_InputLongReal;
  new->pr_size_of_longreal = _sunrpc_SizeOfLongReal;

  new->pr_output_optional =
	(ilu_boolean (*)(ilu_Call,ilu_boolean)) _sunrpc_OutputCardinal;
  new->pr_input_optional =
	(ilu_boolean (*)(ilu_Call,ilu_boolean *)) _sunrpc_InputCardinal;
  new->pr_size_of_optional =
	(ilu_cardinal (*)(ilu_Call,ilu_boolean)) _sunrpc_SizeOfCardinal;

  new->pr_output_enum_code = _sunrpc_OutputEnumeration;
  new->pr_input_enum_code = _sunrpc_InputEnumeration;
  new->pr_size_of_enum_code = _sunrpc_SizeOfEnumeration;

  new->pr_output_byte = _sunrpc_OutputByte;
  new->pr_input_byte = _sunrpc_InputByte;
  new->pr_size_of_byte = _sunrpc_SizeOfByte;

  new->pr_output_character = _sunrpc_OutputShortCardinal;
  new->pr_input_character = _sunrpc_InputShortCardinal;
  new->pr_size_of_character = _sunrpc_SizeOfShortCardinal;

  new->pr_output_boolean =
	(ilu_boolean (*)(ilu_Call,ilu_boolean)) _sunrpc_OutputCardinal;
  new->pr_input_boolean =
	(ilu_boolean (*)(ilu_Call,ilu_boolean *)) _sunrpc_InputCardinal;
  new->pr_size_of_boolean =
	(ilu_cardinal (*)(ilu_Call,ilu_boolean)) _sunrpc_SizeOfCardinal;

  new->pr_output_shortchar = _sunrpc_OutputByte;
  new->pr_input_shortchar = _sunrpc_InputByte;
  new->pr_size_of_shortchar = _sunrpc_SizeOfByte;

  new->pr_output_string = _sunrpc_OutputString;
  new->pr_input_string = _sunrpc_InputString;
  new->pr_size_of_string = _sunrpc_SizeOfString;

  new->pr_output_bytes = _sunrpc_OutputBytes;
  new->pr_input_bytes = _sunrpc_InputBytes;
  new->pr_size_of_bytes = _sunrpc_SizeOfBytes;

  new->pr_output_opaque = _sunrpc_OutputOpaque;
  new->pr_input_opaque = _sunrpc_InputOpaque;
  new->pr_size_of_opaque = _sunrpc_SizeOfOpaque;

  new->pr_output_object_id = _ilu_OutputObjectID;
  new->pr_input_object_id = _ilu_InputObjectID;
  new->pr_size_of_object_id = _ilu_SizeOfObjectID;

  new->pr_output_stringvec =
	(ilu_boolean (*)(ilu_Call,ilu_string,ilu_cardinal))
	_sunrpc_OutputOpaque;
  new->pr_input_stringvec =
	(ilu_boolean (*)(ilu_Call,ilu_string *,ilu_cardinal))
	_sunrpc_InputOpaque;
  new->pr_size_of_stringvec =
	(ilu_cardinal (*)(ilu_Call,ilu_string,ilu_cardinal))
	_sunrpc_SizeOfOpaque;

  new->pr_output_pipe = NULLFN;
  new->pr_input_pipe = NULLFN;
  new->pr_size_of_pipe = NULLFN;

  new->pr_output_sequence = _sunrpc_OutputSequence;
  new->pr_output_sequence_mark = _sunrpc_OutputSequenceMark;
  new->pr_input_sequence = _sunrpc_InputSequence;
  new->pr_input_sequence_mark = _sunrpc_InputSequenceMark;
  new->pr_end_sequence = _sunrpc_EndSequence;
  new->pr_size_of_sequence = _sunrpc_SizeOfSequence;

  new->pr_output_record = _sunrpc_OutputRecord;
  new->pr_input_record = _sunrpc_InputRecord;
  new->pr_end_record = _sunrpc_EndRecord;
  new->pr_size_of_record = _sunrpc_SizeOfRecord;

  new->pr_output_array = _sunrpc_OutputArray;
  new->pr_input_array = _sunrpc_InputArray;
  new->pr_end_array = _sunrpc_EndArray;
  new->pr_size_of_array = _sunrpc_SizeOfArray;

  new->pr_output_union = _sunrpc_OutputUnion;
  new->pr_input_union = _sunrpc_InputUnion;
  new->pr_end_union = _sunrpc_EndUnion;
  new->pr_size_of_union = _sunrpc_SizeOfUnion;

  new->pr_form_handle = _sunrpc_FormProtocolHandle;

  new->pr_read_packet = _sunrpc_ReadPacket;
  new->pr_interpret_pipe_request = _sunrpc_InterpretPipeRequest;
  new->pr_interpret_request = _sunrpc_InterpretRequest;
  new->pr_request_read = _sunrpc_DropInput;
  new->pr_interpret_reply = _sunrpc_InterpretReply;
  new->pr_finish_call = _sunrpc_DropInput;

  new->pr_create_data_block = _sunrpc_CreateDataBlock;
  new->pr_free_data_block = (void (*)(void *)) _sunrpc_FreeDataBlock;
  new->pr_flush = _sunrpc_FlushBuffer;

  return (new);
}

/*L1_sup < prmu*/

ilu_Protocol _ilu_sunrpc_Protocol(void)
{
  /*L1 >= {prmu}*/
  static ilu_Protocol StandardSunRPC = NIL;
  _ilu_AcquireMutex(ilu_prmu);
  if (StandardSunRPC == NIL)
    StandardSunRPC = _sunrpc_NewSunRPC();
  _ilu_ReleaseMutex(ilu_prmu);
  return (StandardSunRPC);
}
