/*
Copyright (c) 1991-1994 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: tcp.c,v 1.61 1995/02/12 01:45:38 janssen Exp $ */
/* Last edited by Mike Spreitzer December 13, 1994 4:38 pm PST */

#ifdef _IS_BSD	/* we believe FIONREAD to be a BSD-ism.  Is this true? */
#define BSD_COMP
#ifndef FIONREAD
#include <sys/ioctl.h>	/* for FIONREAD */
#endif
#endif

#include "iluntrnl.h"

#include "transprt.h"
#include "mooring.h"

#include "socket.h"

#include "oscalls.h"

#include <signal.h>

#ifdef _IS_POSIX
#include <sys/types.h>
#include <fcntl.h>
static int set_non_blocking (int fd)
{
  int flags;

  if ((flags = fcntl(fd, F_GETFL, 0)) < 0 OR
      fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0)
    return (errno);
  else
    return 0;
}
#elif _IS_BSD
static int set_non_blocking (int fd)
{
  int one = 1;

  return (ioctl(fd, FIONBIO, (int) &one));
}
#else	/* not POSIX or BSD */
"Don't know how to do non-blocking I/O for anything but POSIX and BSD just now"
#endif

/*L1_sup < trmu; L2, Main unconstrained*/
ilu_TransportClass _ilu_tcp_TransportClass(void);

typedef struct tcpparms {
  /*L1, L2, Main unconstrained*/
  
  ilu_string hostname;
  ilu_cardinal port;
  
  /*L1 >= {connection's iomu}*/
  ilu_bytes buffer;
  ilu_cardinal size;
  ilu_bytes next;
  ilu_cardinal left;
  ilu_boolean in_msg;	/* TRUE if in middle of output msg */
} *TCPParms;
/* What goes in the data field of a TCP ilu_Transport or ilu_Mooring */

/*L1, L2, Main unconstrained*/

#define TCP_HOSTNAME(a) (((TCPParms)(a))->hostname)
#define TCP_PORT(a) (((TCPParms)(a))->port)

/*L1 >= {connection's iomu}*/
#define TCP_BUFFER(a) (((TCPParms)(a))->buffer)
#define TCPPARMS(a) ((TCPParms)(a))

#define BUFFERSIZE	4096

#define MAXDUMP		10000

/***********************************************************************
************************************************************************
************************************************************************
************************** Some local utilities ************************
************************************************************************
************************************************************************
***********************************************************************/

/*L1 >= {trmu}*/
ilu_string _ilu_tcp_CurrentHostInetName(void)
{
  char *name;
  static ilu_string inetname = NIL;
  struct hostent *he;
  struct in_addr *hea;

  if (inetname == NIL)
    {
      if ((name = _ilu_Hostname()) == NIL) {
	perror("no hostname for this machine!");
	return (NIL);
      }
  
      he = gethostbyname(name);
      hea = (struct in_addr *) (he->h_addr_list[0]);
      inetname = _ilu_Strdup((ilu_string) inet_ntoa(*hea));
    }
  return (inetname);
}

/**********************************************************************
***********************************************************************
***********************************************************************
***** First, the methods for the TCP Transport ************************
***********************************************************************
***********************************************************************
**********************************************************************/

/*L2, Main unconstrained*/

/*L1 >= {trmu}*/
static ilu_boolean SigPIPEHandler = FALSE;

/*L1_sup < trmu*/

static void
_tcp_HandleSigPIPE(void)
{
  /*
   * Ignore SIGPIPEs, since we can occasionally get them. [Thanks
   * to Dave Nichols <nichols@parc.xerox.com> for the original idea
   * of this code.]
   */

#if defined(_IS_POSIX) && defined(SA_NOCLDSTOP)

  _ilu_AcquireMutex(ilu_trmu);
  if (!SigPIPEHandler) {

    struct sigaction old_handler;
    static struct sigaction new_handler;

    if (sigaction(SIGPIPE, NIL, &old_handler) == 0) {
      if (old_handler.sa_handler == SIG_DFL) {	/* no one's using it */
	new_handler.sa_handler = SIG_IGN;
	if (sigaction(SIGPIPE, &new_handler, NIL) != 0) {
	  fprintf(stderr,
		  "ILU:  _tcp_SetSIGPIPE:  Couldn't ignore SIGPIPE signals!\n");
	}
      }
    } else
      fprintf(stderr,
	      "ILU:  _tcp_SetSIGPIPE:  Couldn't read handler for SIGPIPE!\n");
  }
  SigPIPEHandler = TRUE;
  _ilu_ReleaseMutex(ilu_trmu);

#else		/* not really POSIX -- use ANSI C form */

  typedef void    (*handler) (int);
  _ilu_AcquireMutex(ilu_trmu);
  if (!SigPIPEHandler) {
    handler         old_handler;

    if ((old_handler = (handler) signal(SIGPIPE, SIG_IGN)) != SIG_DFL)
      /* Oops!  Someone's using this one */
      signal(SIGPIPE, old_handler);

    SigPIPEHandler = TRUE;
  }
  _ilu_ReleaseMutex(ilu_trmu);

#endif				/* _IS_POSIX */
}

ilu_boolean
ilu_SIGPIPE_Handled(void)
{
  _ilu_AcquireMutex(ilu_trmu);
  if (!SigPIPEHandler) {
    SigPIPEHandler = TRUE;
    _ilu_ReleaseMutex(ilu_trmu);
    return TRUE;
  }
  _ilu_ReleaseMutex(ilu_trmu);
  return FALSE;
}

/*L1 unconstrained*/

static ilu_private NewData (char *hostname, ilu_cardinal port, ilu_cardinal buffersize)
{
  TCPParms new = (TCPParms) ilu_must_malloc(sizeof(struct tcpparms));
  new->hostname = _ilu_Strdup(hostname);
  new->port = port;
  if ((new->size = buffersize) > 0)
    new->buffer = (ilu_bytes) ilu_must_malloc(new->size);
  else
    new->buffer = NIL;
  new->left = new->size - 4;	/* reserve 4 bytes for explicit size */
  new->next = new->buffer + 4;
  new->in_msg = ilu_FALSE;
  return ((ilu_private) new);
}

static void FreeData (ilu_private data)
{
  TCPParms old = (TCPParms) data;

  FREETOKEN(old->hostname);
  FREETOKEN(old->buffer);
  FREETOKEN(old);
}

static ilu_private _tcp_InterpretInfo (ilu_string info)
{
  char hostname[1000];
  ilu_cardinal port;

  if ((sscanf (info, "tcp_%[^_]_%lu", hostname, &port)) == 2)
    return (NewData (hostname, port, BUFFERSIZE));
  else
    return (NIL);
}

static void DumpPacket (ilu_byte *packet, ilu_shortcardinal length)
{
  ilu_cardinal dumplength;
  int n;
  ilu_byte c;
  register ilu_integer i, j;

  if (length > MAXDUMP)
    {
      fprintf (stderr, "Request to dump packet of %u bytes.  Only %u bytes being dumped.\n",
	       length, MAXDUMP);
      dumplength = MAXDUMP;
    }
  else
    dumplength = length;
  if (packet == NIL)
    {
      fprintf (stderr, "Attempt to dump NIL packet.\n");
      return;
    }
  fprintf (stderr, "DumpPacket of packet %p, length ", (void *) packet);
  fprintf (stderr, "%u bytes, dumping %lu bytes:\n", length, dumplength);
  for (i = 0;  i < dumplength;  i += 16)
    {
      fprintf (stderr, "%6ld:  ", i);
      for (j = 0;  j < 16 AND (i + j) < dumplength;  j += 1)
	fprintf (stderr, "%02x%s ", packet[i + j],
		 ((j % 4) == 3) ? " " : "");
      n = (((16-j)/4)*(13)+((16-j)%4)*3)+1; /* padding before ascii */
      fprintf (stderr, "%*.*s", n, n, "");
      for (j = 0;  j < 16 AND (i + j) < dumplength;  j += 1)
	{
	  c = packet[i + j];
	  fprintf (stderr, "%c", ((c >= ' ') && (c <= '~')) ? (char) c
							    : '.');
	}
      fprintf (stderr, "\n");
    }
}

#ifdef _IS_BSD	/* we believe FIONREAD to be a BSD-ism.  Is this true? */

/*L2 >= {conn's iomu}*/
/*L1, Main unconstrained*/
static ilu_integer _tcp_NDataPresent (ilu_Transport conn)
{
  long count;
  int status = ioctl (conn->tr_fd, FIONREAD, (int) &count);
  /* we cast the third arg to ioctl 'cause it's declared as int. */
  if (status==0 && count==0) {
      int errnum = 0;
      SOCKET_SIZE_TYPE errsize = sizeof(errnum);
      status = getsockopt(conn->tr_fd, SOL_SOCKET, SO_ERROR,
                          (char *) &errnum, &errsize);
      if (status < 0 || errnum != 0)
           return -1;
      else return 0;
    }
  return ( (status != 0) ? -1 : (count > 0) ? 1 : 0 );
}

#else

"Don't know how to do FIONREAD on anything but BSD"

#endif

/*Main Invariant holds*/
/*L2 >= {conn's iomu}*/

static ilu_boolean _tcp_WaitForInput (ilu_Transport t, ilu_boolean *sure,
				      ilu_FineTime *limit)
{
  _ilu_WaitForInputOnFD(t->tr_fd, sure, limit);
  return (TRUE);
}

static ILU_ERRS((BadDataStructure, WriteError))
_tcp_WriteBuffer(ilu_Transport self, ilu_boolean EOM)
{
  ilu_Error       status = NIL;
  ilu_integer     i;
  ilu_cardinal    size;
  TCPParms        p;

  if (self == NIL OR(p = TCPPARMS(transport_data(self))) == NIL) {
    BIND_BadDataStructure(&status, self, "ilu_Transport (tcp)",
			  "NIL or bad TCPParms")
  } else if (p->buffer != NIL) {
    size = p->size - p->left - 4;
    DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
	  (stderr, "%s %ld bytes from %p to fd %d.\n",
	   "_tcp_WriteBuffer:  writing", size,
	   (void *) p->buffer, self->tr_fd));

    p->buffer[0] = ((size >> 24) & 0xFF) | (EOM ? 0x80 : 0);
    p->buffer[1] = (size >> 16) & 0xFF;
    p->buffer[2] = (size >> 8) & 0xFF;
    p->buffer[3] = size & 0xFF;

    if ((_ilu_DebugLevel & PACKET_DEBUG) != 0)
      DumpPacket(p->buffer + 4, size);

    if ((i = _ilu_Write(self->tr_fd, p->buffer, size + 4)) < size + 4) {
      int the_errno = errno;
      ILU_BIND_ERR(WriteError, &status, ev) {
	ev->transport = self;
	ev->return_status = i;
	ev->buffer = p->buffer;
	ev->size = size + 4;
	ev->ep_errno = the_errno;
	ev->errstr = strerror(the_errno);

	fprintf(stderr,
		"%s %u of %lu bytes signalled error \"%s\".\n",
	"(_tcp_WriteBuffer):  output to fd", self->tr_fd, size + 4,
		(i < 0) ? strerror(the_errno)
		: ((i == 0) ? "Connection lost"
		   : "not enough bytes written"));
      } ILU_END_BIND_ERR;
    } else {
      p->left = p->size - 4;
      p->next = p->buffer + 4;
      if (EOM)
	p->in_msg = ilu_FALSE;
    }
  }
  return (status);
}

static  ILU_ERRS((BadDataStructure, WriteError))
_tcp_WriteBytes(ilu_Transport self, ilu_bytes b, ilu_cardinal l)
{
  ilu_Error err;
  TCPParms p = TCPPARMS(transport_data(self));
  ilu_cardinal size = l;

  p->in_msg = ilu_TRUE;
  while (l > 0)
    {
      size = MIN(p->left, l);
      memcpy ((void *) p->next, (void *) b, size);
      l -= size;
      b += size;
      p->next += size;
      p->left -= size;
      if (p->left == 0)
	{
	  if ((err = _tcp_WriteBuffer(self, ilu_FALSE)) != NIL)
	    return err;
	}
    }
  return NIL;
}

static          ilu_bytes
_tcp_GetOutputBuffer(ilu_Transport self, ilu_cardinal l,
	    ILU_ERRS((BadDataStructure, InvalidBufferSize, WriteError)) * err)
{
  register TCPParms p = TCPPARMS(transport_data(self));
  register ilu_bytes ptr;

  if (p->size < l)
    return ILU_ERR_CONS3(InvalidBufferSize, err,
			 minsize, 0,
			 maxsize, p->size,
			 requested_size, l, NIL);
  else {
    p->in_msg = ilu_TRUE;
    if (p->left < l) {
      *err = _tcp_WriteBuffer(self, ilu_FALSE);
      if (*err != NIL)
	return NIL;
    }
    ptr = p->next;
    p->next += l;
    p->left -= l;
    *err = ILU_NO_ERR;
    return ptr;
  }
}

static 
ILU_ERRS((BadDataStructure, WriteError))
_tcp_FlushOutput(ilu_Transport self)
{
  TCPParms p = TCPPARMS(transport_data(self));
  if (p->in_msg)
    return (_tcp_WriteBuffer (self, ilu_TRUE));
  else
    return NIL;
}

/*L1, L2, Main unconstrained*/

/*Main Invariant holds*/
static ilu_boolean _tcp_Connect (ilu_Transport self)
{
  int fd, err;
  struct sockaddr_in sin;
  struct hostent *hp;
  ilu_boolean status = TRUE;

  _ilu_AutoSetDebugLevel();
  _tcp_HandleSigPIPE();

  if (self == NIL)
    return (FALSE);

  DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
	(stderr, "_tcp_Connect:  connecting to host %s, port %lu...\n",
	 TCP_HOSTNAME(transport_data(self)),
	 TCP_PORT(transport_data(self))));

  memset((ilu_string ) &sin, 0, sizeof(sin));
  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
	    (stderr, "_tcp_Connect:  socket call failed:  %s.\n",
	     strerror(errno)));
      status = FALSE;
    }
  else
    {
#ifdef SO_REUSEADDR
      (void) setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
			   (ilu_string ) NIL, 0);
#endif /* SO_REUSEADDR */

#ifdef SO_USELOOPBACK
      (void) setsockopt(fd, SOL_SOCKET, SO_USELOOPBACK,
			   (ilu_string ) NIL, 0);
#endif /* SO_USELOOPBACK */

/* The following was suggested by Ken Birman of the Cornell ISIS project.
 * On SunOS it prevents the kernel from delaying this packet in hopes
 * of merging it with a quickly-following one.
 */
#ifdef IPPROTO_TCP
#ifdef TCP_NODELAY
      {
	int one = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
		      (char *) &one, sizeof(one));
      }
#endif /* TCP_NODELAY */
#endif /* IPPROTO_TCP */     

      sin.sin_family = AF_INET;
      sin.sin_port = htons(TCP_PORT(transport_data(self)));

      sin.sin_addr.s_addr = inet_addr( TCP_HOSTNAME(
		transport_data( self)));
      if (sin.sin_addr.s_addr == -1 || sin.sin_addr.s_addr == 0)
	if ((hp = gethostbyname (TCP_HOSTNAME(transport_data(self))))
	    != NIL)
	  memcpy ((char *) &sin.sin_addr, hp->h_addr, hp->h_length);

      if (sin.sin_addr.s_addr == -1 || sin.sin_addr.s_addr == 0)
	{
	  DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		(stderr, "_tcp_Connect:  Invalid host name (%s)\n",
		 TCP_HOSTNAME(transport_data(self))));
	  status = FALSE;
	}
      else if ((err = set_non_blocking(fd)) != 0)
	{
	  DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		(stderr, "_tcp_Connect:  Failed to set socket (descriptor %d, host %s) non-blocking.  Error \"%s\".\n",
		 fd, TCP_HOSTNAME(transport_data(self)), strerror(err)));
	  status = FALSE;
	}
      else
	{
	  if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) == 0)
	    {
	      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		    (stderr, "_tcp_Connect:  connected to %s[%lu]\n",
		     inet_ntoa(sin.sin_addr),
		     TCP_PORT(transport_data(self))));
	      self->tr_fd = fd;
	      status = TRUE;
	    }
	  else if (errno == EINPROGRESS) {
	    struct sockaddr peername;
	    SOCKET_SIZE_TYPE pnlen = sizeof(peername);
	    ilu_boolean sure;
	    _ilu_WaitForOutputOnFD(fd, &sure, NIL);
	    if (getpeername(fd, &peername, &pnlen) == 0) {
	      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		    (stderr,
		     "_tcp_Connect:  eventually connected to %s[%lu]\n",
		     inet_ntoa(sin.sin_addr),
		     TCP_PORT(transport_data(self))));
	      self->tr_fd = fd;
	      status = TRUE;
	      }
	    else {
	      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		    (stderr,
		     "_tcp_Connect: connect to %s[%lu] failed "
		     "(no meaningful errno available).\n",
		     inet_ntoa(sin.sin_addr),
		     TCP_PORT(transport_data(self)) ));
	      close(fd);
	      status = FALSE;
	      }
	    }
	  else
	    {
	      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		    (stderr, "_tcp_Connect: connect to %s[%lu] failed,"
		     " error %d (%s)\n",
		     inet_ntoa(sin.sin_addr),
		     TCP_PORT(transport_data(self)), errno,
		     strerror(errno)));
	      close(fd);
	      status = FALSE;
	    }
	}
    }
  return (status);
}

/*L2 >= {conn's iomu}*/

static void _tcp_Close (ilu_Transport self)
{
  if (self == NIL)
    return;

  ilu_UnregisterInputSource(self->tr_fd);
  ilu_UnregisterOutputSource(self->tr_fd);
  close (self->tr_fd);

  FreeData(self->tr_data);
  FREETOKEN(self->tr_packet);
  return;
}

/*Main Invariant holds*/
static ilu_boolean _tcp_ReadMessage (ilu_Transport self,
				     ilu_bytes *buffer,
				     ilu_cardinal *size)
{
  ilu_byte lenbuf[4];
  int i;
  ilu_cardinal bytesToRead, bytesRead, totalSize = 0;
  ilu_bytes packet = NIL;
  ilu_boolean status = FALSE;
  ilu_boolean lastPacket = FALSE;

  if (self == NIL)
    return(FALSE);

  do {
    if ((i = _ilu_Read (self->tr_fd, lenbuf, 4)) < 4)
      {
	DEBUG((INCOMING_DEBUG | TCP_DEBUG),
	      (stderr,
	       "%s %d returns %d (errno is %s) %s.\n",
	       "_tcp_ReadMessage:  read on fd",
	       self->tr_fd, i, strerror(errno),
	       "while reading packet length"));
	status = FALSE;
      }
    else
      {
	DEBUG((INCOMING_DEBUG | TCP_DEBUG),
	      (stderr, "_tcp_ReadMessage:  lenbuf is %d.%d.%d.%d\n",
	       lenbuf[0], lenbuf[1], lenbuf[2], lenbuf[3]));
	lastPacket = (lenbuf[0] & 0x80) != 0;
	bytesToRead = ((lenbuf[0] & 0x7F) << 24) | (lenbuf[1] << 16)
		      | (lenbuf[2] << 8) | lenbuf[3];
	bytesRead = 0;
	if (packet == NIL)
	  packet = (ilu_bytes) ilu_must_malloc(bytesToRead);
	else
	  packet = (ilu_bytes) ilu_realloc (packet, bytesToRead + totalSize);
	status = TRUE;
	i = _ilu_Read (self->tr_fd, packet + totalSize + bytesRead,
		       bytesToRead);
	if ((i < 0) || ((i == 0) && (bytesToRead > 0)))
	  {
	    DEBUG((INCOMING_DEBUG | TCP_DEBUG),
		  (stderr,
		   "_tcp_ReadMessage:  read on fd %d returns %d (%s)"
		   " while reading packet of length %lu,"
		   " having already read %lu bytes.\n",
		   self->tr_fd, i, strerror(errno),
		   bytesToRead, bytesRead));
	    status = FALSE;
	  }
	else
	  {
	    DEBUG((INCOMING_DEBUG | TCP_DEBUG),
		  (stderr,
		   "_tcp_ReadMessage:  read %d bytes from fd %d"
		   " while asking for %lu bytes.\n",
		   i, self->tr_fd, bytesToRead));
	    bytesRead += i;
	    bytesToRead -= i;
	  }
	totalSize += bytesRead;
      }
  } while (status AND (NOT lastPacket));

  if (status)
    {
      *buffer = packet;
      *size = totalSize;
    }
  else
    {
      if (packet != NIL)
	ilu_free(packet);
    }

  if (status AND (_ilu_DebugLevel & PACKET_DEBUG) != 0)
    DumpPacket(*buffer, *size);

  return (status);
}

/**********************************************************************
***********************************************************************
***********************************************************************
**** Now the methods for the TCP Mooring ******************************
***********************************************************************
***********************************************************************
**********************************************************************/

/*L1, L2, Main unconstrained*/

static ilu_string _tcp_FormHandle (ilu_refany parameterblock)
{
  TCPParms parms = TCPPARMS(parameterblock);
  char buf[1000];

  if (parms == NIL)
    return (NIL);
  sprintf (buf, "tcp_%s_%lu", TCP_HOSTNAME(parms), TCP_PORT(parms));
  return (_ilu_Strdup(buf));
}

/*L1_sup < trmu*/
static ilu_Transport _tcp_AcceptClient (ilu_Mooring self)
{
  register ilu_integer ns;
  struct sockaddr_in sin;
  ilu_Transport new = NIL;
  SOCKET_SIZE_TYPE addrlen;
  int flags;
  char addrDescr[64];

  _ilu_AutoSetDebugLevel();

  if (self == NIL)
    return (NIL);

  DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
	(stderr, "_tcp_AcceptClient: mooring file descriptor is %d\n",
	 self->mo_fd));

  addrlen = sizeof(sin);

  if ((ns = accept(self->mo_fd, (struct sockaddr *) &sin, &addrlen)) < 0)
    {
      perror ("_tcp_AcceptClient:  accept");
    }
  else
    {
      DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
	    (stderr,
	     "_tcp_AcceptClient: new connection, on fd %ld,"
	     " from IP host %s, port %u.\n",
	     ns, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port)));

      new = (ilu_Transport) ilu_must_malloc (sizeof(struct _ilu_Transport_s));
      new->tr_class = _ilu_tcp_TransportClass();
      new->tr_packet = NIL;
      new->tr_fd = ns;
      new->tr_data = NewData (TCP_HOSTNAME(self->mo_data), TCP_PORT(self->mo_data), BUFFERSIZE);
      if ((flags = set_non_blocking(ns)) != 0)
	{
	  DEBUG((CONNECTION_DEBUG | TCP_DEBUG),
		(stderr,
		 "_tcp_AcceptClient:  Failed to set socket (descriptor"
		 " %ld, from %s) non-blocking.  Error \"%s\".\n",
		 ns, addrDescr, strerror(errno)));
        }

/* See above comment on similar code below */
#ifdef IPPROTO_TCP
#ifdef TCP_NODELAY
      {
	int one = 1;
	setsockopt(ns, IPPROTO_TCP, TCP_NODELAY,
		      (char *) &one, sizeof(one));
      }
#endif /* TCP_NODELAY */
#endif /* IPPROTO_TCP */     

    }
  return (new);
}

/*L1_sup < trmu*/
static ilu_Mooring _ilu_tcp_CreateMooring (ilu_private mooringInfo)
{
  TCPParms parms = TCPPARMS(mooringInfo);
  struct sockaddr_in sin;
  struct linger linger;
  SOCKET_SIZE_TYPE namelen;
  ilu_integer skt, on = 1;
  ilu_Mooring self;
  
  _ilu_AutoSetDebugLevel();
  _tcp_HandleSigPIPE();

  /* setup the new server */
  
  if ((skt = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    DEBUG((EXPORT_DEBUG | TCP_DEBUG),
	  (stderr, "_tcp_CreateMooring: socket failed:  %s.\n",
	   strerror(errno)));
    return (NIL);
  }
    
  setsockopt(skt, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
  linger.l_onoff = 0;
  setsockopt(skt, SOL_SOCKET, SO_LINGER,
		(char *) &linger, sizeof(struct linger));

  sin.sin_family = AF_INET;
  sin.sin_port = htons(TCP_PORT(parms));
  sin.sin_addr.s_addr = INADDR_ANY;	/* me, me! */

  if (bind(skt, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
    DEBUG((EXPORT_DEBUG | TCP_DEBUG),
	  (stderr, "_tcp_CreateMooring: bind to port %lu failed:  %s.\n",
	   TCP_PORT(parms), strerror(errno)));
    close (skt);
    return (NIL);
  }

  /* If servicePort was 0 then discover kernel allocated port */
  
  if (sin.sin_port == 0) {
    namelen = sizeof(sin);
    if (getsockname(skt, (struct sockaddr *) &sin, &namelen) < 0)
      {
	DEBUG((EXPORT_DEBUG | TCP_DEBUG),
	      (stderr, "_tcp_CreateMooring: getsockname failed:  %s.\n",
	       strerror(errno)));
	close (skt);
	return (NIL);
      }
  }

  if (listen(skt, 4) < 0) {
    DEBUG((EXPORT_DEBUG | TCP_DEBUG),
	  (stderr,
	   "_tcp_CreateMooring:  listen on port %u failed:  %s.\n",
	   ntohs(sin.sin_port), strerror(errno)));
    close (skt);
    return (NIL);
  }

  self = (ilu_Mooring) ilu_must_malloc (sizeof(struct _ilu_Mooring_s));

  self->mo_fd = skt;
  self->mo_accept_connection = _tcp_AcceptClient;
  TCP_PORT(parms) = ntohs(sin.sin_port);
  if (TCP_HOSTNAME(parms) != NIL)
    ilu_free(TCP_HOSTNAME(parms));
  self->mo_transportClass = _ilu_tcp_TransportClass();
  _ilu_AcquireMutex(ilu_trmu);
  TCP_HOSTNAME(parms) = _ilu_Strdup(_ilu_tcp_CurrentHostInetName());
  _ilu_ReleaseMutex(ilu_trmu);
  self->mo_data = (ilu_private) parms;

  return (self);
}

static void _ilu_tcp_CloseMooring(ilu_Mooring m)
{
  int res;

  ilu_UnregisterInputSource (m->mo_fd);
  res = close(m->mo_fd);
  ASSERT(res==0, buf,
	 (buf, "tcp_CloseMooring: res=%d, errno=%s", res,
	  strerror(errno)));
}

/*L1_sup < trmu*/
ilu_TransportClass _ilu_tcp_TransportClass (void)
{
  static ilu_TransportClass m = NIL;
  _ilu_AcquireMutex(ilu_trmu);
  if (m == NIL)
    {
      m = (ilu_TransportClass)
	  ilu_must_malloc(sizeof(struct _ilu_TransportClass_s));
      m->tc_pFD		= 1;
      m->tc_cFD		= 1;
      m->tc_timesout	= FALSE;
      m->tc_interpret_info	= _tcp_InterpretInfo;
      m->tc_form_info	= _tcp_FormHandle;
      m->tc_connect	= _tcp_Connect;
      m->tc_wait_for_input	= _tcp_WaitForInput;
      m->tc_create_mooring	= _ilu_tcp_CreateMooring;
      m->tc_close_mooring	= _ilu_tcp_CloseMooring;
      m->tc_close		= _tcp_Close;
      m->tc_n_data_present	= _tcp_NDataPresent;
      m->tc_write_bytes		= _tcp_WriteBytes;
      m->tc_get_output_buffer	= _tcp_GetOutputBuffer;
      m->tc_read_message	= _tcp_ReadMessage;
      m->tc_flush_output	= _tcp_FlushOutput;
    }
  _ilu_ReleaseMutex(ilu_trmu);
  return (m);
}

