/*
 * Copyright (c) 1991-1994 Xerox Corporation.  All Rights
 * Reserved.
 * 
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the
 * above copyright notice of Xerox Corporation and this paragraph.
 * Any distribution of this software must comply with all
 * applicable United States export control laws.  This software is
 * made available AS IS, and XEROX CORPORATION DISCLAIMS ALL
 * WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER PROVISION
 * CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM THE
 * SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN
 * IF XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */
/* $Id: type.c,v 1.36 1995/02/10 17:40:48 spreitze Exp $ */
/* Last edited by Mike Spreitzer February 9, 1995 3:12 pm PST */

#define _POSIX_SOURCE

#ifdef MACOS
#pragma segment ilu
#endif

#include "iluntrnl.h"

#include "object.h"
#include "server.h"
#include "type.h"
#include "vector.h"

#include <limits.h>		/* for ULONG_MAX */

/* L1, L2, Main unconstrained */

static struct _ilu_Method_s _ilu_InternalMethods[] = {
  {"GetTypes", 0xFF83,
  TRUE, FALSE, NIL, 0, _ilu_HandleGetTypes},
  {"RegisterGCInterest", 0xFF81,
  FALSE, FALSE, NIL, 0, _ilu_HandleGCInterestRegistration},
  {"UnregisterGCInterest", 0xFF82,
  FALSE, FALSE, NIL, 0, _ilu_HandleGCInterestDeregistration},
  {"Ping", 0xFF84,
  FALSE, FALSE, NIL, 0, _ilu_HandlePing}
};

ilu_Method      _ilu_GetTypesMethod = &_ilu_InternalMethods[0];
ilu_Method      _ilu_RegisterGCInterestMethod = &_ilu_InternalMethods[1];
ilu_Method      _ilu_UnregisterGCInterestMethod = &_ilu_InternalMethods[2];
ilu_Method      _ilu_PingMethod = &_ilu_InternalMethods[3];

static const    nInternal = sizeof(_ilu_InternalMethods)
/ sizeof(struct _ilu_Method_s);

struct _ilu_Class_s _ilu_rootClass_s = {
  "ILU-root",			/* name */
  "com.xerox.parc",		/* brand */
  "ilu:root",			/* unique_id */
  NIL,				/* singleton? */
  TRUE,				/* collectible? */
  NIL,				/* authentication */
  _ilu_InternalMethods,		/* methods */
  sizeof(_ilu_InternalMethods) / sizeof(struct _ilu_Method_s),
  /* method_count */
  0,				/* superclass_count */
  NIL,				/* superclass_ids */
  NIL,				/* superclasses */
  FALSE,			/* shown */
  FALSE				/* optional? */
};

const ilu_Class _ilu_rootClass = &_ilu_rootClass_s;
const ilu_Class ilu_rootClass = &_ilu_rootClass_s;

static struct _ilu_Method_s nometh = {NIL, ULONG_MAX, ilu_FALSE,
ilu_FALSE, NIL, ULONG_MAX, NULLFN};

ilu_Class 
ilu_ConsClass(
	      ilu_string name,	/* ILU name of class */
	      ilu_string brand,	/* brand on class */
	      ilu_string unique_id,	/* unique id for type graph
					 * of class */
	      ilu_string singleton,	/* pinfo if class is
					 * singleton */
	      ilu_boolean collectible,	/* class is collectible? */

	      ilu_string authentication,	/* non-NIL with name of
						 * authentication type,
						 * if any */
	      ilu_cardinal method_count,
	      ilu_cardinal scls_count,	/* number of superclasses */
	      ilu_string * scls_ids,	/* address of vector of
					 * strings */

	      ilu_boolean optional	/* may be NIL? -- IDL
					 * support */
)
{
  ilu_Class       ans = (ilu_Class) ilu_malloc(sizeof(struct _ilu_Class_s));
  int             i;
  if (ans == NIL)
    return ans;
  ans->cl_name = name;
  ans->cl_brand = brand;
  ans->cl_unique_id = unique_id;
  ans->cl_singleton = singleton;
  ans->cl_collectible = collectible;
  ans->cl_authentication = authentication;
  ans->cl_methods = (ilu_Method) ilu_malloc(method_count
				  * sizeof(struct _ilu_Method_s));
  if (ans->cl_methods == NIL) {
    ilu_free(ans);
    return NIL;
  }
  for (i = 0; i < method_count; i++)
    ans->cl_methods[i] = nometh;
  ans->cl_method_count = method_count;
  ans->cl_scls_count = scls_count;
  ans->cl_scls_ids = scls_ids;
  ans->cl_sclses = (ilu_Class *) ilu_malloc(scls_count
					    * sizeof(ilu_Class));
  if (ans->cl_sclses == NIL) {
    ilu_free(ans->cl_methods);
    ilu_free(ans);
    return NIL;
  }
  for (i = 0; i < scls_count; i++)
    ans->cl_sclses[i] = NIL;
  ans->cl_optional = optional != ilu_FALSE;
  return ans;
}

ilu_boolean 
ilu_SetClassMethod(
		   ilu_Class cl,
		   ilu_cardinal i,
		   ilu_string name,
		   ilu_cardinal id,
		   ilu_boolean cacheable,	/* functional? */
		   ilu_boolean asynchronous,	/* need to wait after
						 * calling? */
		   ilu_Exception * exceptionVector,	/* list of possible
							 * exceptions */
		   ilu_cardinal exceptionCount,	/* number of exceptions
						 * in list */

/* Main Invariant holds */
/*
 * L2    >=    {conn's callmu, iomu} before, L2 disjoint {conn's
 * callmu, iomu} after
 */
		   void (*stubproc) (ilu_Call)
)
{
  if (name == NIL || cl == NIL || i >= cl->cl_method_count ||
    cl->cl_methods == NIL || cl->cl_methods[i].me_name != NIL ||
      (exceptionCount > 0 && exceptionVector == NIL))
    return ilu_FALSE;
  cl->cl_methods[i].me_name = name;
  cl->cl_methods[i].me_id = id;
  cl->cl_methods[i].me_cacheable = cacheable;
  cl->cl_methods[i].me_asynchronous = asynchronous;
  cl->cl_methods[i].me_exceptionVector = exceptionVector;
  cl->cl_methods[i].me_exceptionCount = exceptionCount;
  cl->cl_methods[i].me_stubproc = stubproc;
  return ilu_TRUE;
}

/* L1 >= {otmu} */
/* L2, Main unconstrained */

static HashTable ClassNameTable = NIL;
/* Object type name -> object type */

static HashTable ClassIDTable = NIL;
/* Object type id -> object type */

static HashTable UnknownTypeIDs = NIL;
/* Object type id -> ID_DAG */

static HashTable UnlinkedClasses = NIL;
/*
 * Mapping from object type id to vector of object types waiting
 * for id to be registered.
 */

static void 
_ilu_RegisterClass(ilu_Class class)
{
  int             new;
  ilu_Class       p;
  register ilu_cardinal index;
  ilu_Vector      v;

  if (class != NIL AND((ilu_cardinal) (class->cl_singleton)) == 1) {
    fprintf(stderr, "\
Fatal ILU error:  Stubs for type \"%s\" were generated\n\
by a pre-1.6.4-p8 stubber.  Please re-stub, re-compile, and re-link.\n",
	    class->cl_name);
    exit(1);
  }
  DEBUG(OBJECT_DEBUG,
	(stderr, "Registering object type %s, id=%s.\n",
	 class->cl_name, class->cl_unique_id));
  
  if (ClassNameTable == NIL)
    ClassNameTable = (ilu_refany) _ilu_hash_MakeNewTable(
					      CLASS_HASHTABLESIZE,
			     (ilu_hashfnptr) _ilu_hash_HashString,
			 (ilu_compfnptr) _ilu_hash_StringCompare);
  if (ClassIDTable == NIL)
    ClassIDTable = (ilu_refany) _ilu_hash_MakeNewTable(
					      CLASS_HASHTABLESIZE,
			     (ilu_hashfnptr) _ilu_hash_HashString,
			 (ilu_compfnptr) _ilu_hash_StringCompare);
  if (UnlinkedClasses == NIL)
    UnlinkedClasses = (ilu_refany) _ilu_hash_MakeNewTable(
					      CLASS_HASHTABLESIZE,
			     (ilu_hashfnptr) _ilu_hash_HashString,
			 (ilu_compfnptr) _ilu_hash_StringCompare);

  if (_ilu_hash_FindInTable(ClassNameTable, class_name(class)) == NIL)
    _ilu_Assert(_ilu_hash_AddToTable(ClassNameTable, class_name(class),
				     class),
		"RegisterClass AddToTable ClassNameTable");
  new = NIL == _ilu_hash_FindInTable(ClassIDTable,
				      class_unique_id(class));
  if (new)
    _ilu_Assert(_ilu_hash_AddToTable(ClassIDTable,
				   class_unique_id(class), class),
		"RegisterClass AddToTable ClassIDTable");

  /*
   * this handles the case where additional class information is
   * loaded into a running program
   */
  if (new && UnknownTypeIDs != NIL
  && _ilu_hash_FindInTable(UnknownTypeIDs, class_unique_id(class))
      != NIL) {		/* This shouldn't happen! */
    ASSERT(0, buf,
	   (buf, "%s %s %s previously considered unknown.\n",
       "ilu_RegisterClass:  Config bug!  Registering object type",
	    class_name(class), class_unique_id(class)));
  }
  /*
   * this links up the superclass of "class" which have already
   * been registered
   */
  for (index = 0; index < class_superclass_count(class); index++) {
    if (class_superclass_id(class, index) != NIL) {
      p = (ilu_Class) _ilu_hash_FindInTable(ClassIDTable,
			       class_superclass_id(class, index));
      if (p == NIL) {
	v = (ilu_Vector) _ilu_hash_FindInTable(UnlinkedClasses,
			       class_superclass_id(class, index));
	if (v == NIL) {
	  v = _ilu_vector_new(1);
	  _ilu_Assert(_ilu_hash_AddToTable(UnlinkedClasses,
				class_superclass_id(class, index),
					   v),
		      "RegisterClass AddToTable UnlinkedClasses");
	}
	_ilu_vector_add(v, (ilu_refany) class);
      } else
	class_superclass(class, index) = p;
    }
  }

  /*
   * now link those classes which have been waiting for this
   * superclass
   */
  v = (ilu_Vector) _ilu_hash_FindInTable(UnlinkedClasses,
					 class_unique_id(class));
  if (v != NIL) {
    ilu_cardinal    size = _ilu_vector_size(v);
    ilu_Class      *elements = (ilu_Class *) _ilu_vector_elements(v);
    register ilu_cardinal j;

    _ilu_Assert(_ilu_hash_RemoveFromTable(UnlinkedClasses,
					  class_unique_id(class))
		== v,
		"RegisterClass RemoveFromTable UnlinkedClasses");

    for (index = 0; index < size; index++) {
      p = elements[index];
      for (j = 0; j < class_superclass_count(p); j++)
	if (strcmp(class_unique_id(class), class_superclass_id(p, j))
	    == 0) {
	  class_superclass(p, j) = class;
	  break;
	}
    }
    _ilu_vector_destroy(v, NULLFN);
  }
  return;
}

/* L1_sup < otmu */

static void 
PrintClassEntry(ilu_refany entry, ilu_refany junk)
{
  ilu_Class       p;

  if (entry != NIL) {
    p = (ilu_Class) entry;
    fprintf(stderr, "    %s (%s)   %p\n", p->cl_name, p->cl_unique_id, p);
  } else
    fprintf(stderr, "<null>\n");
}

void 
ilu_RegisterClass(ilu_Class class)
{
  _ilu_AcquireMutex(ilu_otmu);
  _ilu_RegisterClass(class);
  _ilu_ReleaseMutex(ilu_otmu);
  return;
}

void _ilu_EnumerateClasses (void (*proc) (ilu_Class c, ilu_refany rock), ilu_refany rock)
{
  _ilu_AcquireMutex(ilu_otmu);
  _ilu_hash_TableEnumerate (ClassNameTable, (void (*) (ilu_refany, ilu_refany)) proc, rock);
  _ilu_ReleaseMutex(ilu_otmu);
  return;
}

ilu_Class 
ilu_GetGcCallbackClass(void)
{
  _ilu_AcquireMutex(ilu_otmu);
  if (ClassIDTable == NIL ||
      _ilu_hash_FindInTable(ClassIDTable, _ilu_GcCallbackClass->cl_unique_id)
      == NIL)
    _ilu_RegisterClass(_ilu_GcCallbackClass);
  _ilu_ReleaseMutex(ilu_otmu);
  return _ilu_GcCallbackClass;
}

ilu_Class 
ilu_FindClassFromName(ilu_string classname)
{
  ilu_Class       c;
  _ilu_AcquireMutex(ilu_otmu);
  if ((_ilu_DebugLevel & OBJECT_DEBUG) == OBJECT_DEBUG) {
    fprintf(stderr, "ilu_FindClassFromName:  classname is %s, class table is %p:\n", classname,
	    ClassNameTable);
    _ilu_hash_TableEnumerate(ClassNameTable, (void (*) (ilu_refany, ilu_refany)) PrintClassEntry, NIL);
  }
  if (strcmp(classname, _ilu_rootClass->cl_name) == 0)
    c = _ilu_rootClass;
  else if (ClassNameTable != NIL)
    c = (ilu_Class) _ilu_hash_FindInTable(ClassNameTable, classname);
  else
    c = NIL;
  _ilu_ReleaseMutex(ilu_otmu);
  return (c);
}

ilu_Class 
ilu_FindClassFromID(ilu_string ID)
{
  ilu_Class       c;
  _ilu_AcquireMutex(ilu_otmu);
  if ((_ilu_DebugLevel & OBJECT_DEBUG) == OBJECT_DEBUG) {
    fprintf(stderr, "ilu_FindClassFromID:  id is %s, class table is %p:\n", ID,
	    ClassIDTable);
    _ilu_hash_TableEnumerate(ClassIDTable, (void (*) (ilu_refany, ilu_refany)) PrintClassEntry, NIL);
  }
  if (strcmp(ID, _ilu_rootClass->cl_unique_id) == 0)
    c = _ilu_rootClass;
  else if (ClassIDTable != NIL)
    c = (ilu_Class) _ilu_hash_FindInTable(ClassIDTable, ID);
  else
    c = NIL;
  _ilu_ReleaseMutex(ilu_otmu);
  return (c);
}

static ilu_boolean 
SeekAncestor(ilu_Class a, ilu_Class b)
{
  int             j;
  ilu_Class       aa;
  if (a == b)
    return (TRUE);
  for (j = 0; j < class_superclass_count(a); j++) {
    aa = class_superclass(a, j);
    if (aa != NIL && SeekAncestor(aa, b))
      return (TRUE);
  }
  return (FALSE);
}

ilu_boolean 
ilu_IsSubObjectType(ilu_Class a, ilu_Class b)
{
  ilu_boolean     ans;
  _ilu_AcquireMutex(ilu_otmu);
  ans = (b == _ilu_rootClass) || SeekAncestor(a, b);
  _ilu_ReleaseMutex(ilu_otmu);
  return (ans);
}

typedef struct _iluBuffer_s {
  /* L1 unconstrained */

  char           *buf;
  unsigned long   allocated;
  unsigned long   used;
}              *iluBuffer;
#define BUF_INCREMENT 100

/* L1 unconstrained */

static void 
addBytesToBuf(iluBuffer buf, char *string, ilu_cardinal len)
{
  if (buf->allocated - buf->used < len) {
    buf->allocated += BUF_INCREMENT;
    buf->buf = (char *) realloc(buf->buf, buf->allocated);
  }
  strncpy(buf->buf + buf->used, string, len);
  buf->used += len;
}

/* L1 >= {otmu} */

static void 
UnshowAncestors(ilu_Class c)
{
  int             i;
  c->cl_shown = FALSE;
  for (i = 0; i < class_superclass_count(c); i++)
    UnshowAncestors(class_superclass(c, i));
  return;
}

static void 
AddTypeName(ilu_Class type, iluBuffer buf)
{
  int             i;
  int             noco = 1;
  addBytesToBuf(buf, class_unique_id(type),
		strlen(class_unique_id(type)));
  if (type->cl_shown != TRUE) {
    type->cl_shown = TRUE;
    if (class_superclass_count(type) == 0)
      return;
    addBytesToBuf(buf, "(", 1);
    for (i = 0; i < class_superclass_count(type); i++) {
      if (noco)
	noco = 0;
      else
	addBytesToBuf(buf, ",", 1);
      AddTypeName(class_superclass(type, i), buf);
    }
    addBytesToBuf(buf, ")", 1);
  }
  return;
}

/* L1_sup < otmu */
static ilu_string 
ClassToString(ilu_Class class)
{
  struct _iluBuffer_s buf;
  buf.allocated = BUF_INCREMENT;
  buf.buf = ilu_malloc(BUF_INCREMENT);
  buf.used = 0;
  _ilu_AcquireMutex(ilu_otmu);
  UnshowAncestors(class);
  AddTypeName(class, &buf);
  addBytesToBuf(&buf, "\0", 1);	/* NUL-terminate */
  _ilu_ReleaseMutex(ilu_otmu);
  return (buf.buf);
}

/* L1 >= {otmu} */

typedef struct _ilu_ID_DAG_s ID_DAG_s, *ID_DAG;

typedef struct _ilu_ID_DAG_Cons {
  /* L1 >= {otmu} */

  ID_DAG          idd;
  struct _ilu_ID_DAG_Cons *next;
}               ID_DAG_Cons, *ID_DAG_List;

struct _ilu_ID_DAG_s {
  /* L1 >= {otmu} */

  char           *id;
  ilu_Class       it;		/* non-NIL if id registered */
  ID_DAG_List     supers;
  ilu_boolean     computed;	/* TRUE => mska, saak significant */
  ilu_Class       mska;		/* the one Most Specific Known
				 * Ancestor, or NIL if not
				 * well-defined */
  ilu_boolean     saak;		/* Some Ancestors Are Known */
  ilu_boolean     visited;	/* is this an ancestor of a counted
				 * type? */
  ID_DAG          anext;	/* threaded list of most spcfc
				 * known anc'rs */
  ID_DAG          aprev;	/* it's doubly-linked */
};

static ID_DAG_s mska_head;
/*
 * Head of threaded list of most spcfc known anc'rs. No one member
 * is an ancestor of another.
 */

/* L1_sup < otmu */
ilu_Class 
_ilu_FindMSKA(ilu_string tid)
{
  ID_DAG          n;
  ilu_Class       ans;
  _ilu_AcquireMutex(ilu_otmu);
  n = (ID_DAG) _ilu_hash_FindInTable(UnknownTypeIDs, tid);
  if (n != NIL && n->computed)
    ans = n->mska;
  else
    ans = NIL;
  _ilu_ReleaseMutex(ilu_otmu);
  return ans;
}

/*
 * spec ::== id [ "(" spec ("," spec)* ")" ] At entry, *names is at
 * the start of a spec; at exit, *names is at the first char past
 * that spec. Result is the ID_DAG having the id in the spec.
 * Caller owns storage for names.
 */
static ID_DAG 
ClassNamesToID_DAG(ilu_string * names)
{
  ID_DAG          n;
  char           *pl, *pr, *pc, *pe, cr;
  ilu_integer     len;

  DEBUG(OBJECT_DEBUG,
  (stderr, "(ClassNamesToID_DAG):  Called with \"%s\"\n", *names));

  if (*names == NIL OR ** names == '\0')
    return (NIL);
  len = strlen(*names);
  pl = strchr(*names, '(');
  pr = strchr(*names, ')');
  pc = strchr(*names, ',');
  if (pl == NIL)
    pl = (*names) + len;
  if (pr == NIL)
    pr = (*names) + len;
  if (pc == NIL)
    pc = (*names) + len;
  pe = (pr < pl) ? pr : pl;
  pe = (pc < pe) ? pc : pe;
  cr = *pe;
  *pe = '\0';
  n = (ID_DAG) _ilu_hash_FindInTable(UnknownTypeIDs, *names);
  if (n == NIL) {
    ilu_string      id = _ilu_Strdup(*names);
    n = (ID_DAG) ilu_malloc(sizeof(ID_DAG_s));
    _ilu_Assert(_ilu_hash_AddToTable(UnknownTypeIDs, id, n),
		"ClassNamesToID_DAG AddToTable UnknownTypeIDs");
    n->id = id;
    n->it = (ilu_Class) _ilu_hash_FindInTable(ClassIDTable, id);
    n->supers = NIL;
    n->mska = n->it;
    n->computed = n->saak = n->mska != NIL;
    n->anext = n->aprev = NIL;
    if (cr == '(') {
      ID_DAG_List    *pp = &(n->supers);
      *names = pl + 1;
      while ((**names) != ')') {
	ID_DAG          d2 = ClassNamesToID_DAG(names);
	ID_DAG_List     l = (ID_DAG_List) ilu_malloc(sizeof(ID_DAG_Cons));
	l->idd = d2;
	l->next = NIL;
	*pp = l;
	pp = &(l->next);
	if ((**names) == ',')
	  (*names)++;
      }
      (*names)++;
    } else
      *names = pe;
  } else if (pl < pc) {		/* check for consistency */
    ID_DAG_List     l = n->supers;
    *names = pl + 1;
    while ((**names) != ')') {
      ID_DAG          d2 = ClassNamesToID_DAG(names);
      ASSERT(d2 == l->idd, buf,
	     (buf, "ClassNamesToID_DAG:  %s %s: %s vs. %s.\n",
	      "Disagreement on superclasses of", n->id, d2->id,
	      l->idd->id));
      l = l->next;
      if ((**names) == ',')
	(*names)++;
    }
    (*names)++;
  } else
    *names = pe;
  *pe = cr;
  return (n);
}

static void 
ClearVisited(ID_DAG t)
{
  ID_DAG_List     p;
  t->visited = FALSE;
  for (p = t->supers; p != NIL; p = p->next)
    ClearVisited(p->idd);
  return;
}

/*
 * Before and after: every node in the mska list is visited; for
 * each visited node n: 1. if n is known then n is an ancestor of a
 * node in the mska list; 2. every ancestor of n is visited.
 */

/*
 * As above, plus... After: all of t's ancestors are visited, and
 * none are in the list.
 */
static void 
MarkAncestors(ID_DAG t)
{
  ID_DAG_List     p;
  for (p = t->supers; p != NIL; p = p->next) {
    ID_DAG          u = p->idd;
    if (u->anext != NIL) {
      /*
       * must already be visited, and have no proper ancestors in
       * the mska list; remove from list.
       */
      u->aprev->anext = u->anext;
      u->anext->aprev = u->aprev;
    } else if (!(u->visited && (u->it != NIL))) {
      MarkAncestors(u);
      u->visited = TRUE;
    }
  }
  return;
}

static void 
FindMostSpecificType(ID_DAG t)
{
  ID_DAG_List     p;
  DEBUG(OBJECT_DEBUG,
	(stderr, "(FindMostSpecificType):  at %s, %s visited.\n",
	 t->id, (t->visited == FALSE) ? "not" : ""));
  if (t->visited == TRUE)
    return;
  if (t->it != NIL) {
    MarkAncestors(t);
    t->anext = mska_head.anext;
    t->aprev = &mska_head;
    t->anext->aprev = t;
    t->aprev->anext = t;
  } else {
    for (p = t->supers; p != NIL; p = p->next)
      FindMostSpecificType(p->idd);
  }
  t->visited = TRUE;
  return;
}

/* L1_sup < otmu */
static ilu_Class 
StringToClass(ilu_string s)
{
  ID_DAG          t;
  char           *q = s;
  ilu_Class       ans;

  if (s == NIL)
    return (NIL);
  _ilu_AcquireMutex(ilu_otmu);
  if (UnknownTypeIDs == NIL)
    UnknownTypeIDs = (ilu_refany) _ilu_hash_MakeNewTable(
					      CLASS_HASHTABLESIZE,
			     (ilu_hashfnptr) _ilu_hash_HashString,
			 (ilu_compfnptr) _ilu_hash_StringCompare);
  t = ClassNamesToID_DAG(&q);
  if (t == NIL)
    return (NIL);
  DEBUG(OBJECT_DEBUG,
  (stderr, "(StringToClass):  Converted names <%s> to DAG.\n", s));
  if (t->computed) {
    DEBUG(OBJECT_DEBUG, (stderr, "(StringToClass):  Old problem.\n"));
  } else {
    ClearVisited(t);
    mska_head.anext = mska_head.aprev = &mska_head;
    FindMostSpecificType(t);
    t->saak = mska_head.anext != &mska_head;
    if (t->saak && (mska_head.anext->anext == &mska_head))
      t->mska = mska_head.anext->it;
    else
      t->mska = NIL;
    t->computed = TRUE;
  }
  ans = t->mska;
  _ilu_ReleaseMutex(ilu_otmu);
  if (ans == NIL) {
    fprintf(stderr,
	    "(StringToClass):  %s ancestors of type <%s> are known!  Unable to resolve type.\n",
	    t->saak ? "Multiple" : "No", s);
  } else {
    DEBUG(OBJECT_DEBUG,
	  (stderr, "(StringToClass):  Found class %s.\n",
	   class_name(ans)));
  }
  return (ans);
}

/* Main Invariant holds; L2 otherwise unconstrained */

static ilu_ProtocolException 
ObtainTypes(ilu_Object o, ilu_string * types, ilu_cardinal * typeslen)
{
  ilu_Call        call;
  ilu_cardinal	  reqSize;
  ilu_cardinal    estatus = 0;
  ilu_ProtocolException internal;
  ilu_Class       pclass = object_class(o);
  ilu_Server      s = object_server(o);

  DEBUG(OBJECT_DEBUG, (stderr, "_ilu_FindClassViaRPC:  object %p...\n", o));

  if (class_singleton(pclass)) {
    DEBUG(OBJECT_DEBUG,
	  (stderr,
	"%s %s is singleton, not attempting GetTypes RPC call.\n",
	   "_ilu_FindClassViaRPC:  pclass", class_name(pclass)));
    return (ilu_ProtocolException_NoSuchMethodOnClass);
  }
  if ((call = ilu_BeginCall(s)) == NIL)
    return (ilu_ProtocolException_Unknown);
  _ilu_AcquireMutex(server_lock(s));
  reqSize = ilu_SizeOfObjectID(call, o, TRUE, _ilu_rootClass);
  _ilu_ReleaseMutex(server_lock(s));
  if (NOT ilu_BeginRequest(call, _ilu_rootClass, _ilu_GetTypesMethod, reqSize))
    internal = ilu_ProtocolException_Unknown;
  else {
    ilu_EnterServer(s, object_class(o));
    ilu_OutputObjectID(call, o, TRUE, _ilu_rootClass);
    ilu_FinishRequest(call);
    internal = ilu_GetReply(call, &estatus);
    if (internal == ilu_ProtocolException_Success AND estatus == 0)
      /* successful reply */
      ilu_InputString(call, types, typeslen, 0xFFFF, FALSE);
  }
  ilu_FinishCall(call);
  return internal;
}

ilu_Class 
_ilu_FindClassViaRPC(ilu_Object o)
{
  ilu_string      types = NIL;
  ilu_ProtocolException internal;
  ilu_cardinal    typeslen;
  ilu_Class       c = NIL;

  internal = ObtainTypes(o, &types, &typeslen);

  if (types == NIL)
    DEBUG(OBJECT_DEBUG,
	  (stderr, "%s  Status of RPC call is %u.\n",
	   "_ilu_FindClassViaRPC:  no types.", internal));
  else
    DEBUG(OBJECT_DEBUG,
	(stderr, "_ilu_FindClassViaRPC:  typestring is <%s>...\n",
	 types));

  if (types != NIL) {
    c = StringToClass(types);
    FREETOKEN(types);
  }
  DEBUG(OBJECT_DEBUG,
	(stderr, "_ilu_FindClassViaRPC:  class is \"%s\".\n",
	 (c == NIL) ? "*unknown*" : class_name(c)));
  return (c);
}

/*
 * L2    >=    {conn's callmu, iomu} before, L2 disjoint {conn's
 * callmu, iomu} after
 */
void 
_ilu_HandleGetTypes(ilu_Call call)
{
  ilu_Object      disc;
  ilu_string      names = NIL;
  ilu_Error       err;

  ilu_InputObjectID(call, &disc, TRUE, _ilu_rootClass);
  if (disc != NIL) {
    err = _ilu_DeltaHolds(disc, 1);
    ILU_MUST_BE_SUCCESS(err);
    ilu_ExitServer(object_server(disc), _ilu_rootClass);
  }
  ilu_RequestRead(call);

  if (disc != NIL && object_class(disc) != NIL)
    names = ClassToString(object_class(disc));

  if (names == NIL) {
    if (ilu_BeginException(call, 0,
			   ilu_ProtocolException_GarbageArguments))
      ilu_FinishException(call);
  } else {
    ilu_cardinal    len = strlen(names);

    if (ilu_BeginReply(call, FALSE,
		       ilu_SizeOfString(call, names, len, 0xFFFF,
					FALSE))) {
      ilu_OutputString(call, names, len, 0xFFFF, FALSE);
      ilu_FinishReply(call);
    }
  }

  if (disc != NIL) {
    ilu_Server      s = object_server(disc);
    ilu_Class       cl = object_class(disc);
    ilu_EnterServer(s, cl);
    err = _ilu_DeltaHolds(disc, -1);
    ILU_MUST_BE_SUCCESS(err);
    ilu_ExitServer(s, cl);
  }
  return;
}
