#|
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: ilu-server.lisp,v 1.21 1995/02/21 01:09:49 janssen Exp $
|#

(in-package :ilu)		; in this package

(eval-when (compile load eval)
  (export '(signal-exception
	    kernel-server
	    server-c-server
	    server-unix-port
	    server-id
	    define-server-class
	    )))
	  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Actual code for dealing with server objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *default-server* nil)

(defclass kernel-server ()
  ((unix-port :reader server-unix-port :initarg :port :initform 0)
   (c-server :initform nil :reader server-c-server)
   (id :initarg :id :initform nil :reader server-id))
  )

(defclass stateless-kernel-server (kernel-server) ())

(defmethod has-state ((self kernel-server))
  t)

(defmethod has-state ((self stateless-kernel-server))
  nil)

(defun create-true-c-server (server-id unix-ip-port)
  (ensure-initialized)
  (let* ((c-server (ilu_create-true-server server-id 0))
	 (port (open-port c-server unix-ip-port)))
    (set-server-default-port c-server port)
    c-server))

(defmethod initialize-instance ((self kernel-server) &key)
  (call-next-method)
  (unless (server-id self)
    (setf (slot-value self 'id)
      (generate-server-id)))
  (unless (server-unix-port self)
    (error "No UNIX port specified for export of service ~s.~%" self))
  (setf (slot-value self 'c-server)
    (create-true-c-server (server-id self) (server-unix-port self)))
  (setf *default-server* self)
  )

(defmethod print-object ((self ilu:kernel-server) stream)
  (format stream "#<~s ~s>"
	  (type-of self)
	  (server-id self)))

(defun ensure-server ()
  (unless *default-server*
    (setf *default-server* (make-instance 'kernel-server :port 0)))
  *default-server*)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Code for dealing with true objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; before: Main invariant holds
;; after:  Inside(server,pclass)
(defun find-or-create-true-kernel-object (ih server pclass lspo)
  (let ((kserver (server-c-server server)))
    (ilu_enter-server kserver pclass)
    (ilu_find-or-create-true-object
     ih kserver pclass
     (if lspo (register-lisp-object lspo :reftype :strong)
       0))))

(defclass ilu-true-object (ilu-object) ()); only appears in servers

(defparameter *object-id-counter* 1)

(defun create-default-id ()
  (setf *object-id-counter* (1+ *object-id-counter*))
  (format nil "~a" *object-id-counter*))

;; before: Inside(kernel-object->ob_server, kernel-object->ob_class)
;; after:  Main Invariant holds
(defmethod setup-object-links ((lisp-obj ilu-true-object)
			       kernel-object &optional &key (server nil))
  (setf (ilu-server lisp-obj) (server-c-server server))
  (setf (ilu-class lisp-obj) (ilu_ilu-class kernel-object))
  (setf (ilu-cached-kernel-obj lisp-obj) kernel-object)
  (ilu_exit-server (ilu-server lisp-obj) (ilu-class lisp-obj)))

(defmethod initial-pointer-reftype ((self ilu-true-object))
  :strong)

(defmethod initialize-instance :after ((self ilu-true-object)
				       &key
				       (ilu-kernel-server nil)
				       (ilu-instance-handle nil)
				       &allow-other-keys)
  (let ((server (or ilu-kernel-server (ensure-server))))
    (setup-object-links
     self
     (find-or-create-true-kernel-object
      (or ilu-instance-handle (create-default-id))
      server (find-ilu-class-record self) self)
     :server server)))

(defmethod object-write-prep (call (self ilu-true-object) pclass)
  (declare (ignore pclass call))
  (when (null (ilu-cached-kernel-obj self))
    (let ((server (or (ilu-server self) (ensure-server))))
      (setup-object-links
       self
       (find-or-create-true-kernel-object
	(or (slot-value 'ilu-instance-handle self) (create-default-id))
	server (find-ilu-class-record self) self)
       :server server))))

(defvar *gc-callback-object* nil)	;; hold onto this obj to prevent GC

(defclass ilu-gc-callback-class (ilu-true-object) ())

(defmethod find-ilu-class-record ((self ilu-gc-callback-class))
  (ilu_gc-callback-class))

(defun initialize-gc-callback ()
  (unless *gc-callback-object*
    (setq *gc-callback-object* (make-instance 'ilu-gc-callback-class))
    (let ((gc-callback-kobj (kernel-obj *gc-callback-object*)))
      (ilu_set-gc-callback gc-callback-kobj)
      (ilu_exit-server (ilu_ilu-server gc-callback-kobj)
		       (ilu_ilu-class gc-callback-kobj))
      (values))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Server functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *debug-uncaught-conditions* nil
  "If T, causes break to debugger.  Default to NIL.")

(defvar *trace-calls* nil
  "If T, causes message to be printed on each request serviced.  Default NIL.")

(defun handle-uncaught-condition (call cond lspo conn method-name sn)
  (declare (ignore conn))
  (unwind-protect
      (if *debug-uncaught-conditions*
	  (invoke-debugger cond)
	(warn "Uncaught condition in call ~D: method ~A on ~A.~% Condition: ~A"
	      sn method-name lspo cond))
    (begin-exception call 0 +protocol-exception-unknown+)	;; send "unknown error"
    (finish-exception call)))

(defun get-discriminator (call pclass)
  (multiple-value-bind (ilu-obj class-record) (object-id-read call pclass 1)
    (if (and ilu-obj (not (zerop ilu-obj)))
	(ilu-object->instance
	 (gethash class-record *class-table*
		  (gethash pclass *class-table*))
	 ilu-obj)
      (error "Couldn't read kernel object~%"))))

(defun handle-input-from-connection (conn server)
  (declare (ignore server))
  (multiple-value-bind (sn call type method-name)
      (ilulisp_get-request conn)
    (if (= sn -2)
	(progn
	  (warn "Problem reading or interpreting request.  Dropping connection 0x~x.~%" conn)
	  nil)
      (if (= sn -1)
	  (progn
	    (warn "Client on connection 0x~x dropped connection.~%" conn)
	    nil)
	(if (= sn 0)
	    ;; already handled internally
	    t
	  (let ((lspo (get-discriminator call type)))
	    (when *trace-calls*
	      (format t "~d:  ~s, ~a~%" sn lspo method-name))
	    (handler-bind ((serious-condition
			    #'(lambda (cond)
				(handle-uncaught-condition call
							   cond lspo
							   conn method-name sn)
				(return-from handle-input-from-connection t))))
;	      (format t "calling method ~s on object ~s with class precedence list ~s.~%"
;		      method-name lspo (clos:class-precedence-list (class-of lspo)))
	      (call-server-stub-for-method-name call (intern method-name :keyword) lspo))
	    t))))))	    
  
(defun handle-new-connection (server port)
  (let ((conn (handle-ilu-connection port)))
    (unless (zerop conn)
      (setup-watch-connection #'handle-input-from-connection
			      conn server))))

(defun open-port (server &optional (ip-port 0))
  (if (zerop server)
      (error "Can't open port on nil server~%"))
  (let* ((s-port (create-port server "sunrpc_" (format nil "tcp_localhost_~a" ip-port)))
	 (fd (and s-port (file-descriptor-of-mooring-of-port s-port)))
	 (sid (id-of-kernel-server server)))
    (if (zerop s-port)
	(error "Can't open RPC port on UNIX port ~a for server ~s.~%" ip-port sid)
      (if (or (null fd) (< fd 0))
	  (error "Can't obtain file descriptor of mooring of port ~s.~%" s-port)
	(progn
	  (set-server-default-port server s-port)
	  (setup-new-connection-handler #'handle-new-connection server s-port)
	  s-port)))))
