/*
 ** Copyright (c) 1991, 1992, 1993 Xerox Corporation. All Rights Reserved.
 **
 ** Unlimited use, reproduction, and distribution of this software is
 ** permitted. Any copy of this software must include both the above
 ** copyright notice of Xerox Corporation and this paragraph. Any
 ** distribution of this software must comply with all applicable United
 ** States export control laws. This software is made available AS IS,
 ** and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 ** INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 ** AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 ** PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 ** THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 ** CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 ** XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
/* Last edited by Mike Spreitzer December 6, 1994 3:38 pm PST */


#include <stdlib.h>
#include <stdarg.h>
#include "cstubber.h"

static boolean makedir (char *);
static void usage(void);

static boolean GenerateTrue = FALSE;
static boolean GenerateCommon = FALSE;
static boolean GenerateSurrogate = FALSE;
static boolean GenerateHeaders = FALSE;

static list Includes = NULL;

extern cardinal MethodRecordID;
char *NamesFile = NULL;
static char *OutputDirectory = ".";
static char *ProgramName = "c-stubber";

static void generateCBoilerplate (FILE *file, Interface parse)
{
  static char *prefixes[2] = { " *", " *" };

  fprintf (file, "/*\n");
  iluparser_GenerateBoilerplate (file, parse, ProgramName, prefixes);
  fprintf (file, " */\n\n");
}

static void generateCommon (Interface parse)
{
  FILE *file;
  char filename[1000];

  MethodRecordID = 0;
  sprintf(filename, "%s/%s-common.c", OutputDirectory, c_interface_name(parse));
  if ((file = fopen(filename, "w")) == NULL)
    fatal("Couldn't open client stubs output file %s.\n", filename);
  printf("common code for interface %s to %s...\n", c_interface_name(parse), filename);
  setFile(file);
  generateCBoilerplate (file, parse);
  generateCommonCode(parse, file);
  fclose (file);
}

static void generateSurrogate (Interface parse)
{
  FILE *file;
  char filename[1000];

  MethodRecordID = 0;
  sprintf(filename, "%s/%s-surrogate.c", OutputDirectory, c_interface_name(parse));
  if ((file = fopen(filename, "w")) == NULL)
    fatal("Couldn't open surrogate stubs output file %s.\n", filename);
  printf("code for surrogate stubs of interface %s to %s...\n", c_interface_name(parse), filename);
  setFile(file);
  generateCBoilerplate (file, parse);
  generateClientCode(parse, file);
  fclose (file);
}

static void generateTrue (Interface parse)
{
  FILE *file;
  char filename[1000];

  sprintf(filename, "%s/%s-true.c", OutputDirectory, c_interface_name(parse));
  if ((file = fopen(filename, "w")) == NULL)
    fatal("Couldn't open true stubs output file %s.\n", filename);
  printf("code for true stubs of interface %s to %s...\n", c_interface_name(parse), filename);
  setFile(file);
  generateCBoilerplate (file, parse);
  generateServerCode(parse, file);
  fclose(file);
}

static void generateHeaders(Interface parse)
{
  FILE *file;
  char filename[1000];

  sprintf(filename, "%s/%s.h", OutputDirectory, c_interface_name(parse));
  if ((file = fopen(filename, "w")) == NULL) {
    fatal("Couldn't open output file %s.\n", filename);
  }
  printf("header file for interface %s to %s...\n", c_interface_name(parse), filename);
  setFile(file);
  generateCBoilerplate (file, parse);
  generateCHeaders(parse, file);
  fclose(file);
}

static void generateStubs (Interface parse)
{
  if (!makedir(OutputDirectory))
    fatal("Couldn't create output directory %s.\n", OutputDirectory);
  if (GenerateHeaders)
    generateHeaders (parse);
  if (GenerateCommon)
    generateCommon (parse);
  if (GenerateSurrogate)
    generateSurrogate (parse);
  if (GenerateTrue)
    generateTrue (parse);
}

#if (defined(_IS_BSD) || defined(_IS_POSIX))
#include <sys/stat.h>
#ifdef _IS_POSIX
#include <unistd.h>
#endif
#endif

static boolean makedir (char *dirname)
{
#if (defined(_IS_BSD) || defined(_IS_POSIX))

  struct stat s;

  if (stat(dirname, &s) == 0)
    {
      if (S_ISDIR(s.st_mode))
	{

#ifdef _IS_POSIX

	  return (access (dirname, W_OK) == 0);

#else /* not POSIX */

	  return (TRUE);

#endif
	  
	}
      return (FALSE);
    }
  else
    {
      char buf[1000];
      sprintf(buf, "mkdir %s", dirname);
      return((system(buf) >> 8) == 0);
    }

#else

  fprintf (stderr, "%s:  Can't test directory \"%s\".\n",
	   ProgramName, dirname);
  return (TRUE);

#endif
}

void main(int ac, char **av, char **envp)
{
  list s;
  char **interfacename;

  if (ac < 2) {
    usage();
    exit(1);
  }
  Includes = new_list();

  if ((ProgramName = iluparser_GetProgramName(*av)) == NULL)
    ProgramName = "c-stubber";
  ac--;
  av++;
  while(*av[0] == '-') {
    if (strcmp(*av, "-dir") == 0)
      OutputDirectory = *++av;
    else if (strcmp(*av, "-I") == 0)
      list_insert(Includes, *++av);
    else if (strcmp(*av, "-true") == 0)
      GenerateTrue = TRUE;
    else if (strcmp(*av, "-renames") == 0)
      NamesFile = *++av;
    else if (strcmp(*av, "-surrogate") == 0)
      GenerateSurrogate = TRUE;
    else if (strcmp(*av, "-headers") == 0)
      GenerateHeaders = TRUE;
    else if (strcmp(*av, "-common") == 0)
      GenerateCommon = TRUE;
    else {
      usage();
      fatal("%s: Invalid switch \"%s\".\n", ProgramName, *av);
    }
    ac--;
    av++;
  }
  if (!(GenerateTrue || GenerateHeaders || GenerateSurrogate || GenerateCommon)) {
    /* do everything, if nothing is specified */
    GenerateCommon = TRUE;
    GenerateHeaders = TRUE;
    GenerateSurrogate = TRUE;
    GenerateTrue = TRUE;
  }
  if (NamesFile != NULL) {
    if (!ReadSynonyms(NamesFile)) {
      fatal("Couldn't read names file %s.\n", NamesFile);
    }
  }

  iluparser_RegisterInterfaceDirectories(Includes);

  for(interfacename = av; *interfacename != NULL; interfacename++) {
    if ((s = ParseFile(*interfacename)) == NULL) {
      fatal("Couldn't find or parse %s.\n", *interfacename);
    }
    list_enumerate(s, (void (*)(refany, refany)) generateStubs, NULL);
  }
  exit(0);
}

static void usage()
{
  fprintf(stderr, "Usage: %s [-I ISLDIRECTORY] [-dir OUTPUTDIRECTORY] [-true] [-surrogate] [-headers] [-common] [-renames NAMESFILE] ISLFILE [ISLFILE ...]\n", ProgramName);
}

