/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: client.c,v 1.28 1994/11/20 22:43:53 janssen Exp $
*/
/* Last edited by Mike Spreitzer November 10, 1994 4:29 pm PST */

#include "m3.h"

static cardinal MethodIndex = 0;

typedef struct _class_node_s {
  Type type;
  boolean leaf;
  list parents;		/* list of ClassNode */
  list children;	/* list of ClassNode */
} ClassNode;

static list AllNodes = NULL;

static ClassNode *newClassNode (Type t)
{
  ClassNode *n = (ClassNode *) iluparser_Malloc(sizeof(ClassNode));

  n->type = t;
  n->leaf = FALSE;
  n->parents = new_list();
  n->children = new_list();

  if (AllNodes == NULL)
    fprintf (stderr, "Null AllNodes\n");
  else
    list_insert(AllNodes, n);

  return (n);
}

static boolean CompareTypes (refany element, refany rock)
{
  ClassNode *c = (ClassNode *) element;
  return (c->type == rock);
}

ClassNode *FindPreviouslyCreatedNode (Type t)
{
  if (AllNodes == NULL)
    return (NULL);
  else
    return ((ClassNode *) list_find (AllNodes, CompareTypes, t));
}

static void AddNode (refany element, refany rock);

static ClassNode * AddNodeWork (refany element, refany rock)
{
  Type            ut, st = (Type) element;
  ClassNode      *tn, *cn = (ClassNode *) rock;

  ut = ur_type(st);
  if (ut->marked)
    tn = FindPreviouslyCreatedNode(ut);
  else {
    tn = newClassNode(ut);
    list_enumerate(class_object(ut)->superclasses, AddNode, tn);
    ut->marked = TRUE;
  }
  if (cn != NULL) {
    list_insert(cn->parents, tn);
    list_insert(tn->children, cn);
  }
  return (tn);
}

static void AddNode (refany element, refany rock)
{
  AddNodeWork(element, rock);
  return;
}

static void freeEdges (ClassNode *n)
{
  list_clear (n->parents, FALSE);
  iluparser_Free(n->parents);
  list_clear (n->children, FALSE);
  iluparser_Free(n->children);
}

static void MarkLeaves2 (ClassNode *n, int *c)
{
  int count = 0;
  if (n->type->marked)
    return;
  else
    *c += 1;
  n->type->marked = TRUE;
  list_enumerate (n->children, (EnumProc) MarkLeaves2, &count);
  if (count == 0)
    n->leaf = TRUE;
  else
    n->leaf = FALSE;
}

static void MarkLeaves (ClassNode *r, Type t)
{
  int             count;
  if (list_size(r->parents) == 0 /* root */
      && !r->type->marked)
    MarkLeaves2(r, &count);
}

static void FindFollowers (ClassNode *r, list l)
{
  if (r->leaf)
    list_insert(l, r);
}

list GetFollowers (Type t)
{
  ClassNode *n;
  list l;

  if (t == NULL OR (type_kind(t) != object_Type))
    return (NULL);

  UnmarkSupertypes(t);

  if (AllNodes == NULL)
    AllNodes = new_list();
  else
    {
      list_enumerate (AllNodes, (EnumProc) freeEdges, NULL);
      list_clear (AllNodes, TRUE);
    }
  n = AddNodeWork (t, NULL);	/* actually builds graph */
  UnmarkSupertypes (t);
  list_enumerate (AllNodes, (EnumProc) MarkLeaves, t);
  l = new_list();	/* list of follower nodes */
  list_enumerate (n->parents, (EnumProc) FindFollowers, l);

  return (l);
}

static void GenerateFollowerType (ClassNode *f, Context context)
{
  Type follower = ur_type(f->type);
  char *un;
  char *tn;

  if (context->class == follower)
    return;

  un = TwoTypeName(context->class, follower);
  tn = M3_SHORT_TYPE_NAME(context->class);

  fprintf(context->file,
	  "TYPE Following__%s = %sSurrogate__%s OBJECT\n",
	  un, SupportPrefix(follower, context),
	  M3_SHORT_TYPE_NAME(follower));
  fprintf (context->file, "    ilu_the_leader : Lead__%s;\n", tn);
  fprintf (context->file, "  OVERRIDES\n");
  fprintf (context->file, "    ILU_Get_Type := ILU_Get_Type_%s;\n", tn);
  fprintf(context->file, "    ILU_Qua_Type := Qua_From_Following__%s\n",
	  un);
  fprintf (context->file, "  END;\n\n");

  fprintf(context->file,
	  "PROCEDURE Qua_From_Following__%s (self: Following__%s; ot: Ilu.ObjectType) : Ilu.Object =\n",
	  un, un);
  fprintf (context->file, "  BEGIN\n");
  fprintf(context->file,
	  "    RETURN Qua_From_Lead_%s(self.ilu_the_leader, ot)\n",
	  tn);
  fprintf (context->file, "  END Qua_From_Following__%s;\n\n", un);
}

static void generate_method_assignment (refany element, refany rock)
{
  Procedure       m = (Procedure) element;
  Context         context = (Context) rock;
  fprintf(context->file, "    %s := ", M3_SHORT_METHOD_NAME(m));
  fprintf(context->file, "%sClient__%s_%s;\n",
	  SupportPrefix(m->object, context),
	  M3_SHORT_TYPE_NAME(m->object),
	  M3_SHORT_METHOD_NAME(m));
}

static void GenerateMethodOverrides (Type st, Context context)
{
  Class od;
  Type t = ur_type(st);

  if (st == NULL OR t == NULL
      OR type_kind(t) != object_Type
      OR (od = class_object(t)) == NULL)
    return;
  t->marked = TRUE;
  if (list_size(od->superclasses) > 0)
    GenerateMethodOverrides ((Type) list_car(od->superclasses), context);
  list_enumerate (od->methods, generate_method_assignment, context);
}

static void GenerateFollowerSlot (ClassNode *f, Context context)
{
  Type follower = f->type;

  if (context->class == follower)
    return;

  fprintf (context->file, "    ilu_following__%s_%s: Following__%s;\n",
	   m3_interface_name(ur_type(follower)->interface),
	   M3_SHORT_TYPE_NAME(follower),
	   TwoTypeName(context->class, follower));
}

static void InitializeFollowerSlot (ClassNode *f, Context context)
{
  Type follower = f->type;

  if (context->class == follower)
    return;

  fprintf (context->file, ", ilu_following__%s_%s := NEW(Following__%s)",
	   m3_interface_name(ur_type(follower)->interface),
	   M3_SHORT_TYPE_NAME(follower),
	   TwoTypeName(context->class, follower));
}

static void InitFollowers (ClassNode *f, Context context)
{
  Type follower = f->type;
  Type uf = ur_type(follower);

  if (context->class == follower)
    return;

  fprintf(context->file,
	  "    x.ilu_following__%s_%s.ilu_the_leader := x;\n",
	  m3_interface_name(ur_type(follower)->interface),
	  M3_SHORT_TYPE_NAME(follower));
  fprintf(context->file,
  "    %sInit_Surrogate__%s(x.ilu_following__%s_%s, server);\n",
	  SupportPrefix(uf, context),
	  M3_SHORT_TYPE_NAME(uf),
	  m3_interface_name(ur_type(follower)->interface),
	  M3_SHORT_TYPE_NAME(follower));
}

static void ClimbQuaTree (ClassNode *follower, Context context)
{
  ClassNode      *n;
  Type            t = ur_type(follower->type);

  for (n = follower;
       n != NULL && (!n->type->marked);
       n = list_car(n->parents)) {
    Type            un = ur_type(n->type);
    fprintf(context->file, "    ELSIF ot = %sObjectType_%s THEN\n",
	    SupportPrefix(un, context),
	    M3_SHORT_TYPE_NAME(un));
    fprintf(context->file, "      RETURN self.ilu_following__%s_%s\n",
	  m3_interface_name(t->interface), M3_SHORT_TYPE_NAME(t));
    n->type->marked = TRUE;
  }
}

static void GenerateObjectSupport (Type type, Context context)
{
  char           *tn;
  char           *stn;
  list            followers;
  Type            n;

  tn = M3_TYPE_NAME(type);
  stn = M3_SHORT_TYPE_NAME(type);

  fprintf(context->file, "PROCEDURE ILU_SBH_To_%s", stn);
  fprintf(context->file, "(sbh : TEXT; mstid: TEXT) : %s\n",
	  stn);
  fprintf(context->file,
	  "  RAISES {IluBasics.Failed, Thread.Alerted} =\n");
  if (type->supertype != NULL) {
    Type            u = under_type(type);
    fprintf(context->file, "  BEGIN\n");
    fprintf(context->file, "    RETURN %s.ILU_SBH_To_%s(sbh, mstid)\n",
	    m3_interface_name(u->interface),
	    M3_SHORT_TYPE_NAME(u));
    fprintf(context->file, "  END ILU_SBH_To_%s;\n\n", stn);
    return;
  }
  fprintf(context->file, "  VAR x := IluRuntime.M3ObjectFromSbh");
  fprintf(context->file, "(sbh, ObjectType_%s, mstid);\n",
	  stn);
  fprintf(context->file, "  BEGIN\n");
  fprintf(context->file, "    TYPECASE x OF\n");
  fprintf(context->file, "      | %s (xx) => RETURN xx;\n", tn);
  fprintf(context->file, "      ELSE RAISE IluBasics.Failed(NEW(IluBasics.Failure,");
  fprintf(context->file, " info := \"sbh \" & sbh & \" leads to non-%s object\"))\n", tn);
  fprintf(context->file, "    END (* typecase *);\n");
  fprintf(context->file, "  END ILU_SBH_To_%s;\n\n", stn);
  
  followers = GetFollowers(type);
  context->class = type;

  fprintf(context->file, "REVEAL Surrogate__%s = %s BRANDED OBJECT\n",
	  stn, stn);
  fprintf(context->file, "    server: Ilu.SurrogateServer := NIL;\n");
  fprintf(context->file, "  OVERRIDES\n");
  GenerateMethodOverrides(type, context);
  fprintf(context->file,
	  "    ILU_Get_Server := Get_Surrogate_Server_From_%s;\n",
	  stn);
  fprintf(context->file, "  END;\n\n");

  fprintf(context->file,
	  "TYPE Lead__%s = Surrogate__%s BRANDED OBJECT\n",
	  stn, stn);
  list_enumerate(followers, (EnumProc) GenerateFollowerSlot, context);
  fprintf(context->file, "  OVERRIDES\n");
  fprintf(context->file, "    ILU_Qua_Type := Qua_From_Lead_%s\n", stn);
  fprintf(context->file, "  END;\n\n");

  list_enumerate(followers, (EnumProc) GenerateFollowerType, context);

  fprintf(context->file, "PROCEDURE Create_Lead__%s", stn);
  fprintf(context->file,
	  "(<*UNUSED*> self: IluRuntime.ObjectCreator;\n");
  fprintf(context->file,
  "    server : Ilu.SurrogateServer) : Ilu.Object RAISES {} =\n");
  fprintf(context->file, "  VAR x := NEW(Lead__%s", stn);
  list_enumerate(followers, (EnumProc) InitializeFollowerSlot, context);
  fprintf(context->file, ");\n");
  fprintf(context->file, "  BEGIN\n");
  list_enumerate(followers, (EnumProc) InitFollowers, context);
  fprintf(context->file, "    Init_Surrogate__%s (x, server);\n", stn);
  fprintf(context->file, "    RETURN x;\n");
  fprintf(context->file, "  END Create_Lead__%s;\n\n", stn);

  fprintf(context->file, "PROCEDURE Init_Surrogate__%s", stn);
  fprintf(context->file, "(self: Surrogate__%s;", stn);
  fprintf(context->file, "server : Ilu.SurrogateServer)\n");
  fprintf(context->file, "  RAISES {} =\n");
  fprintf(context->file, "  BEGIN\n");
  fprintf(context->file, "    self.ilu_is_surrogate := TRUE;\n");
  fprintf(context->file, "    self.server := server;\n");
  fprintf(context->file, "  END Init_Surrogate__%s;\n\n", stn);

  fprintf(context->file, "PROCEDURE Qua_From_Lead_%s", stn);
  fprintf(context->file, "(self: Lead__%s; ot : IluRuntime.ObjectType)",
	  stn);
  fprintf(context->file, ": Ilu.Object =\n");
  fprintf(context->file, "  BEGIN\n");
  fprintf(context->file, "    IF ot = ObjectType_%s THEN\n", stn);
  fprintf(context->file, "      RETURN self\n");
  UnmarkSupertypes(type);
  for (n = (Type) list_car(class_object(type)->superclasses);
  n != NULL; n = (Type) list_car(class_object(n)->superclasses)) {
    Type            un = ur_type(n);
    fprintf(context->file, "    ELSIF ot = %sObjectType_%s THEN\n",
	    SupportPrefix(un, context), M3_SHORT_TYPE_NAME(un));
    fprintf(context->file, "      RETURN self\n");
    n->marked = TRUE;
  }
  type->marked = TRUE;
  list_enumerate(followers, (EnumProc) ClimbQuaTree, context);
  fprintf(context->file, "    ELSE\n      RETURN NIL\n    END;\n");
  fprintf(context->file, "  END Qua_From_Lead_%s;\n\n", stn);

  fprintf(context->file, "PROCEDURE Get_Surrogate_Server_From_%s", stn);
  fprintf(context->file, "(self : Surrogate__%s): Ilu.Server =\n", stn);
  fprintf(context->file, "  BEGIN\n    RETURN self.server;\n");
  fprintf(context->file, "  END Get_Surrogate_Server_From_%s;\n\n", stn);
}

void DeclareCallerReturnValue (Type type, Context context)
{
  if (type_kind(type) != void_Type) {
    fprintf(context->file, "  VAR return_value_: %s;\n",
	    M3_TYPE_NAME(type));
  }
}

void ListExceptionName (Exception e, Context context)
{
  fprintf (context->file, ", %s", M3_EXCEPTION_NAME(e));
}

void ListArgument (Argument arg, Context context)
{
  char           *mode;
  switch (arg->direction) {
  case In:
    mode = "";
    break;
  case Out:
  case InOut:
    mode = "VAR ";
    break;
  }
  fprintf(context->file, "; %s%s: %s", mode, m3_argument_name(arg),
	  M3_TYPE_NAME(arg->type));
}

static void EncodeArgument (Argument arg, Context context)
{
  switch (arg->direction) {
  case Out:
    return;
  case In:
  case InOut:
    MarshallValue(arg->type, m3_argument_name(arg), context, 8);
  }
}

static void ReceiveOut (Argument arg, Context context)
{
  switch (arg->direction) {
  case In:
    return;
  case Out:
  case InOut:
    UnmarshallValue(context, arg->type, arg->def,
		    m3_argument_name(arg), 12, TRUE);
  }
}

static void RemoveFromList (Exception e, list l)
{
  list_remove(l, e);
}

static void AddToList (refany e, list l)
{
  list_insert(l, e);
}

static void list_copy (list to, list from)
{
  list_enumerate (from, (EnumProc) AddToList, to);
}

static void declare_unhandled_exceptions (list elist, Context context)
{
  list l = new_list();
  list_copy (l, context->interface->exceptions);
  list_enumerate (elist, (EnumProc) RemoveFromList, l);
  if (list_size(l) == 0)
    ;
  else
    {
      fprintf (context->file, "  <*FATAL %s", M3_EXCEPTION_NAME((Exception)list_car(l)));
      if (list_size(l) > 1)
	{
	  list c = (list) list_cdr(l);
	  list_enumerate(c, (EnumProc) ListExceptionName, context);
	  iluparser_Free(c);
	}
      fprintf (context->file, "*>\n");
    }
  list_clear (l, FALSE);
  iluparser_Free(l);
}

static void SizeArgument (Argument arg, Context context)
{
  switch (arg->direction) {
  case Out:
    return;
  case In:
  case InOut:
    fprintf(context->file, "\n          + ");
    SizeValue(arg->type, m3_argument_name(arg), context);
  }
}

static void generate_method_code (Procedure m, Context context)
{
  TypeKind t = type_kind(m->returnType);
  char *stn = M3_SHORT_TYPE_NAME(m->object);
  int             haveRet = m->returnType != NULL
  && t != void_Type;

  /* print out formal params */

  fprintf (context->file, "PROCEDURE Client__%s_%s (self_: %s",
	   stn, M3_SHORT_METHOD_NAME(m), stn);
  list_enumerate (m->arguments, (EnumProc) ListArgument, context);
  fprintf (context->file, ")%s%s\n",
	   t != void_Type ? ": " : "",
	   t != void_Type ? M3_TYPE_NAME(m->returnType) : "");
  fprintf (context->file, "  RAISES {IluBasics.Failed, Thread.Alerted");
  if (list_size(m->exceptions) > 0)
    list_enumerate(m->exceptions, (EnumProc) ListExceptionName, context);
  fprintf (context->file, "} =\n");
  declare_unhandled_exceptions (m->exceptions, context);

  /* declare local variables */
/*  DeclareCallerReturnValue(m->returnType, context); */
  fprintf (context->file, "  VAR call_: IluRuntime.Call;\n");
  if (NOT m->asynch)
    fprintf (context->file, "  VAR scode_: INTEGER;\n");

  /* if method is functional, see if it is cached, and if so, return the cached value */

  if (IsCacheableMethod(m))
    {
    }

  /* start code */
  fprintf (context->file, "  BEGIN\n");
  fprintf (context->file, "    call_ := IluRuntime.BeginCall (self_);\n");
  fprintf (context->file, "    IF call_ = NIL THEN\n");
  fprintf (context->file, "      RAISE IluBasics.Failed(NEW(IluBasics.Failure, info := \"lost connection\"))\n");
  fprintf (context->file, "    END;\n");
  fprintf (context->file, "    TRY\n");
  fprintf (context->file, "      TRY\n");
  fprintf (context->file, "        IluRuntime.BeginRequest(call_, ObjectType_%s, %s_x.Methods__%s[%s_x.MethodNames__%s.%s], 0",
	   stn,
	   m3_interface_name(m->object->interface), stn,
	   m3_interface_name(m->object->interface), stn,
	   M3_SHORT_METHOD_NAME(m));
  list_enumerate (m->arguments, (EnumProc) SizeArgument, context);
  if (NOT context->class->description->structuredDes.object->singleton)
    fprintf (context->file, "\n          + IluRuntime.SizeM3Object(call_, self_, TRUE, ObjectType_%s)",
	     M3_SHORT_TYPE_NAME(context->class));
  fprintf (context->file, ");\n");
  if (!context->class->description->structuredDes.object->singleton)
    fprintf (context->file, "        IluRuntime.OutputM3Object(call_, self_, TRUE, ObjectType_%s);\n",
	     M3_SHORT_TYPE_NAME(context->class));
  list_enumerate (m->arguments, (EnumProc) EncodeArgument, context);
  fprintf (context->file, "        IluRuntime.FinishRequest (call_);\n");
  fprintf (context->file, "      EXCEPT\n");
  fprintf (context->file, "        IluBasics.Failed(e) =>\n");
  fprintf (context->file, "          RAISE IluBasics.Failed(NEW(IluBasics.Failure, info := \"problem marshalling call\", subArg := e));\n");
  fprintf (context->file, "      END (* try *);\n");
  if (BlockingCall(m))
    {
      fprintf(context->file, "      TRY\n");
      fprintf(context->file, "        scode_ := IluRuntime.GetReply(call_);\n");
      fprintf(context->file, "        IF scode_ = 0 THEN\n");
      if (haveRet)
	fprintf(context->file, "          VAR returnvalue_ : %s;\n",
		M3_TYPE_NAME(m->returnType));
      fprintf(context->file, "          BEGIN\n");
      if (haveRet)
	UnmarshallValue(context, m->returnType, m->def, "returnvalue_",
			12, TRUE);
      list_enumerate(m->arguments, (EnumProc) ReceiveOut, context);
      fprintf(context->file, "            RETURN%s;\n",
	      haveRet ? " returnvalue_" : "");
      fprintf(context->file, "          END\n");
      fprintf(context->file, "        ELSE\n");
      if (list_size(m->exceptions) > 0)
	fprintf(context->file, "          %s_x.ReSignal_Exception(call_, scode_);\n",
		m3_interface_name(m->object->interface));
      fprintf(context->file, "            <*ASSERT FALSE*>\n");
      fprintf(context->file, "        END; (* if *)\n");
      fprintf(context->file, "      EXCEPT\n        IluBasics.Failed(e) =>\n");
      fprintf(context->file, "          RAISE IluBasics.Failed(NEW(IluBasics.Failure, info := \"reply problem\", subArg := e))\n");
      fprintf(context->file, "      END (* try *);\n");
    }
  fprintf (context->file, "    FINALLY\n      IluRuntime.FinishCall(call_);\n    END (* try-finally *);\n");
  fprintf(context->file, "  END Client__%s_%s;\n\n", stn,
	  M3_SHORT_METHOD_NAME(m));
}

static void generate_class_code (refany element, refany rock)
{
  Type            type = (Type) element;
  Context         context = (Context) rock;
  if (type->importInterfaceName == NULL
      && type_ur_kind(type) == object_Type) {
    context->class = type;

    GenerateObjectSupport(type, context);
    if (type->supertype != NULL)
      return;
    MethodIndex = 0;
    list_enumerate((class_object(type))->methods,
		   (EnumProc) generate_method_code, context);
  }
}

static void generate_client_includes (Context context)
{
  fprintf(context->file, "IMPORT Ilu; <*NOWARN*>\n");
  fprintf(context->file, "IMPORT IluBasics, IluRuntime, Thread;\n");
  fprintf(context->file, "IMPORT %s, %s_x; <*NOWARN*>\n",
	  m3_interface_name(context->interface),
	  m3_interface_name(context->interface));
  list_enumerate(context->interface->imports, (EnumProc) ImportInterface,
		 context);
  fprintf(context->file, "\n");
}

void generate_client_code (Interface i, FILE *file)
{
  struct context_s context;

  context.file = file;
  context.interface = i;
  context.class = NULL;
  context.module = ClientModule;

  generate_boilerplate(file, i);
  fprintf (file, "UNSAFE MODULE %s_c EXPORTS %s, %s_x;\n",
	   m3_interface_name(i), m3_interface_name(i),
	   m3_interface_name(i));
  generate_client_includes (&context);

  list_enumerate(i->types, generate_class_code, &context);

  fprintf(file, "BEGIN\n  %s_x.BeStarted();\nEND %s_c.\n",
	  m3_interface_name(i), m3_interface_name(i));
}
