/*
Copyright (c) 1991, 1992, 1993 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: headers.c,v 1.24 1995/03/03 23:39:16 janssen Exp $
*/
/* Last edited by Mike Spreitzer November 10, 1994 3:36 pm PST */

#include "m3.h"
#include <time.h>

#define IsHexDigit(x)	((((x)>='0')&&((x)<='9'))||(((x)>='a')&&((x)<='f'))||(((x)>='A')&&((x)<='F')))
#define HexValue(x)	(((x)<='9')?((x)-'0'):(((x)<='F')?((x)-'A'):(((x)<='f')?((x)-'a'):0)))

static boolean FirstItem;
static void sort_types_for_declaration (Type type,list sorted);

static void PrintStringChar (FILE *file, int quote, int ch)
{
  if (ch < ' ' || '~' < ch)
    fprintf (file, "\\%03o", ch);
  else
    {
      if (ch == quote || ch == '\\')
        putc ('\\', file);
      putc (ch, file);
    }
}

static void PrintString (FILE *file, const string str)
{
  unsigned char *p = (unsigned char *) str;
  int ch;

  putc ('"', file);
  while ((ch = *p++) != 0)
    PrintStringChar (file, '"', ch);
  putc ('"', file);
}

static void declare_constant_value (string cbname, string csname,
                                    Type ctype, ConstantValue cvalue,
                                    Interface cifc, Context context)
{
  TypeKind t = type_ur_kind(ctype);

  if ((t == cardinal_Type || t == shortcardinal_Type || t == byte_Type)
      && (cvalue->type == integer_Type)) {
    fprintf(context->file, "CONST %s : %s = %u;\n",
	    iname(csname, cifc),
	    (t == cardinal_Type) ? "CARDINAL"
	    : ((t == byte_Type) ? "[0..256]" : "[0..65535]"),
	    cvalue->val.i.value);
  } else if ((t == integer_Type || t == shortinteger_Type)
	     && (cvalue->type == integer_Type)) {
    fprintf(context->file, "CONST %s : %s = %s%u;\n",
	    iname(csname, cifc),
	    (t == integer_Type) ? "INTEGER" : "[-32768..32767]",
	    (cvalue->val.i.sign < 0) ? "-" : "",
	    cvalue->val.i.value);
  } else if (t == shortcharacter_Type
	     && (cvalue->type == integer_Type)) {
    if (cvalue->val.i.sign >= 0) {
      fprintf(context->file, "CONST %s = '",
	      iname(csname, cifc));
      PrintStringChar(context->file, '\'', cvalue->val.i.value);
      fprintf(context->file, "';\n");
    } else
      fprintf(stderr,
	      "Invalid constant, %s, encountered: negative ORD character.\n",
	      cbname);
  } else if (t == shortcharacter_Type
	     && (cvalue->type == shortcharacter_Type)) {
    fprintf(context->file, "CONST %s = '",
	    iname(csname, cifc));
    PrintStringChar(context->file, '\'', cvalue->val.s[0]);
    fprintf(context->file, "';\n");
  } else if ((t == real_Type || t == shortreal_Type)
	     && (cvalue->type == real_Type)) {
    fprintf(context->file, "CONST %s : %s = %s%s.%s%s%d;\n",
	    iname(csname, cifc),
	    (t == real_Type) ? "LONGREAL" : "REAL",
	    (cvalue->val.r.sign < 0) ? "-" : "",
	    cvalue->val.r.value,
	    (cvalue->val.r.fraction == NULL)
	    ? "0" : cvalue->val.r.fraction,
	    (t == real_Type) ? "D" : "E",
	    cvalue->val.r.exponent);
  } else if ((t == real_Type || t == shortreal_Type)
	     && (cvalue->type == integer_Type)) {
    fprintf(context->file, "CONST %s = FLOAT(%s%lu, %s);\n",
	    iname(csname, cifc),
	    (cvalue->val.i.sign < 0) ? "-" : "",
	    cvalue->val.i.value,
	    (t == real_Type) ? "LONGREAL" : "REAL");
  } else if (t == boolean_Type) {
    fprintf(context->file, "CONST %s : BOOLEAN = %s;\n",
	    iname(csname, cifc),
	    (cvalue->val.b) ? "TRUE" : "FALSE");
  } else if (t == sequence_Type
	     && (type_ur_kind(type_description(ctype)
				 ->structuredDes.sequence.type)
		 == shortcharacter_Type)
	     && cvalue->type == shortcharacter_Type) {
    fprintf(context->file, "CONST %s = ",
	    iname(csname, cifc));
    PrintString(context->file, cvalue->val.s);
    fprintf(context->file, ";\n");
  } else if (t == enumeration_Type
	     && cvalue->type == shortcharacter_Type) {
    /* This case needed at least for union arm constants. */
    fprintf(context->file, "CONST %s = %s.%s;\n", iname(csname, cifc),
	    trname(ctype), cvalue->val.s);
  } else
    fprintf(stderr,
	    "Invalid constant, %s, encountered.  Bad type for ILU constant.\n",
	    cbname);
}

static void declare_constant (Constant c, Context context)
{
  if (c->interface != context->interface)
    return;
  declare_constant_value(name_base_name(c->name), M3_CONSTANT_NAME(c),
			 c->type, c->value,
			 c->interface, context);
}

static boolean MatchPointer (refany p1, refany p2)
{
  return (p1 == p2);
}

static void SortArgTypes (Argument arg, list sorted)
{
  sort_types_for_declaration (arg->type, sorted);  
}

static void SortMethodTypes (Procedure m, list sorted)
{
  if (!m->returnOptional)
    sort_types_for_declaration (m->returnType, sorted);
  list_enumerate (m->arguments, (void (*) (void *, void *)) SortArgTypes, sorted);
}

static Interface SortContextInterface = NULL;

static void sort_types_for_declaration (Type type, list sorted)
{
  TypeKind t;
  static list     pending = NULL;

  if (type == NULL)
    return;

  if (pending == NULL)
    pending = new_list();

  t = type_ur_kind(type);

  if (type->interface != SortContextInterface
      || type->importInterfaceName != NULL)
    return;
  if (list_find(sorted, MatchPointer, type) != NULL
      || list_find(pending, MatchPointer, type) != NULL)
    return;
  list_insert(pending, type);
  if (type->importInterfaceName == NULL &&
      (t == record_Type || t == array_Type || t == union_Type
       || t == sequence_Type || t == object_Type)) {
    switch (t) {
    case record_Type:

      list_enumerate(type_description(type)->structuredDes.record,
		(void (*) (void *, void *)) SortArgTypes, sorted);
      break;

    case union_Type:

      list_enumerate(type_description(type)->structuredDes.uniond.types,
		(void (*) (void *, void *)) SortArgTypes, sorted);
      break;

    case array_Type:

      if (!type_description(type)->structuredDes.array.optional)
	sort_types_for_declaration(
	type_description(type)->structuredDes.array.type, sorted);
      break;

    case sequence_Type:

      sort_types_for_declaration(
      type_description(type)->structuredDes.sequence.type, sorted);
      break;

    case object_Type:

      if (type->marked)
	break;
      list_enumerate(class_object(type)->superclasses,
		     (void (*) (void *, void *)) sort_types_for_declaration, sorted);
      list_enumerate(class_object(type)->methods,
	     (void (*) (void *, void *)) SortMethodTypes, sorted);
      type->marked = TRUE;
      break;

    default:
      break;

    }
  }
  if (list_find(sorted, MatchPointer, type) == NULL)
    list_insert(sorted, type);
  list_remove(pending, type);
}

static void GenerateRecordField (Argument a, Context context)
{
  fprintf (context->file, "   %s : %s;\n",
	   m3_argument_name(a), trname(a->type));
}

static void GenerateRecordDeclaration (Type t, Context context)
{
  TypeDescription type_description();

  fprintf (context->file, "TYPE %s = RECORD\n", M3_SHORT_TYPE_NAME(t));
  list_enumerate (type_description(t)->structuredDes.record, (void (*) (void *, void *)) GenerateRecordField, context);
  fprintf (context->file, "END;\n");
}

struct double_s {
  Type t;
  Context c;
};

static void GenerateUnionField (Argument a, struct double_s *d)
{
  string          cn = UnionCaseName(d->t, a);
  fprintf(d->c->file, "TYPE  %s      = %s BRANDED OBJECT v: %s END;\n",
	  cn,
	  M3_SHORT_TYPE_NAME(d->t),
	  trname(a->type));
  if (list_size(a->values) == 1) {
    char            buf[1000];
    TypeDescription td = type_description(d->t);
    Type            disc = td->structuredDes.uniond.discriminator_type;
    ConstantValue   v = (ConstantValue) a->values->head->data;
    sprintf(buf, "%s__Tag", cn);
    declare_constant_value(buf, buf, disc, v, d->t->interface, d->c);
  }
  return;
}

static void GenerateUnionDeclaration (Type t, Context context)
{
  TypeDescription td = type_description(t);
  Type            disc = td->structuredDes.uniond.discriminator_type;
  list            e = td->structuredDes.uniond.types;
  Argument        da = td->structuredDes.uniond.default_arm;
  struct double_s s;

  fprintf(context->file, "TYPE  %s = BRANDED OBJECT d: %s END;",
	  M3_SHORT_TYPE_NAME(t),
	  trname(disc));
  fprintf(context->file, "  (* NIL not allowed *)\n");
  s.c = context;
  s.t = t;
  list_enumerate(e, (void (*) (void *, void *)) GenerateUnionField, &s);
  if (td->structuredDes.uniond.others_allowed)
    fprintf(context->file, "TYPE  %s_OTHERS = %s BRANDED OBJECT END;\n",
	    M3_SHORT_TYPE_NAME(t),
	    M3_SHORT_TYPE_NAME(t));
  return;
}

static void GenerateUnionFieldAlias (Argument a, struct double_s *d)
{
  Type            u = d->t->supertype;
  fprintf(d->c->file, "TYPE  %s", UnionCaseName(d->t, a));
  fprintf(d->c->file, "      = %s;\n", UnionCaseName(u, a));
  if (list_size(a->values) == 1) {
    fprintf(d->c->file, "CONST %s__Tag", UnionCaseName(d->t, a));
    fprintf(d->c->file, " = %s__Tag;\n", UnionCaseName(u, a));
  }
  return;
}

static void GenerateUnionAlias (Type t, Context context)
{
  TypeDescription td = type_description(t);
  Type            disc = td->structuredDes.uniond.discriminator_type;
  list            e = td->structuredDes.uniond.types;
  Argument        da = td->structuredDes.uniond.default_arm;
  struct double_s s;

  s.c = context;
  s.t = t;
  list_enumerate(e, (void (*) (void *, void *)) GenerateUnionFieldAlias,
		 &s);
  if (td->structuredDes.uniond.others_allowed)
    fprintf(context->file, "TYPE  %s_OTHERS = %s_OTHERS;\n",
	    M3_SHORT_TYPE_NAME(t),
	    trname(t->supertype));
  return;
}

static void PrintEnumField (EnumField e, Context context)
{
  if (FirstItem)
    FirstItem = FALSE;
  else
    fprintf (context->file, ",\n");

  fprintf (context->file, "  %s", m3_string(e->name));
}

static void GenerateEnumerationDeclaration (Type t, Context context)
{
  boolean oldFirst;

  fprintf (context->file, "TYPE %s = {\n", M3_SHORT_TYPE_NAME(t));
  context->class = t;
  oldFirst = FirstItem;  FirstItem = TRUE;
  list_enumerate (type_description(t)->structuredDes.enumeration, (EnumProc) PrintEnumField, context);
  FirstItem = oldFirst;
  fprintf (context->file, "};\n");
}

void ListArrayDimension (cardinal dim, Context context)
{
  fprintf (context->file, "ARRAY [0..%u] OF ", (dim < 1) ? 0 : dim - 1);
}

static void GenerateArrayDeclaration (Type array, Context context)
{
  Type atype = type_description(array)->structuredDes.array.type;

  fprintf (context->file, "TYPE %s = ", M3_SHORT_TYPE_NAME(array));
  context->class = array;
  list_enumerate (type_description(array)->structuredDes.array.dimensions, (EnumProc) ListArrayDimension, context);
  if (type_ur_kind(atype) == byte_Type)
    fprintf (context->file, "Ilu.PackedByte;\n");
  else if (type_ur_kind(atype) == shortcharacter_Type)
    fprintf (context->file, "Ilu.PackedShortChar;\n");
  else
    fprintf (context->file, "%s;\n", trname(atype));
}

static void GenerateOptionalDeclaration (Type t, Context context)
{
  TypeKind bt = type_ur_kind (type_description(t)->structuredDes.optional);
  Type otype = type_description(t)->structuredDes.optional;
  if (bt == object_Type OR bt == sequence_Type OR bt == union_Type)
  fprintf (context->file, "TYPE %s = %s;  (* Union of %s and NULL, so NIL allowed *)\n",
	   M3_SHORT_TYPE_NAME(t), trname(otype),
	   M3_SHORT_TYPE_NAME(otype));
  else
    fprintf (context->file, "TYPE %s = REF %s;  (* Union of %s and NULL, so NIL allowed *)\n",
	     M3_SHORT_TYPE_NAME(t), trname(otype),
	     M3_SHORT_TYPE_NAME(otype));
}

static void GenerateSequenceDeclaration (Type s, Context context)
{ 
  Type d = type_description(s)->structuredDes.sequence.type;
  if (type_ur_kind(d) == shortcharacter_Type)
    fprintf (context->file, "TYPE %s = TEXT;  (* NIL not allowed *)\n", M3_SHORT_TYPE_NAME(s));
  else if (type_ur_kind(d) == byte_Type)
    fprintf (context->file, "TYPE %s = REF ARRAY OF BITS 8 FOR Ilu.Byte;  (* NIL not allowed *)\n",
	     M3_SHORT_TYPE_NAME(s));
  else
    fprintf (context->file, "TYPE %s = REF ARRAY OF %s;  (* NIL not allowed *)\n", M3_SHORT_TYPE_NAME(s),
	     trname(d));
}

void declare_type (Type t, Context context)
{
  TypeKind tb;
#if 0
  printf("%s %d%s  file is 0x%x\n", type_name(t), list_size(t->refs),
	 t->builtIn ? " builtin" : "", context->file);
#endif
  if (t->builtIn)
    return;

  if (t->importInterfaceName != NULL)
    return;

  tb = type_ur_kind(t);
  if (t->supertype != NULL) {
    fprintf(context->file, "TYPE %s = %s;\n", M3_SHORT_TYPE_NAME(t),
	    trname(t->supertype));
    switch (tb) {
    case object_Type:
      declare_object_type(t, context);	/* see declare-object.c */
      break;
      
    case union_Type:
      GenerateUnionAlias(t, context);
      break;

    default:;			/* nothing to do */
    }
  } else
    switch (tb) {
    case void_Type:
      break;

    case integer_Type:
      fprintf(context->file, "TYPE %s = INTEGER;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case cardinal_Type:
      fprintf(context->file, "TYPE %s = CARDINAL;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case shortinteger_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.ShortInteger;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case shortcardinal_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.ShortCardinal;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case longinteger_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.LongInteger;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case longcardinal_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.LongCardinal;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case real_Type:
      fprintf(context->file, "TYPE %s = LONGREAL;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case shortreal_Type:
      fprintf(context->file, "TYPE %s = REAL;\n", M3_SHORT_TYPE_NAME(t));
      break;

    case longreal_Type:
      fprintf(context->file, "TYPE %s = LONGREAL;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case boolean_Type:
      fprintf(context->file, "TYPE %s = BOOLEAN;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case byte_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.Byte;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case character_Type:
      fprintf(context->file, "TYPE %s = IluRuntime.Character;\n",
	      M3_SHORT_TYPE_NAME(t));
      break;

    case shortcharacter_Type:
      fprintf(context->file, "TYPE %s = CHAR;\n", M3_SHORT_TYPE_NAME(t));
      break;

    case optional_Type:
      GenerateOptionalDeclaration(t, context);
      break;

    case array_Type:
      GenerateArrayDeclaration(t, context);
      return;

    case object_Type:
      declare_object_type(t, context);	/* see declare-object.c */
      break;

    case record_Type:
      GenerateRecordDeclaration(t, context);
      break;

    case union_Type:
      GenerateUnionDeclaration(t, context);
      break;

    case sequence_Type:
      GenerateSequenceDeclaration(t, context);
      break;

    case enumeration_Type:
      GenerateEnumerationDeclaration(t, context);
      break;

    default:
      fprintf(stderr,
	  "Error:  Can't cope with declaration of type %s yet.\n",
	      M3_TYPE_NAME(t));
      SEGFAULT;
    }
}

static void DeclareException (Exception e, Context context)
{
  if (e->importInterfaceName != NULL)
    return;

  fprintf (context->file, "EXCEPTION %s", M3_SHORT_EXCEPTION_NAME(e));
  if (e->type != NULL AND type_ur_kind(e->type) != void_Type)
    fprintf (context->file, " (%s)", trname(e->type));
  fprintf (context->file, ";\n");
}

static void generate_exception_table (Interface i, Context c)
{
  if (list_size(i->exceptions) > 0)
    {
      fprintf (c->file, "(* Exceptions *)\n\n");
      list_enumerate (i->exceptions, (EnumProc) DeclareException, c);
      fprintf (c->file, "\n");
    }
}

void generate_headers (Interface interface, FILE *file)
{
  struct context_s context;
  list            sorted = new_list();

  context.file = file;
  context.interface = interface;
  context.class = NULL;
  context.module = GeneralInterface;

  generate_boilerplate(file, interface);

  fprintf(file, "INTERFACE %s;\n\n", m3_interface_name(interface));
  fprintf(file, "IMPORT Ilu, IluBasics, Thread;\n");
  list_enumerate(interface->imports, (EnumProc) ImportInterface,
		 &context);
  fprintf(file, "\n");

  list_enumerate(interface->types, (EnumProc) UnmarkSupertypes, NULL);
  SortContextInterface = interface;
  list_enumerate(interface->types, (EnumProc) sort_types_for_declaration,
		 sorted);
  list_enumerate(sorted, (EnumProc) declare_type, &context);
  fprintf(file, "\n");

  list_enumerate(interface->constants, (EnumProc) declare_constant,
		 &context);
  fprintf(file, "\n");

  generate_exception_table(interface, &context);

  fprintf(file, "END %s.\n", m3_interface_name(interface));
}
